import { Node } from "./parser";
import { Function, Scope, Variable, ScopeElement, FunctionParameter } from "./scope";
import { Type } from "./types/";
import { TypeChecker } from './typecheck/';
import * as ssa from "./ssa";
import * as backend from "./backend/backend";
import { Package } from "./pkg";
/**
 * A DestructorInstruction tells which actions must be taken to destruct the result of an expression
 * once this result is no longer required.
 */
export declare class DestructorInstruction {
    constructor(v: ssa.Variable, t: Type, action: "none" | "decref" | "destruct" | "unlock");
    v: ssa.Variable;
    action: "none" | "decref" | "destruct" | "unlock";
    t: Type;
}
export declare class CodeGenerator {
    constructor(tc: TypeChecker, backend: backend.Backend, disableNullCheck: boolean);
    processModule(mnode: Node, emitIR: boolean, initPackages: Array<Package> | null, duplicateCodePackages: Array<Package> | null): string;
    getSSAType(t: Type, finalize?: boolean): ssa.Type | ssa.StructType | ssa.PointerType;
    private getSSAFunctionType;
    processFunction(f: Function, wf: backend.Function): ssa.Node;
    processScopeVariables(b: ssa.Builder, vars: Map<ScopeElement, ssa.Variable>, scope: Scope): void;
    freeScopeVariables(ignoreVariables: Array<Variable | FunctionParameter>, b: ssa.Builder, vars: Map<ScopeElement, ssa.Variable>, scope: Scope): void;
    processStatement(f: Function, scope: Scope, snode: Node, b: ssa.Builder, vars: Map<ScopeElement, ssa.Variable>, blocks: {
        body: ssa.Node;
        outer: ssa.Node;
    } | null): void;
    /**
     * Returns a local variable that can be assigned to or a pointer to a memory location that can be used to store a value.
     */
    processLeftHandExpression(f: Function, scope: Scope, enode: Node, b: ssa.Builder, vars: Map<ScopeElement, ssa.Variable>, dtor: Array<DestructorInstruction>, keepAlive: "none" | "hold" | "lock", noNullPointer?: boolean): ssa.Variable | ssa.Pointer;
    /**
     * Creates code that checks that a pointer-like value is non-null.
     * If a `ssa.Variable` is passed, the value of the variable is tested.
     * If a `ssa.Pointer` is passed, the value stored at the location pointed to is checked.
     */
    private processNullCheck;
    private processPureLiteral;
    private processPureLiteralInternal;
    /**
     * Returns true if the expression denoted by 'node' must be evaluted with processLeftHandSide.
     * For example in "a.b.c" the expression "a.b" must be processed with processLeftHandSide, because
     * otherwise we would get a copy of the struct denoted by "a.b".
     */
    isLeftHandSide(node: Node): boolean;
    private createInterfaceTable;
    private createInterfaceDescriptor;
    private createTypeDescriptor;
    autoConvertData(data: ssa.Variable | number | ssa.Pointer, targetType: Type, fromType: Type, b: ssa.Builder): ssa.Variable | number;
    /**
     * The same as processExpression except that it destructs/decrefs/unlocks all data structures that have been
     * increfed/locked while computing this expression.
     * This is safe if the type of the expression is a pure value type or if the expression is a take expression.
     * For all other expressions, the destruction might free the memory to which returned pointers are pointing.
     */
    private processValueExpression;
    /**
     * Returns a pointer to a value if possible and otherwise a variable that holds the value.
     * This is required for example by `expr[index]` if `expr` is an array.
     * It does not make sense to copy the array into a variable and then access one of its elements.
     * However, if `expr` is a function call that returns an array, then the array is in a variable anyway.
     * In this case it makes no sense to return a pointer to it.
     */
    private processInnerExpression;
    /**
     *
     * @param dtor is a list with destructor instructions that must be executed after the value computed by the expression
     *             is no longer required. processExpression will append to this list.
     * @param keepAlive determines whether pointers/slices returned by processExpression should either be locked or donated.
     *             processExpression will ignore keep alive if used on types that do not contain any pointers.
     *             Holding and locking both require that the returned pointer is pointing to allocated memory.
     *             Locking means in addition, that the data structure in this memory has not been destructed.
     *             The `dtor` carries information on how to release the holding or locking later.
     */
    private processExpression;
    private processFillZeros;
    private processCompare;
    private generateZeroStruct;
    private typecode;
    private isThis;
    private mangleDestructorName;
    private generateSliceDestructor;
    private generateTupleDestructor;
    private generateStructDestructor;
    /**
     * Generates a destructor that takes two arguments.
     * One is a pointer to the data, the other is the size.
     * This destructor can be used for slices, too, since a slice
     * is a pointer to an array which has a runtime-dependent size.
     *
     * The destructor does not free the memory used by the array.
     */
    private generateArrayDestructor;
    /**
     * Generates a destructor that takes one argument, which is
     * a pointer to the data
     * This destructor can be used for or-types, where the size of the array is known
     * and the destructor must have only one argument.
     *
     * The destructor does not free the memory used by the array.
     */
    private generateFixedArrayDestructor;
    private generatePointerDestructor;
    /**
     * Generates a destructor that takes one argument.
     * The argument is a pointer to the interface header.
     */
    private generateInterfaceDestructor;
    /**
     * Generates a destructor that takes two arguments.
     * The first is a pointer to the struct behind the interface.
     * The second is a pointer to the interface table.
     */
    /**
     * Generates a destructor that takes one argument.
     * The argument is a pointer to the or type struct.
     */
    private generateOrTypeDestructor;
    /**
     * Generate a destructor function that takes only one argument, which is a pointer
     * to the value that is to be destructed.
     *
     * If the type needs no destructors, the function returns null.
     */
    private generateDestructor;
    /**
     * pointer is the address of a value and t is the type of the value being pointed to.
     */
    private callDestructor;
    private callDestructorOnPointer;
    private callDestructorOnVariable;
    private scopeNeedsDestructors;
    /**
     * Determines whether the expression enode needs an incref before passing it as argument to a function call.
     * References stored in local variables on the stack do not need an incref, if no pointer to said local variables have been passed as arguments already.
     * The reason is that the callee cannot modify the stack variables of the caller.
     * Furthermore, references to objects owned directly via a strong pointer stored on the stack, do not need incref as well.
     * The reason is that local variables of the caller are not modified, hence said object must continue exist, because the local variable holds a strong pointer on it.
     */
    /**
     * Returns whether the expression yields a value that requires a cleanup via callDestructor().
     */
    private processDestructorInstructions;
    private processIncref;
    private processLock;
    private spawnNeedsWrapper;
    /**
     * This function is used to load a package that implements some built-in functionality
     * like decoding a string while iterating over it in a for-loop.
     *
     * Throws SyntaxError, ImportError or TypeError.
     */
    private loadPackage;
    private loadFunction;
    hasDestructors(): boolean;
    hasSymbols(): boolean;
    private backend;
    private tc;
    private imports;
    private funcs;
    private globalVars;
    private slicePointer;
    private localSlicePointer;
    private ifaceHeader;
    private mapHead;
    private disableNullCheck;
    private typeCodeMap;
    private destructors;
    private structs;
    private ifaceDescriptors;
    private symbols;
}
