import ast = require("./parser/ast");
import { Scope } from "./scope";
import { TypeChecker } from './typecheck';
import { CodeGenerator } from "./codegen";
export declare enum SystemCalls {
    heap = -1,
    currentMemory = -2,
    growMemory = -3,
    heapTypemap = -4,
    pageSize = -5,
    defaultStackSize = -6,
    stackPointer = -8,
    createMap = -15,
    setMap = -16,
    hashString = -17,
    lookupMap = -18,
    removeMapKey = -19,
    setNumericMap = -20,
    lookupNumericMap = -21,
    removeNumericMapKey = -22,
    abs32 = -23,
    abs64 = -24,
    sqrt32 = -25,
    sqrt64 = -26,
    trunc32 = -27,
    trunc64 = -28,
    nearest32 = -29,
    nearest64 = -30,
    ceil32 = -31,
    ceil64 = -32,
    floor32 = -33,
    floor64 = -34,
    max32 = -35,
    max64 = -36,
    min32 = -37,
    min64 = -38,
    copysign32 = -39,
    copysign64 = -49,
    continueCoroutine = -52,
    scheduleCoroutine = -53,
    coroutine = -54
}
export declare class Package {
    constructor(mainPackage?: boolean);
    sourcePath(): string;
    /**
     * Might throw ImportError
     */
    findSources(fyrPath: string, pkgPath: string): void;
    setSources(files: Array<string>, nativeFiles: Array<string>): void;
    /**
     * Might throw SyntaxError or ImportError or TypeError.
     * The function loads the sources, parses them and applies the first phase of type checking.
     */
    loadSources(): void;
    /**
     * Might throw TypeError
     */
    checkTypesPassTwo(): void;
    /**
     * Might throw TypeError
     */
    checkTypesPassThree(): void;
    /**
     * Might throw TypeError
     */
    checkTypesPassFour(): void;
    generateCode(backend: "C" | "WASM" | null, emitIR: boolean, initPackages: Array<Package> | null, duplicateCodePackages: Array<Package>, disableNullCheck: boolean): void;
    generateObjectFiles(backend: "C" | "WASM" | null, nativePackages: Array<Package>): void;
    /**
     * Might throw ImportError
     */
    private createObjFilePath;
    hasTemplateInstantiations(): boolean;
    /**
     * Checks the types of all packages imported so far.
     * Calling this function multiple times is ok, as it will not check the same package twice,
     * but it will check new imported packages.
     */
    static checkTypesForPackages(): void;
    /**
     * Generates C or WASM files and optionally compiles and links these files to create a native executable.
     */
    static generateCodeForPackages(backend: "C" | "WASM" | null, emitIR: boolean, emitNative: boolean, disableNullCheck: boolean): void;
    static getFyrPaths(): Array<string>;
    /**
     * @param pkgPath is of the form "/some/fyr/package".
     * @param loc is the location that is used for reporting an import error.
     *
     * Throws ImportError of the package could not be resolved.
     * Can throw TypeError or SyntaxError if loading of the sources detecs a syntax error.
     */
    static resolve(pkgPath: string, loc: ast.Location): Package;
    static registerPackage(p: Package): void;
    static clearPackages(): void;
    pkgNode: ast.Node;
    pkgPath: string;
    fyrPath: string;
    scope: Scope;
    tc: TypeChecker;
    files: Array<string>;
    nativeFiles: Array<string>;
    codegen: CodeGenerator;
    objFilePath: string;
    objFileName: string;
    binFilePath: string;
    binFileName: string;
    isInternal: boolean;
    isImported: boolean;
    hasMain: boolean;
    hasInitFunction: boolean;
    compileCmdLineArgs: Array<string>;
    linkCmdLineArgs: Array<string>;
    private typeCheckPass;
    /**
     * The package we are generating an executable or library for or null if
     * no such target is being built;
     */
    static mainPackage: Package | null;
    private static packagesByPath;
    private static packages;
    private static fyrPaths;
    static fyrBase: string;
}
