"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typecheck_1 = require("../typecheck");
const helper_1 = require("../typecheck/helper");
const errors_1 = require("../errors");
const Type_1 = require("./Type");
const TupleType_1 = require("./TupleType");
class FunctionType extends Type_1.Type {
    constructor() {
        super();
        this.callingConvention = "fyr";
        this.parameters = [];
    }
    parameterTypes() {
        let lst = [];
        for (let p of this.parameters) {
            lst.push(p.type);
        }
        return lst;
    }
    toString() {
        if (this.name) {
            if (this.objectType) {
                return this.objectType.toString() + "." + this.name;
            }
            return this.name;
        }
        let name = "(";
        for (let p of this.parameters) {
            if (name != "(") {
                name += ",";
            }
            if (p.ellipsis) {
                name += "...";
            }
            name += p.type.toString();
        }
        if (this.returnType) {
            name += ") => " + this.returnType.toString();
        }
        else {
            name += ") => undefined";
        }
        return name;
    }
    toTypeCodeString() {
        if (this.name) { // TODO: is this intentional?
            if (this.objectType) {
                return this.objectType.toTypeCodeString() + "." + this.name;
            }
            return this.name;
        }
        let name = "(";
        for (let p of this.parameters) {
            if (name != "(") {
                name += ",";
            }
            if (p.ellipsis) {
                name += "...";
            }
            name += p.type.toTypeCodeString();
        }
        name += ") => " + this.returnType.toTypeCodeString();
        return name;
    }
    hasEllipsis() {
        return (this.parameters.length > 0 && this.parameters[this.parameters.length - 1].ellipsis);
    }
    lastParameter() {
        return this.parameters[this.parameters.length - 1];
    }
    requiredParameterCount() {
        let i = 0;
        for (let t of this.parameters) {
            if (!t.ellipsis) {
                i++;
            }
        }
        return i;
    }
    createGroups() {
        let defaultGroup = new typecheck_1.Group(typecheck_1.GroupKind.Bound);
        let groups = new Map();
        let groupNames = new Map();
        for (let p of this.parameters) {
            if (p.type.groupName) {
                if (helper_1.isUnique(p.type)) {
                    throw new errors_1.TypeError("Unique pointers must not be marked with a group name", p.loc);
                }
                let g = groupNames.get(p.type.groupName);
                if (!g) {
                    g = new typecheck_1.Group(typecheck_1.GroupKind.Bound);
                    groupNames.set(p.type.groupName, g);
                }
                groups.set(p.name, g);
            }
            else {
                if (helper_1.isUnique(p.type)) {
                    groups.set(p.name, new typecheck_1.Group(typecheck_1.GroupKind.Free));
                }
                else {
                    groups.set(p.name, defaultGroup);
                }
            }
        }
        if (this.returnType) {
            if (this.returnType instanceof TupleType_1.TupleType) {
                for (let i = 0; i < this.returnType.types.length; i++) {
                    let t = this.returnType.types[i];
                    if (t.groupName) {
                        if (helper_1.isUnique(t)) {
                            throw new errors_1.TypeError("Unique pointers must not be marked with a group name", t.loc);
                        }
                        let g = groupNames.get(t.groupName);
                        if (!g) {
                            g = new typecheck_1.Group(typecheck_1.GroupKind.Bound);
                            groupNames.set(t.groupName, g);
                        }
                        groups.set("return " + i.toString(), g);
                    }
                    else {
                        if (helper_1.isUnique(t)) {
                            groups.set("return " + i.toString(), new typecheck_1.Group(typecheck_1.GroupKind.Free));
                        }
                        else {
                            groups.set("return " + i.toString(), defaultGroup);
                        }
                    }
                }
            }
            else {
                if (this.returnType.groupName) {
                    if (helper_1.isUnique(this.returnType)) {
                        throw new errors_1.TypeError("Unique pointers must not be marked with a group name", this.returnType.loc);
                    }
                    let g = groupNames.get(this.returnType.groupName);
                    if (!g) {
                        g = new typecheck_1.Group(typecheck_1.GroupKind.Bound);
                        groupNames.set(this.returnType.groupName, g);
                    }
                    groups.set("return", g);
                }
                else {
                    if (helper_1.isUnique(this.returnType)) {
                        groups.set("return", new typecheck_1.Group(typecheck_1.GroupKind.Free));
                    }
                    else {
                        groups.set("return", defaultGroup);
                    }
                }
            }
        }
        if (this.objectType) {
            groups.set("this", defaultGroup);
        }
        return groups;
    }
}
exports.FunctionType = FunctionType;
//# sourceMappingURL=FunctionType.js.map