"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const parser_1 = require("../parser");
const pkg_1 = require("../pkg");
const types_1 = require("../types/");
const group_1 = require("./group");
const scope_1 = require("../scope");
const errors_1 = require("../errors");
const crypto_1 = require("crypto");
const helper = require("./helper");
const Static_1 = require("./Static");
/**
 * Typedef represents the result of a 'type' statement, i.e.
 * a named type which is of course subject to a scope.
 */
class Typedef {
}
exports.Typedef = Typedef;
class TypeChecker {
    constructor(pkg) {
        /**
         * List of all interfaces. These are checked for possible errors after they have been defined.
         */
        this.ifaces = [];
        this.structs = [];
        this.templateTypeInstantiations = new Map();
        this.templateFunctionInstantiations = new Map();
        this.typedefs = [];
        this.functions = [];
        this.globalVariables = [];
        this.stringLiteralTypes = new Map();
        /**
         * During group checking, the compiler verifies that a variable that is taken in a statement,
         * is not used a second time in the statement.
         */
        this.modifiedVariabes = new Set();
        /**
         * During group checking, the compiler verifies that a variable that is taken in a statement,
         * is not used a second time in the statement.
         */
        this.usedVariables = new Set();
        if (!Static_1.Static.isInitialized()) {
            Static_1.Static.init();
            this.ifaces.push(Static_1.Static.t_error);
        }
        this.pkg = pkg;
        this.globalGroup = new group_1.Group(group_1.GroupKind.Bound, "$global");
    }
    createType(tnode, scope, mode = "default") {
        let t = this.createTypeIntern(tnode, scope, mode);
        if (tnode.groupName) {
            t.groupName = tnode.groupName.value;
        }
        return t;
    }
    createTypeIntern(tnode, scope, mode = "default") {
        let originalMode = mode;
        if (mode == "parameter_toplevel") {
            mode = "parameter";
        }
        if (mode == "variable_toplevel") {
            mode = "variable";
        }
        if (tnode.op == "basicType") {
            if (tnode.nspace) {
                let p = scope.resolveElement(tnode.nspace);
                if (!p) {
                    throw new errors_1.TypeError("Unknown package " + tnode.nspace, tnode.loc);
                }
                if (!(p.type instanceof types_1.PackageType)) {
                    throw new errors_1.TypeError(tnode.nspace + " is not a package", tnode.loc);
                }
                let t = p.type.pkg.scope.resolveType(tnode.value);
                if (!t) {
                    throw new errors_1.TypeError("Unknown type " + tnode.value + " in package " + p.type.pkg.pkgPath, tnode.loc);
                }
                return t;
            }
            let t = scope.resolveType(tnode.value);
            if (!t) {
                throw new errors_1.TypeError("Unknown type " + tnode.value, tnode.loc);
            }
            return t;
        }
        else if (tnode.op == "str") {
            let t = this.stringLiteralType(tnode.value);
            return t;
        }
        else if (tnode.op == "constType") {
            let c = this.createType(tnode.rhs, scope, originalMode);
            if (helper.isSafePointer(c)) {
                let ptr = types_1.RestrictedType.strip(c);
                ptr.elementType = helper.makeConst(ptr.elementType, tnode.loc);
                return c;
            }
            else if (helper.isSlice(c)) {
                let ptr = types_1.RestrictedType.strip(c);
                ptr.arrayType = helper.makeConst(ptr.arrayType, tnode.loc);
                return c;
            }
            // TODO: Map
            return helper.makeConst(c, tnode.loc);
        }
        else if (tnode.op == "pointerType") {
            let t = this.createType(tnode.rhs, scope, mode);
            return new types_1.PointerType(t, "strong");
        }
        else if (tnode.op == "uniquePointerType") {
            let c = this.createType(tnode.rhs, scope, mode);
            return new types_1.PointerType(c, "unique");
        }
        else if (tnode.op == "referenceType") {
            let c = this.createType(tnode.rhs, scope, mode);
            return new types_1.PointerType(c, "reference");
        }
        else if (tnode.op == "localReferenceType") {
            if (originalMode != "parameter_toplevel") {
                throw new errors_1.TypeError("A local reference is not allowed in this place", tnode.loc);
            }
            let c = this.createType(tnode.rhs, scope, mode);
            return new types_1.PointerType(c, "local_reference");
        }
        else if (tnode.op == "unsafePointerType") {
            let t = this.createType(tnode.rhs, scope, mode);
            if (helper.isInterface(t)) {
                throw new errors_1.TypeError("Unsafe pointers to interfaces are not possible", tnode.loc);
            }
            return new types_1.UnsafePointerType(t);
        }
        else if (tnode.op == "sliceType") {
            let t = this.createType(tnode.rhs, scope, mode);
            let s = new types_1.SliceType(t, "strong");
            if (tnode.value == "^[]") {
                s.mode = "unique";
            }
            else if (tnode.value == "~[]") {
                s.mode = "reference";
            }
            else if (tnode.value == "&[]") {
                if (originalMode != "parameter_toplevel") {
                    throw new errors_1.TypeError("A local reference slice is not allowed in this place", tnode.loc);
                }
                s.mode = "local_reference";
            }
            return s;
        }
        else if (tnode.op == "tupleType") {
            let types = [];
            for (let p of tnode.parameters) {
                let pt = this.createType(p, scope, mode);
                types.push(pt);
            }
            let t = new types_1.TupleType(types);
            return t;
        }
        else if (tnode.op == "arrayType") {
            if (tnode.lhs === null) {
                return new types_1.ArrayType(this.createType(tnode.rhs, scope, mode), -1);
            }
            this.checkExpression(tnode.lhs, scope);
            if (tnode.lhs.op != "int") {
                throw new errors_1.TypeError("Expected a constant number for array size", tnode.lhs.loc);
            }
            // TODO: Check range before parseInt
            return new types_1.ArrayType(this.createType(tnode.rhs, scope, mode), parseInt(tnode.lhs.value));
        }
        else if (tnode.op == "funcType") {
            let t = new types_1.FunctionType();
            t.loc = tnode.loc;
            if (tnode.parameters) {
                for (let pnode of tnode.parameters) {
                    var p = new scope_1.FunctionParameter();
                    if (pnode.op == "ellipsisParam") {
                        p.ellipsis = true;
                        pnode = pnode.lhs;
                    }
                    p.type = this.createType(pnode, scope, "parameter_toplevel");
                    if (helper.isReference(p.type) || helper.isLocalReference(p.type) || helper.isString(p.type)) {
                        p.isConst = true;
                    }
                    if (p.ellipsis && !(p.type instanceof types_1.SliceType)) {
                        throw new errors_1.TypeError("Ellipsis parameters must be of a slice type", pnode.loc);
                    }
                    this.checkVariableType(p.type, pnode.loc);
                    p.loc = pnode.loc;
                    t.parameters.push(p);
                }
            }
            if (tnode.rhs) {
                t.returnType = this.createType(tnode.rhs, scope, "parameter");
                this.checkVariableType(t.returnType, tnode.rhs.loc);
            }
            else {
                t.returnType = Static_1.Static.t_void;
            }
            return t;
        }
        else if (tnode.op == "mapType") {
            let k = this.createType(tnode.lhs, scope, mode);
            let v = this.createType(tnode.rhs, scope, mode);
            if (!helper.isIntNumber(k) && !helper.isString(k) && !helper.isSafePointer(k) && !helper.isUnsafePointer(k)) {
                throw new errors_1.TypeError("Map keys must be integers, strings, or pointers", tnode.loc);
            }
            return new types_1.MapType(k, v);
        }
        else if (tnode.op == "genericType") {
            let baset;
            if (tnode.nspace) {
                let p = scope.resolveElement(tnode.nspace);
                if (!p) {
                    throw new errors_1.TypeError("Unknown package " + tnode.nspace, tnode.loc);
                }
                if (!(p.type instanceof types_1.PackageType)) {
                    throw new errors_1.TypeError(tnode.nspace + " is not a package", tnode.loc);
                }
                baset = p.type.pkg.scope.resolveType(tnode.lhs.value);
                if (!baset) {
                    throw new errors_1.TypeError("Unknown type " + tnode.lhs.value + " in package " + p.type.pkg.pkgPath, tnode.loc);
                }
            }
            else {
                baset = scope.resolveType(tnode.lhs.value);
                if (!baset) {
                    throw new errors_1.TypeError("Unknown type " + tnode.lhs.value, tnode.loc);
                }
            }
            if (!(baset instanceof types_1.TemplateType)) {
                throw new errors_1.TypeError("Type " + baset.toString() + " is not a template type", tnode.loc);
            }
            let types = [];
            for (let i = 0; i < tnode.genericParameters.length; i++) {
                let t = this.createType(tnode.genericParameters[i], scope, mode);
                types.push(t);
            }
            return this.instantiateTemplateType(baset, types, tnode.loc, mode);
        }
        else if (tnode.op == "orType") {
            return this.createOrType(tnode, scope, null, mode);
        }
        else if (tnode.op == "andType") {
            return this.createInterfaceType(tnode, scope, null, mode);
        }
        else if (tnode.op == "structType") {
            return this.createStructType(tnode, scope, null, mode);
        }
        else if (tnode.op == "interfaceType" || tnode.op == "componentInterfaceType") {
            if (originalMode == "parameter_toplevel" || originalMode == "variable_toplevel") {
                throw new errors_1.TypeError("Interface types are not allowed in this place. Use a pointer to an interface instead", tnode.loc);
            }
            let iface = this.createInterfaceType(tnode, scope, null, mode);
            return new types_1.PointerType(iface, "strong");
        }
        else if (tnode.op == "copyType") {
            let t = this.createType(tnode.lhs, scope, mode);
            return this.createCopyType(t, tnode.loc);
        }
        else if (tnode.op == "opaqueType") {
            if (originalMode == "parameter_toplevel" || originalMode == "variable_toplevel") {
                throw new errors_1.TypeError("Opaque types are not allowed in this place. Use a pointer to the type instead", tnode.loc);
            }
            let s = new types_1.StructType();
            s.pkg = this.pkg;
            s.loc = tnode.loc;
            s.opaque = true;
            return s;
        }
        throw new errors_1.ImplementationError("type " + tnode.op);
    }
    createCopyType(t, loc) {
        if (helper.isUnsafePointer(t)) {
            return t;
        }
        if (helper.isSafePointer(t)) {
            if (helper.isStrong(t) || helper.isUnique(t)) {
                let stripped = types_1.RestrictedType.strip(t);
                let t2 = new types_1.PointerType(stripped.elementType, "reference");
                if (helper.isConst(t)) {
                    t2 = helper.makeConst(t2, loc);
                }
                return t2;
            }
            return t;
        }
        if (helper.isSlice(t)) {
            if (helper.isStrong(t) || helper.isUnique(t)) {
                let stripped = types_1.RestrictedType.strip(t);
                let t2 = new types_1.SliceType(stripped.arrayType, "reference");
                if (helper.isConst(t)) {
                    t2 = helper.makeConst(t2, loc);
                }
                return t2;
            }
            return t;
        }
        if (helper.isArray(t) || helper.isStruct(t) || helper.isTuple(t)) {
            if (helper.isPureValue(t)) {
                throw new errors_1.TypeError("The type " + t.toString() + " cannot be copied", loc);
            }
            return t;
        }
        if (helper.isNumber(t)) {
            return t;
        }
        throw new errors_1.ImplementationError();
    }
    createArrayType(tnode, scope, t, mode) {
        let e = this.createType(tnode.rhs, scope, mode ? mode : "default");
        t.elementType = e;
        return t;
    }
    createOrType(tnode, scope, t, mode) {
        // TODO: Avoid double entries
        if (!t) {
            t = new types_1.OrType();
        }
        for (let i = 0; i < tnode.parameters.length; i++) {
            let pnode = tnode.parameters[i];
            let pt = this.createType(pnode, scope, mode ? mode : "default");
            if (pt instanceof types_1.OrType) {
                t.types = t.types.concat(pt.types);
            }
            else {
                t.types.push(pt);
            }
        }
        return t;
    }
    createInterfaceType(tnode, scope, iface, mode) {
        if (!iface) {
            iface = new types_1.InterfaceType();
            iface.isComponent = (tnode.op == "componentInterfaceType");
            iface.pkg = this.pkg;
            iface.loc = tnode.loc;
            this.ifaces.push(iface);
        }
        iface.loc = tnode.loc;
        if (tnode.op == "andType") {
            for (let i = 0; i < tnode.parameters.length; i++) {
                let pnode = tnode.parameters[i];
                let pt = this.createType(pnode, scope, mode ? mode : "default");
                if (!(pt instanceof types_1.InterfaceType)) {
                    throw new errors_1.TypeError(pt.toString() + " is not an interface", pnode.loc);
                }
                iface.extendsInterfaces.push(pt);
            }
            return iface;
        }
        for (let mnode of tnode.parameters) {
            if (mnode.op == "extends") {
                let t = this.createType(mnode.rhs, scope, mode ? mode : "default");
                if (!(t instanceof types_1.InterfaceType)) {
                    throw new errors_1.TypeError(t.toString() + " is not an interface", mnode.loc);
                }
                iface.extendsInterfaces.push(t);
            }
        }
        for (let mnode of tnode.parameters) {
            if (mnode.op == "extends") {
                // Do nothing by intention
            }
            else if (mnode.op == "funcType") {
                let ft = this.createType(mnode, scope, mode ? mode : "default");
                ft.name = mnode.name.value;
                if (iface.methods.has(ft.name)) {
                    throw new errors_1.TypeError("Duplicate member name " + ft.name, mnode.loc);
                }
                iface.methods.set(ft.name, ft);
                let fscope = new scope_1.Scope(scope); // This scope is required for box names
                let ptrMode = "reference";
                if ((mnode.lhs.flags & parser_1.AstFlags.ReferenceObjectMember) == parser_1.AstFlags.ReferenceObjectMember) {
                    ptrMode = "local_reference";
                }
                let isConst = false;
                if (mnode.lhs.op == "constType") {
                    isConst = true;
                }
                ft.objectType = new types_1.PointerType(iface, ptrMode);
                if (isConst) {
                    ft.objectType = helper.makeConst(ft.objectType, mnode.loc);
                }
            }
            else {
                throw new errors_1.ImplementationError(mnode.op + " " + iface.name);
            }
        }
        return iface;
    }
    checkInterfaceType(iface) {
        if (iface._markChecked) {
            return;
        }
        iface._markChecked = true;
        let bases = iface.getAllBaseTypes();
        if (bases && bases.indexOf(iface) != -1) {
            throw new errors_1.TypeError("Interface " + iface.name + " is extending itself", iface.loc);
        }
        // Find all methods that exist in more than one inherited interface or are defined on this interface.
        // These method types must be equal
        let methodNames = new Map();
        let maps = new Map();
        maps.set(iface, iface.methods);
        for (let b of iface.extendsInterfaces) {
            maps.set(b, b.getAllMethods());
        }
        for (let m of maps.values()) {
            for (let key of m.keys()) {
                methodNames.set(key, true);
            }
        }
        for (let key of methodNames.keys()) {
            let ft = null;
            let ft_iface;
            for (let entry of maps.entries()) {
                if (entry[1].has(key)) {
                    if (ft) {
                        let ft2 = entry[1].get(key);
                        // Both functions must have the same signatures and either they are both const or both non-const (the only restriction possible on interface functions)
                        if (!this.checkTypeEquality(ft, ft2, iface.loc, false)) {
                            throw new errors_1.TypeError("Incompatible definition of " + key + " in " + ft_iface.toString() + " and " + entry[0].toString(), iface.loc);
                        }
                    }
                    else {
                        ft = entry[1].get(key);
                        ft_iface = entry[0];
                    }
                }
            }
        }
    }
    createStructType(tnode, scope, s, mode) {
        if (!s) {
            s = new types_1.StructType();
            s.pkg = this.pkg;
            s.loc = tnode.loc;
            this.structs.push(s);
        }
        if (tnode.op == "opaqueType") {
            return s;
        }
        for (let fnode of tnode.parameters) {
            if (fnode.op == "extends") {
                let ext = this.createType(fnode.rhs, scope, mode ? mode : "default");
                if (s.extends) {
                    throw new errors_1.TypeError("Struct cannot extend only one type", fnode.loc);
                }
                s.extends = ext;
                if (s.extends.name != "") {
                    if (s.field(s.extends.name)) {
                        throw new errors_1.TypeError("Duplicate field name " + s.extends.name, fnode.loc);
                    }
                }
                if (s.field("base", true)) {
                    throw new errors_1.TypeError("Duplicate field name base", fnode.loc);
                }
                let f = new types_1.StructField();
                f.name = "base";
                f.type = s.extends;
                s.fields.unshift(f);
            }
            else if (fnode.op == "implements") {
                let ext = this.createType(fnode.rhs, scope, mode ? mode : "default");
                if (!(ext instanceof types_1.InterfaceType)) {
                    throw new errors_1.TypeError(ext.toString() + " is not an interface type", tnode.rhs.loc);
                }
                s.implements.push(ext);
            }
            else if (fnode.op == "structField") {
                if (s.field(fnode.lhs.value, true)) {
                    throw new errors_1.TypeError("Duplicate field name " + fnode.lhs.value, fnode.lhs.loc);
                }
                // TODO: Check for duplicate names in the structs extends by this struct
                let field = new types_1.StructField();
                field.name = fnode.lhs.value;
                field.type = this.createType(fnode.rhs, scope, mode ? mode : "default");
                s.fields.push(field);
            }
            else {
                throw new errors_1.ImplementationError();
            }
        }
        return s;
    }
    checkStructType(s) {
        if (s._markChecked) {
            return;
        }
        s._markChecked = true;
        if (s.extends) {
            let bases = s.getAllBaseTypes();
            if (bases && bases.indexOf(s) != -1) {
                throw new errors_1.TypeError("Struct " + s.name + " is extending itself", s.loc);
            }
            if (s.extends instanceof types_1.StructType) {
                if (s.extends.opaque) {
                    throw new errors_1.TypeError("Cannot extend opaque type " + s.extends.toString(), s.loc);
                }
                this.checkStructType(s.extends);
                let inheritedMethods = s.extends.getAllMethodsAndFields();
                for (let key of s.methods.keys()) {
                    if (inheritedMethods.has(key)) {
                        throw new errors_1.TypeError("Method " + key + " shadows field or method of " + s.extends.toString(), s.loc);
                    }
                }
                for (let f of s.fields) {
                    if (inheritedMethods.has(f.name)) {
                        throw new errors_1.TypeError("Field " + f.name + " shadows field or method of " + s.extends.toString(), s.loc);
                    }
                }
            }
        }
        for (let iface of s.implements) {
            this.checkIsAssignableType(new types_1.PointerType(iface, "strong"), new types_1.PointerType(s, "strong"), s.loc, "assign", true);
        }
    }
    instantiateTemplateType(t, types, loc, mode) {
        // Check whether the template has already been instantiated with these type parameters
        let a = this.templateTypeInstantiations.get(t);
        if (a) {
            for (let s of a) {
                let ok = true;
                for (let k = 0; k < types.length; k++) {
                    let type = types[k];
                    let hasType = s.templateParameterTypes[k];
                    if (!this.checkTypeEquality(type, hasType, loc, false)) {
                        ok = false;
                        break;
                    }
                }
                if (ok) {
                    return s;
                }
            }
        }
        if (types.length != t.templateParameterNames.length) {
            throw new errors_1.TypeError("Mismatch in template parameter types while instantiating " + t.toString(), loc);
        }
        let scope = new scope_1.Scope(t.parentScope);
        for (let i = 0; i < t.templateParameterNames.length; i++) {
            if (t.templateParameterTypes[i]) {
                let tp = this.createType(t.templateParameterTypes[i], scope, mode);
                if (!(tp instanceof types_1.OrType)) {
                    throw new errors_1.TypeError("Template parameter type constraints must be an or'ed type", loc);
                }
                let ok = false;
                for (let o of tp.types) {
                    if (this.checkTypeEquality(o, types[i], loc, false)) {
                        ok = true;
                        break;
                    }
                }
                if (!ok) {
                    throw new errors_1.TypeError(types[i].toString() + " does not match with template parameter constraint " + tp.toString() + " of template parameter " + t.templateParameterNames[i] + " of template type " + t.name, loc);
                }
            }
            scope.registerType(t.templateParameterNames[i], types[i]);
        }
        let node = t.node.rhs.clone();
        if (node.op == "structType") {
            let s = new types_1.TemplateStructType();
            s.pkg = this.pkg;
            s.base = t;
            s.name = t.name;
            s.loc = t.loc;
            s.templateParameterTypes = types;
            if (a) {
                a.push(s);
            }
            else {
                this.templateTypeInstantiations.set(t, [s]);
            }
            this.createStructType(node, scope, s);
            for (let m of t.methods) {
                this.instantiateTemplateMemberFunction(t, s, m);
            }
            return s;
        }
        else if (node.op == "interfaceType" || node.op == "andType" || node.op == "componentInterfaceType") {
            let s = new types_1.TemplateInterfaceType();
            s.base = t;
            s.name = t.name;
            s.loc = t.loc;
            s.templateParameterTypes = types;
            if (a) {
                a.push(s);
            }
            else {
                this.templateTypeInstantiations.set(t, [s]);
            }
            this.createInterfaceType(node, scope, s);
            return s;
        }
        else if (node.op == "funcType") {
            let s = new types_1.TemplateFunctionType();
            s.base = t;
            s.name = t.name;
            s.loc = t.loc;
            s.templateParameterTypes = types;
            if (a) {
                a.push(s);
            }
            else {
                this.templateTypeInstantiations.set(t, [s]);
            }
            throw new errors_1.TodoError();
        }
        else if (node.op == "orType") {
            throw new errors_1.TodoError();
        }
        throw new errors_1.ImplementationError();
    }
    instantiateTemplateMemberFunction(t, s, m) {
        let scope = new scope_1.Scope(t.parentScope);
        // TODO: Register the fully qualified name, too
        // scope.registerType(s.name, s);
        for (let i = 0; i < t.templateParameterNames.length; i++) {
            scope.registerType(t.templateParameterNames[i], s.templateParameterTypes[i]);
        }
        let node = m.node.clone();
        let f = this.createFunction(node, scope, this.moduleNode.scope, null, null, s);
        if (f instanceof scope_1.Function) {
            f.isTemplateInstance = true;
            this.checkFunctionBody(f);
        }
        return f;
    }
    /**
     * Parses the instantiation of a template function, e.g. in "max<int>(4,5)" this function parses "max<int>".
     */
    instantiateTemplateFunctionFromNode(tnode, scope) {
        if (tnode.op != "genericInstance") {
            throw new errors_1.ImplementationError();
        }
        let baset = tnode.lhs.type;
        // Is the type a template type?
        if (!(baset instanceof types_1.TemplateType)) {
            throw new errors_1.TypeError("Type " + baset.toString() + " is not a template function", tnode.loc);
        }
        // Is the type a function template type?
        if (baset.node.op != "func" && baset.node.op != "export_func") {
            throw new errors_1.TypeError("Type " + baset.toString() + " is not a template function", tnode.loc);
        }
        if (tnode.genericParameters.length != baset.templateParameterNames.length) {
            throw new errors_1.TypeError("Mismatch in template parameter types while instantiating " + baset.toString(), tnode.loc);
        }
        // Parse the types in brackets as in max<int>(4,5)
        let types = [];
        for (let i = 0; i < tnode.genericParameters.length; i++) {
            let pnode = tnode.genericParameters[i];
            let pt = this.createType(pnode, scope, "parameter");
            pnode.type = pt;
            types.push(pt);
        }
        return this.instantiateTemplateFunction(baset, types, tnode.loc);
    }
    /**
     * Instantiates a template function.
     */
    instantiateTemplateFunction(t, types, loc) {
        // Check whether this function has already been instantiated.
        let a = this.templateFunctionInstantiations.get(t);
        if (a) {
            for (let f of a) {
                let ok = true;
                for (let k = 0; k < types.length; k++) {
                    let type = types[k];
                    let hasType = f.type.templateParameterTypes[k];
                    if (!this.checkTypeEquality(type, hasType, loc, false)) {
                        ok = false;
                        break;
                    }
                }
                if (ok) {
                    return f;
                }
            }
        }
        // Check whether the template instantiation is consistent with the template parameter types.
        // Register the template parameters as types.
        // Non-exported templates can see package variables via t.parentScope. Exported templates cannot do this.
        // TODO: Need a vanilla scope
        // let scope = new Scope(t.node.op == "export_func" ? null : t.parentScope);
        let scope = new scope_1.Scope(t.parentScope);
        for (let i = 0; i < t.templateParameterNames.length; i++) {
            if (t.templateParameterTypes[i]) {
                let tp = this.createType(t.templateParameterTypes[i], scope, "parameter");
                if (!(tp instanceof types_1.OrType)) {
                    throw new errors_1.TypeError("Template parameter type constraints must be an or'ed type", loc);
                }
                let ok = false;
                for (let o of tp.types) {
                    if (this.checkTypeEquality(o, types[i], loc, false)) {
                        ok = true;
                        break;
                    }
                }
                if (!ok) {
                    throw new errors_1.TypeError(types[i].toString() + " does not match with template parameter constraint " + tp.toString() + " of parameter " + t.templateParameterNames[i] + " of template function " + t.name, loc);
                }
            }
            scope.registerType(t.templateParameterNames[i], types[i]);
        }
        // Create a copy the template AST and parse the template function's type signature.
        // Store the type, so that the same template instantiation does not occur twice.
        let node = t.node.clone();
        let f = this.createFunction(node, scope, this.moduleNode.scope, t, types);
        if (!(f instanceof scope_1.Function)) {
            throw new errors_1.ImplementationError();
        }
        //        if (!(f.type instanceof TemplateFunctionType)) {
        //            throw new ImplementationError()
        //        }
        //        f.isTemplateInstance = true;
        //        f.type.base = t;
        //        f.type.templateParameterTypes = types;
        if (a) {
            a.push(f);
        }
        else {
            this.templateFunctionInstantiations.set(t, [f]);
        }
        // Typecheck the template function body.
        this.checkFunctionBody(f);
        return f;
    }
    // TODO: move to helper.ts
    static mangleTemplateParameters(types) {
        let str = "<";
        for (let g of types) {
            str += g.toTypeCodeString() + ",";
        }
        str += ">";
        let hash = crypto_1.createHash("md5");
        hash.update(str);
        return "_" + hash.digest("hex");
    }
    // TODO: move to helper.ts
    static mangledTypeName(t) {
        if (t instanceof types_1.TemplateStructType) {
            return t.base.pkg.pkgPath + "/" + t.name + TypeChecker.mangleTemplateParameters(t.templateParameterTypes);
        }
        if (t instanceof types_1.StructType) {
            return t.pkg.pkgPath + "/" + t.name;
        }
        return t.name;
    }
    createFunction(fnode, parentScope, registerScope, templateBase = null, templateParameterTypes = null, templateTypeInstance = null) {
        if (!fnode.name) {
            throw new errors_1.TypeError("Function must be named", fnode.loc);
        }
        let objectType;
        let structType;
        let templateType;
        // A member function?
        if (fnode.lhs) {
            var obj;
            if (templateTypeInstance) {
                let s = new scope_1.Scope(parentScope);
                s.registerType(templateTypeInstance.name, templateTypeInstance);
                obj = this.createType(fnode.lhs, s, "parameter");
            }
            else {
                obj = this.createType(fnode.lhs, parentScope, "parameter");
            }
            let obj2 = types_1.RestrictedType.strip(obj);
            if (obj2.name == "") {
                throw new errors_1.TypeError(obj.toString() + " is not a named struct", fnode.lhs.loc);
            }
            if (obj2 instanceof types_1.StructType) {
                structType = obj2;
            }
            else if (obj2 instanceof types_1.TemplateType) {
                templateType = obj2;
            }
            else {
                throw new errors_1.TypeError(obj.toString() + " is not a named struct", fnode.lhs.loc);
            }
            let mode = "reference";
            if ((fnode.lhs.flags & parser_1.AstFlags.ReferenceObjectMember) == parser_1.AstFlags.ReferenceObjectMember) {
                mode = "local_reference";
            }
            objectType = new types_1.PointerType(obj, mode);
        }
        let f;
        if ((fnode.genericParameters && !templateBase) || helper.isTemplateType(templateType)) {
            f = new scope_1.TemplateFunction();
            f.node = fnode;
            if (templateType) {
                f.owner = templateType;
                templateType.methods.push(f);
            }
            f.name = fnode.name.value;
        }
        else if (templateBase) {
            f = new scope_1.Function();
            fnode.scope = f.scope;
            let tt = new types_1.TemplateFunctionType();
            tt.templateParameterTypes = templateParameterTypes;
            tt.base = templateBase;
            f.type = tt;
            f.isTemplateInstance = true;
            f.name = templateBase.pkg.pkgPath + "/" + fnode.name.value + TypeChecker.mangleTemplateParameters(templateParameterTypes);
        }
        else {
            f = new scope_1.Function();
            fnode.scope = f.scope;
            f.type = new types_1.FunctionType();
            f.name = fnode.name.value;
        }
        f.node = fnode;
        f.loc = fnode.loc;
        f.isExported = (fnode.op == "export_func");
        if (f instanceof scope_1.TemplateFunction) {
            let gt = new types_1.TemplateType();
            gt.pkg = this.pkg;
            gt.name = fnode.name.value;
            gt.node = fnode;
            gt.parentScope = parentScope;
            gt.registerScope = registerScope;
            // let scope = new Scope(parentScope);
            if (fnode.genericParameters && !templateBase) {
                for (let g of fnode.genericParameters) {
                    gt.templateParameterTypes.push(g.condition ? g.condition : null);
                    gt.templateParameterNames.push(g.value);
                }
            }
            f.type = gt;
            f.type.loc = fnode.loc;
            if (!templateType) {
                registerScope.registerElement(f.name, f);
            }
            // Do not process any further. This is done upon template instantiation
            return f;
        }
        f.scope.parent = parentScope;
        f.type.loc = fnode.loc;
        // A member function?
        if (objectType) {
            f.type.objectType = objectType;
            let p = new scope_1.FunctionParameter();
            p.name = "this";
            p.loc = fnode.lhs.loc;
            p.type = objectType;
            p.isConst = true;
            f.scope.registerElement("this", p);
        }
        if (fnode.parameters) {
            for (let pnode of fnode.parameters) {
                let original_pnode = pnode;
                var p = new scope_1.FunctionParameter();
                if (pnode.op == "ellipsisParam") {
                    // TODO: Ellipsis must be the last parameter
                    p.ellipsis = true;
                    pnode = pnode.lhs;
                }
                p.name = pnode.name.value;
                for (let param of f.type.parameters) {
                    if (param.name == p.name) {
                        throw new errors_1.TypeError("Duplicate parameter name " + p.name, pnode.loc);
                    }
                }
                p.type = this.createType(pnode, f.scope, "parameter_toplevel");
                if (helper.isReference(p.type) || helper.isLocalReference(p.type) || helper.isString(p.type)) {
                    p.isConst = true;
                }
                if (p.ellipsis && (!(p.type instanceof types_1.SliceType) || p.type.mode != "local_reference")) {
                    throw new errors_1.TypeError("Ellipsis parameters must be of a local reference slice type, i.e. &[]", pnode.loc);
                }
                this.checkVariableType(p.type, pnode.loc);
                p.loc = pnode.loc;
                f.type.parameters.push(p);
                f.scope.registerElement(p.name, p);
            }
        }
        f.hasNamedReturnVariables = false;
        // A return type?
        if (fnode.rhs) {
            f.type.returnType = this.createType(fnode.rhs, f.scope, "parameter");
            if (fnode.rhs.op == "tupleType") {
                for (let i = 0; i < fnode.rhs.parameters.length; i++) {
                    let pnode = fnode.rhs.parameters[i];
                    if (i == 0) {
                        f.hasNamedReturnVariables = !!pnode.name;
                    }
                    else if ((f.hasNamedReturnVariables && !pnode.name) || (!f.hasNamedReturnVariables && !!pnode.name)) {
                        throw new errors_1.TypeError("Either all or no return variables are named", pnode.loc);
                    }
                }
            }
            if (fnode.rhs.op == "tupleType") {
                for (let i = 0; i < fnode.rhs.parameters.length; i++) {
                    let pnode = fnode.rhs.parameters[i];
                    let v = new scope_1.Variable();
                    v.isResult = true;
                    v.loc = pnode.loc;
                    if (pnode.name) {
                        v.name = pnode.name.value;
                    }
                    else {
                        v.name = "return " + i.toString();
                    }
                    v.type = f.type.returnType.types[i];
                    f.scope.registerElement(v.name, v);
                    if (!f.namedReturnVariables) {
                        f.namedReturnVariables = [];
                    }
                    f.namedReturnVariables.push(v);
                    f.type.returnType.types[i] = v.type;
                }
            }
            else {
                let v = new scope_1.Variable();
                v.isResult = true;
                v.loc = fnode.loc;
                v.name = "return";
                v.type = f.type.returnType;
                f.unnamedReturnVariable = v;
            }
            this.checkVariableType(f.type.returnType, fnode.rhs.loc);
        }
        else {
            f.type.returnType = Static_1.Static.t_void;
        }
        // The function is a member function
        if (structType) {
            if (structType.methods.has(f.name)) {
                let loc = structType.methods.get(f.name).loc;
                throw new errors_1.TypeError("Method " + structType.toString() + "." + f.name + " is already defined at " + loc.file + " (" + loc.start.line + "," + loc.start.column + ")", fnode.loc);
            }
            if (structType.field(f.name)) {
                throw new errors_1.TypeError("Field " + structType.toString() + "." + f.name + " is already defined", fnode.loc);
            }
            structType.methods.set(f.name, f.type);
            registerScope.registerElement(TypeChecker.mangledTypeName(structType) + "." + f.name, f);
        }
        else {
            registerScope.registerElement(f.name, f);
            registerScope.setGroup(f, new group_1.Group(group_1.GroupKind.Free));
        }
        return f;
    }
    createVar(vnode, scope, needType = true, isConst = false, isGlobal = false) {
        let v = new scope_1.Variable();
        v.loc = vnode.loc;
        v.name = vnode.value;
        v.isGlobal = isGlobal;
        v.isConst = isConst;
        if (!vnode.rhs) {
            if (needType) {
                throw new errors_1.TypeError("Variable declaration of " + vnode.value + " without type information", vnode.loc);
            }
        }
        else {
            v.type = this.createType(vnode.rhs, scope, "variable_toplevel");
            this.checkVariableType(v.type, vnode.loc);
        }
        if (v.name != "_") {
            scope.registerElement(v.name, v);
        }
        return v;
    }
    createTypedef(tnode, scope) {
        let t = new Typedef();
        t.name = tnode.name.value;
        t.node = tnode;
        t.scope = scope;
        if (t.node.genericParameters) {
            let tmpl = new types_1.TemplateType();
            tmpl.pkg = this.pkg;
            tmpl.node = tnode;
            tmpl.name = t.name;
            tmpl.loc = tnode.loc;
            tmpl.parentScope = scope;
            tmpl.registerScope = scope;
            for (let g of tnode.genericParameters) {
                tmpl.templateParameterNames.push(g.value);
                tmpl.templateParameterTypes.push(g.condition ? g.condition : null);
            }
            t.type = tmpl;
            scope.registerType(t.name, tmpl, tnode.loc);
        }
        else if (t.node.rhs.op == "structType" || t.node.rhs.op == "opaqueType") {
            let s = new types_1.StructType();
            s.pkg = this.pkg;
            s.loc = t.node.loc;
            s.name = t.name;
            this.structs.push(s);
            t.type = s;
            scope.registerType(t.name, s, tnode.loc);
        }
        else if (t.node.rhs.op == "interfaceType" || t.node.rhs.op == "andType" || t.node.rhs.op == "componentInterfaceType") {
            let iface = new types_1.InterfaceType();
            iface.pkg = this.pkg;
            iface.loc = t.node.loc;
            iface.name = t.name;
            iface.isComponent = (t.node.rhs.op == "interfaceType");
            this.ifaces.push(iface);
            t.type = iface;
            scope.registerType(t.name, iface, tnode.loc);
        }
        else if (t.node.rhs.op == "orType") {
            let newt = new types_1.OrType();
            newt.loc = t.node.loc;
            newt.name = t.name;
            t.type = newt;
            scope.registerType(t.name, newt, tnode.loc);
        }
        else if (t.node.rhs.op == "arrayType") {
            let newt = new types_1.ArrayType(null, parseInt(t.node.rhs.lhs.value));
            newt.loc = t.node.loc;
            newt.name = t.name;
            t.type = newt;
            scope.registerType(t.name, newt, tnode.loc);
        }
        else {
            throw new errors_1.TypeError("A type must be a struct, interface, array-type, or or-type", tnode.loc);
        }
        return t;
    }
    createImport(inode, scope) {
        if (inode.rhs.op == "importNative") {
            let ip;
            let importPath = inode.rhs.rhs.value;
            if (!inode.lhs) {
                // Syntax of the kind: import "<include.h>" { func ... }
                let name;
                if (inode.name) {
                    name = inode.name.value;
                }
                else {
                    let importPathElements = importPath.split("/");
                    // TODO: Sanitize the name
                    name = importPathElements[importPathElements.length - 1];
                }
                let pkg = new pkg_1.Package();
                pkg.scope = new scope_1.Scope(null);
                // TODO: Sanitize the name
                ip = new scope_1.ImportedPackage(name, pkg, inode.loc);
                scope.registerElement(name, ip);
            }
            else if (inode.lhs.op == "id") {
                // Syntax of the kind: import identifier from "<include.h>" { func ... }
                if (inode.name) {
                    throw new errors_1.TypeError("An import selection must not be used together with an import alias", inode.name.loc);
                }
                let pkg = new pkg_1.Package();
                pkg.scope = new scope_1.Scope(null);
                ip = new scope_1.ImportedPackage(inode.lhs.value, pkg, inode.loc);
                scope.registerElement(ip.name, ip);
            }
            else if (inode.lhs.op == ".") {
                // Syntax of the kind: import . from "<include.h>" { func ... }
                if (inode.name) {
                    throw new errors_1.TypeError("An import selection must not be used together with an import alias", inode.name.loc);
                }
                // Do nothing by intention
            }
            else if (inode.lhs.op == "identifierList") {
                // Syntax of the kind: import {id1, id2, ...} from "path/to/module"
                if (inode.name) {
                    throw new errors_1.TypeError("An import selection must not be used together with an import alias", inode.name.loc);
                }
                // Do nothing by intention
            }
            else {
                throw new errors_1.ImplementationError("import lhs " + inode.lhs.op);
            }
        }
        else {
            let importPath = inode.rhs.value;
            let p = pkg_1.Package.resolve(importPath, inode.rhs.loc);
            let ip;
            if (!inode.lhs) {
                // Syntax of the kind: import "path/to/module"
                let name;
                if (inode.name) {
                    name = inode.name.value;
                }
                else {
                    let importPathElements = importPath.split("/");
                    // TODO: Sanitize the name
                    name = importPathElements[importPathElements.length - 1];
                }
                ip = new scope_1.ImportedPackage(name, p, inode.loc);
                scope.registerElement(name, ip);
            }
            else if (inode.lhs.op == "identifierList") {
                // Syntax of the kind: import {id1, id2, ...} from "path/to/module"
                if (inode.name) {
                    throw new errors_1.TypeError("An import selection must not be used together with an import alias", inode.name.loc);
                }
            }
            else if (inode.lhs.op == "id") {
                // Syntax of the kind: import identifier from "path/to/module"
                ip = new scope_1.ImportedPackage(inode.lhs.value, p, inode.loc);
                scope.registerElement(ip.name, ip);
            }
            else if (inode.lhs.op == ".") {
                // Syntax of the kind: import . from "path/to/module"
            }
            else {
                throw new errors_1.ImplementationError("import lhs " + inode.lhs.op);
            }
        }
    }
    createNativeConstImport(nativePackageName, node, scope) {
        let v = new scope_1.Variable();
        v.nativePackageName = nativePackageName;
        v.name = node.name.value;
        v.node = node;
        v.loc = node.loc;
        v.type = this.createType(node.lhs, scope, "variable_toplevel");
        v.isConst = true;
        v.isGlobal = true;
        scope.registerElement(v.name, v);
        // Imported variables are considered to be free.
        // This can lead to race conditions.
        scope.setGroup(v, new group_1.Group(group_1.GroupKind.Free));
        return v;
    }
    createNativeFunctionImport(nativePackageName, fnode, scope) {
        let f = new scope_1.Function();
        f.nativePackageName = nativePackageName;
        f.name = fnode.name.value;
        f.scope.parent = scope;
        f.node = fnode;
        f.loc = fnode.loc;
        f.type = new types_1.FunctionType();
        f.type.callingConvention = "native";
        f.type.loc = fnode.loc;
        //        f.type.callingConvention = "system";
        let i = 0;
        if (fnode.parameters) {
            for (let pnode of fnode.parameters) {
                let p = new scope_1.FunctionParameter();
                p.name = "p" + i.toString();
                i++;
                p.type = this.createType(pnode, f.scope, "parameter_toplevel");
                p.loc = pnode.loc;
                f.type.parameters.push(p);
                f.scope.registerElement(p.name, p);
            }
        }
        if (fnode.rhs) {
            f.type.returnType = this.createType(fnode.rhs, f.scope, "parameter");
            if (fnode.rhs.op == "tupleType") {
                for (let i = 0; i < fnode.rhs.parameters.length; i++) {
                    let pnode = fnode.rhs.parameters[i];
                    if (pnode.name) {
                        let v = new scope_1.Variable();
                        v.isResult = true;
                        v.loc = pnode.loc;
                        v.name = "p" + i.toString();
                        i++;
                        v.name = pnode.name.value;
                        v.type = f.type.returnType.types[i];
                        f.scope.registerElement(v.name, v);
                        if (!f.namedReturnVariables) {
                            f.namedReturnVariables = [];
                        }
                        f.namedReturnVariables.push(v);
                        f.type.returnType.types[i] = v.type;
                    }
                }
            }
            else {
                f.type.returnType = f.type.returnType;
                let v = new scope_1.Variable();
                v.isResult = true;
                v.loc = fnode.loc;
                v.name = "return";
                v.type = f.type.returnType;
                f.unnamedReturnVariable = v;
            }
        }
        else {
            f.type.returnType = Static_1.Static.t_void;
        }
        scope.registerElement(f.name, f);
        scope.setGroup(f, new group_1.Group(group_1.GroupKind.Free));
        return f;
    }
    importTypes(inode, scope) {
        if (inode.rhs.op == "importNative") {
            let ip;
            let importPath = inode.rhs.rhs.value;
            if (!inode.lhs) {
                // Syntax of the kind: import "native/path" { func ... }
                let importPathElements = importPath.split("/");
                let name = importPathElements[importPathElements.length - 1];
                // TODO: Sanitize the name
                let e = scope.resolveElement(name);
                if (!(e instanceof scope_1.ImportedPackage)) {
                    throw new errors_1.ImplementationError();
                }
                ip = e;
            }
            else if (inode.lhs.op == "id") {
                // Syntax of the kind: import identifier from "native/path" { func ... }
                // TODO: Sanitize the name
                let e = scope.resolveElement(inode.lhs.value);
                if (!(e instanceof scope_1.ImportedPackage)) {
                    throw new errors_1.ImplementationError();
                }
                ip = e;
            }
            else if (inode.lhs.op == ".") {
                // Syntax of the kind: import . from "native/path" { func ... }
            }
            else {
                throw new errors_1.ImplementationError("import lhs " + inode.lhs.op);
            }
            for (let n of inode.rhs.parameters) {
                if (n.op == "funcType") {
                    // Do nothing by intention
                }
                else if (n.op == "typedef") {
                    this.typedefs.push(this.createTypedef(n, ip ? ip.pkg.scope : scope));
                }
                else if (n.op == "constValue") {
                    // Do nothing by intention
                }
                else {
                    throw new errors_1.ImplementationError("import " + n.op);
                }
            }
        }
        else {
            let importPath = inode.rhs.value;
            let p = pkg_1.Package.resolve(importPath, inode.rhs.loc);
            let ip;
            if (!inode.lhs) {
                // Syntax of the kind: import "path/to/module"
            }
            else if (inode.lhs.op == "identifierList") {
                // Syntax of the kind: import (id1, id2, ...) "path/to/module"
                for (let pnode of inode.lhs.parameters) {
                    if (p.scope.types.has(pnode.value)) {
                        var t = p.scope.types.get(pnode.value);
                        scope.registerType(pnode.value, t, pnode.loc);
                    }
                }
            }
            else if (inode.lhs.op == "id") {
                // Syntax of the kind: import identifier "path/to/module"
            }
            else if (inode.lhs.op == ".") {
                // Syntax of the kind: import . "path/to/module"
                for (var key of p.scope.types.keys()) {
                    var t = p.scope.types.get(key);
                    scope.registerType(key, t, inode.loc);
                }
            }
            else {
                throw new errors_1.ImplementationError("import lhs " + inode.lhs.op);
            }
        }
    }
    importFunctionsAndVariables(inode, scope) {
        if (inode.rhs.op == "importNative") {
            let ip;
            let importPath = inode.rhs.rhs.value;
            if (!inode.lhs) {
                // Syntax of the kind: import "native/path" { func ... }
                let importPathElements = importPath.split("/");
                let name = importPathElements[importPathElements.length - 1];
                // TODO: Sanitize the name
                let e = scope.resolveElement(name);
                if (!(e instanceof scope_1.ImportedPackage)) {
                    throw new errors_1.ImplementationError();
                }
                ip = e;
            }
            else if (inode.lhs.op == "id") {
                // Syntax of the kind: import identifier from "native/path" { func ... }
                // TODO: Sanitize the name
                let e = scope.resolveElement(inode.lhs.value);
                if (!(e instanceof scope_1.ImportedPackage)) {
                    throw new errors_1.ImplementationError();
                }
                ip = e;
            }
            else if (inode.lhs.op == ".") {
                // Syntax of the kind: import . from "native/path" { func ... }
            }
            else {
                throw new errors_1.ImplementationError("import lhs " + inode.lhs.op);
            }
            for (let n of inode.rhs.parameters) {
                if (n.op == "funcType") {
                    this.createNativeFunctionImport(inode.rhs.rhs.value, n, ip ? ip.pkg.scope : scope);
                }
                else if (n.op == "typedef") {
                    // Do nothing by intention
                }
                else if (n.op == "constValue") {
                    this.createNativeConstImport(inode.rhs.rhs.value, n, ip ? ip.pkg.scope : scope);
                }
                else {
                    throw new errors_1.ImplementationError("import " + n.op);
                }
            }
        }
        else {
            let importPath = inode.rhs.value;
            let p = pkg_1.Package.resolve(importPath, inode.rhs.loc);
            let ip;
            if (!inode.lhs) {
                // Syntax of the kind: import "path/to/module"
            }
            else if (inode.lhs.op == "identifierList") {
                // Syntax of the kind: import (id1, id2, ...) "path/to/module"
                for (let pnode of inode.lhs.parameters) {
                    if (p.scope.elements.has(pnode.value)) {
                        var el = p.scope.elements.get(pnode.value);
                        scope.registerElement(pnode.value, el, pnode.loc);
                        scope.setGroup(el, new group_1.Group(group_1.GroupKind.Free));
                    }
                    else if (p.scope.types.has(pnode.value)) {
                        // Do nothing by intention
                    }
                    else {
                        throw new errors_1.TypeError("Unknown identifier " + pnode.value + " in package \"" + p.pkgPath + "\"", pnode.loc);
                    }
                }
            }
            else if (inode.lhs.op == "id") {
                // Syntax of the kind: import identifier "path/to/module"
            }
            else if (inode.lhs.op == ".") {
                // Syntax of the kind: import . "path/to/module"
                for (var key of p.scope.elements.keys()) {
                    var el = p.scope.elements.get(key);
                    scope.registerElement(key, el, inode.loc);
                    scope.setGroup(el, new group_1.Group(group_1.GroupKind.Free));
                }
            }
            else {
                throw new errors_1.ImplementationError("import lhs " + inode.lhs.op);
            }
        }
    }
    processBuildInstructions(snode, pkg) {
        if (snode.parameters) {
            for (let p of snode.parameters) {
                if (p.op == "build_link") {
                    if (p.parameters) {
                        let args = [];
                        for (let a of p.parameters) {
                            args.push(a.value);
                        }
                        pkg.linkCmdLineArgs = pkg.linkCmdLineArgs ? pkg.linkCmdLineArgs.concat(args) : args;
                    }
                }
                else if (p.op == "build_compile") {
                    if (p.parameters) {
                        let args = [];
                        for (let a of p.parameters) {
                            args.push(a.value);
                        }
                        pkg.compileCmdLineArgs = pkg.compileCmdLineArgs ? pkg.compileCmdLineArgs.concat(args) : args;
                    }
                }
                else {
                    throw new errors_1.ImplementationError(snode.op);
                }
            }
        }
    }
    /**
     * The main function of the Typechecker that checks the types of an entire module.
     * However, this function just handles all imports and declares typedefs (but does not yet define them).
     *
     * Use checkModulePassTwo() and checkModulePassThree() to complete type checking.
     * The reason for splitting type checking in phases is that each phase is applied to all imported packages first,
     * before proceeding with the next phase.
     */
    checkModule(pkg) {
        let scope = new scope_1.Scope(null);
        scope.pkg = pkg;
        scope.registerType("bool", Static_1.Static.t_bool);
        scope.registerType("float", Static_1.Static.t_float);
        scope.registerType("double", Static_1.Static.t_double);
        scope.registerType("null", Static_1.Static.t_null);
        scope.registerType("byte", Static_1.Static.t_byte);
        scope.registerType("char", Static_1.Static.t_char);
        scope.registerType("int8", Static_1.Static.t_int8);
        scope.registerType("int16", Static_1.Static.t_int16);
        scope.registerType("int32", Static_1.Static.t_int32);
        scope.registerType("int64", Static_1.Static.t_int64);
        scope.registerType("int", Static_1.Static.t_int);
        scope.registerType("uint8", Static_1.Static.t_uint8);
        scope.registerType("uint16", Static_1.Static.t_uint16);
        scope.registerType("uint32", Static_1.Static.t_uint32);
        scope.registerType("uint64", Static_1.Static.t_uint64);
        scope.registerType("uint", Static_1.Static.t_uint);
        scope.registerType("string", Static_1.Static.t_string);
        scope.registerType("void", Static_1.Static.t_void);
        scope.registerType("error", Static_1.Static.t_error);
        scope.registerType("rune", Static_1.Static.t_rune);
        scope.registerType("coroutine", Static_1.Static.t_coroutine);
        pkg.pkgNode.scope = scope;
        this.moduleNode = pkg.pkgNode;
        // Iterate over all files and process all imports
        for (let fnode of this.moduleNode.statements) {
            fnode.scope = new scope_1.Scope(scope);
            for (let snode of fnode.statements) {
                if (snode.op == "import") {
                    this.createImport(snode, fnode.scope);
                }
                else if (snode.op == "build") {
                    this.processBuildInstructions(snode, pkg);
                }
            }
        }
        // Iterate over all files and declare all types.
        // The body of structs and interfaces is processed after all types are declared,
        // because types can reference themselves or each other cross-wise.
        for (let fnode of this.moduleNode.statements) {
            for (let snode of fnode.statements) {
                if (snode.op == "typedef") {
                    let t = this.createTypedef(snode, scope);
                    this.typedefs.push(t);
                }
            }
        }
        return scope;
    }
    /**
     * Imports types from other modules, defines typedefs, defines global variables
     * and declares functions.
     */
    checkModulePassTwo() {
        let scope = this.moduleNode.scope;
        // Iterate over all files and process all imports
        for (let fnode of this.moduleNode.statements) {
            for (let snode of fnode.statements) {
                if (snode.op == "import") {
                    this.importTypes(snode, fnode.scope);
                }
            }
        }
        // Define all types which have been declared before
        for (let t of this.typedefs) {
            if (t.type instanceof types_1.StructType) {
                this.createStructType(t.node.rhs, t.scope, t.type);
            }
            else if (t.type instanceof types_1.InterfaceType) {
                this.createInterfaceType(t.node.rhs, t.scope, t.type);
            }
            else if (t.type instanceof types_1.OrType) {
                this.createOrType(t.node.rhs, t.scope, t.type);
            }
            else if (t.type instanceof types_1.ArrayType) {
                this.createArrayType(t.node.rhs, t.scope, t.type);
            }
        }
        // Iterate over all files and declare all functions and global variables.
        for (let fnode of this.moduleNode.statements) {
            for (let snode of fnode.statements) {
                if (snode.op == "func" || snode.op == "export_func") {
                    let f = this.createFunction(snode, fnode.scope, scope);
                    if (f instanceof scope_1.Function) {
                        this.functions.push(f);
                    }
                }
                else if (snode.op == "var") {
                    let v = this.createVar(snode.lhs, scope, false, false, true);
                    v.node = snode;
                    this.globalVariables.push(v);
                }
                else if (snode.op == "let") {
                    let v = this.createVar(snode.lhs, scope, false, true, true);
                    v.node = snode;
                    this.globalVariables.push(v);
                }
                else if (snode.op == "import") {
                    // Do nothing by intention
                }
                else if (snode.op == "typedef") {
                    // Do nothing by intention
                }
                else if (snode.op == "comment") {
                    // Do nothing by intention
                }
                else if (snode.op == "build") {
                    // Do nothing by intention
                }
                else if (snode.op == "export_as") {
                    // Do nothing by intention
                }
                else {
                    throw new errors_1.ImplementationError(snode.op);
                }
            }
        }
    }
    /**
     * Imports functions from other modules, checks interfaces and structs for errors,
     * checks global variables.
     */
    checkModulePassThree() {
        let scope = this.moduleNode.scope;
        // Iterate over all files and process all imports for functions and variables
        for (let fnode of this.moduleNode.statements) {
            for (let snode of fnode.statements) {
                if (snode.op == "import") {
                    this.importFunctionsAndVariables(snode, fnode.scope);
                }
            }
        }
        // Check all interfaces for conflicting names
        for (let iface of this.ifaces) {
            this.checkInterfaceType(iface);
        }
        // Check all structs for conflicting names and whether they implement the promised interfaces
        for (let s of this.structs) {
            this.checkStructType(s);
        }
        // Check global variable assignments and determine the type of the global variables
        for (let v of this.globalVariables) {
            // Unique global pointers are subject to their own group.
            // All other global variables belong to the same group.
            if (helper.isUnique(v.type)) {
                scope.setGroup(v, new group_1.Group(group_1.GroupKind.Free, v.name));
            }
            else {
                scope.setGroup(v, this.globalGroup);
            }
            this.checkGlobalVariable(v, scope);
        }
        // Rename elements for export
        for (let fnode of this.moduleNode.statements) {
            for (let snode of fnode.statements) {
                if (snode.op == "export_as") {
                    for (let exp of snode.parameters) {
                        if (exp.op == "exportFuncAs" || exp.op == "exportConstAs" || exp.op == "exportVarAs") {
                            let e = fnode.scope.resolveElement(exp.lhs.value);
                            if (!e) {
                                throw new errors_1.TypeError("Unknown function/const/variable " + exp.lhs.value, exp.loc);
                            }
                            if (exp.lhs.value != exp.rhs.value) {
                                if (scope.resolveElement(exp.rhs.value)) {
                                    throw new errors_1.TypeError("A function/const/variable of name " + exp.rhs.value + " is already exported", exp.rhs.loc);
                                }
                                scope.registerElement(exp.rhs.value, e, exp.loc);
                            }
                        }
                        else if (exp.op == "exportTypeAs") {
                            let t = fnode.scope.resolveType(exp.lhs.value);
                            if (!t) {
                                throw new errors_1.TypeError("Unknown type " + exp.lhs.value, exp.loc);
                            }
                            if (exp.lhs.value != exp.rhs.value) {
                                if (scope.resolveType(exp.rhs.value)) {
                                    throw new errors_1.TypeError("A type of name " + exp.rhs.value + " is already exported", exp.rhs.loc);
                                }
                                scope.registerType(exp.rhs.value, t, exp.loc);
                            }
                        }
                        else {
                            throw new errors_1.ImplementationError(exp.op);
                        }
                    }
                }
            }
        }
    }
    /**
     * Checks all function bodies
     */
    checkModulePassFour() {
        // Check function bodies
        for (let e of this.functions) {
            this.checkFunctionBody(e);
        }
        // Determine which functions could block and hence needs special coroutine treatment.
        /*
        let changes = false;
        do {
            for(let f of this.callGraph.keys()) {
                if (f.type.callingConvention == "fyrCoroutine") {
                    continue;
                }
                if (!this.callGraph.has(f)) {
                    continue;
                }
                let arr = this.callGraph.get(f);
                for(let a of arr) {
                    if (a.callingConvention == "fyrCoroutine") {
                        f.type.callingConvention = "fyrCoroutine";
                        changes = true;
                        break;
                    }
                }
            }
        } while(changes);
        */
    }
    /**
     * Typecheck the template function body.
     */
    checkFunctionBody(f) {
        let scopeExit;
        if (f.node.statements) {
            scopeExit = this.checkStatements(f.node.statements, f.scope);
        }
        else {
            scopeExit = new scope_1.ScopeExit();
            scopeExit.fallthrough = f.scope;
        }
        let needsReturn = !!f.node.rhs;
        if (needsReturn) {
            if (scopeExit.fallthrough) {
                throw new errors_1.TypeError("Missing return at end of function", f.loc);
            }
        }
        this.checkGroupsInFunction(f);
    }
    checkVarAssignment(isConst, scope, vnode, rtype, rnode = null) {
        // TODO: const and box are not handled properly here
        if (vnode.op == "id" || vnode.op == "optionalId") {
            let v = this.createVar(vnode, scope, false, isConst);
            if (!v.type) {
                if (rtype instanceof types_1.PackageType) {
                    throw new errors_1.TypeError("Package types cannot be assigned", vnode.loc);
                }
                if (rtype instanceof types_1.ArrayLiteralType || rtype instanceof types_1.TupleLiteralType || rtype instanceof types_1.ObjectLiteralType) {
                    v.type = this.defaultLiteralType(rnode);
                }
                else {
                    v.type = rtype;
                }
            }
            else {
                if (rnode) {
                    this.checkIsAssignableNode(v.type, rnode, scope);
                }
                else {
                    this.checkIsAssignableType(v.type, rtype, vnode.loc, "assign", true);
                }
            }
            if (isConst && rnode && (rnode.op == "array" || rnode.op == "object")) {
                v.isNotNull = true;
            }
            vnode.type = v.type;
        }
        else if (vnode.op == "tuple") {
            let rtypeStripped = types_1.RestrictedType.strip(rtype);
            if (!(rtypeStripped instanceof types_1.TupleType) && !(rtypeStripped instanceof types_1.TupleLiteralType)) {
                throw new errors_1.TypeError("Expected tuple expression or json on right hand side", vnode.loc);
            }
            let hasEllipsis = false;
            for (let i = 0; i < vnode.parameters.length; i++) {
                if (i >= rtypeStripped.types.length) {
                    throw new errors_1.TypeError("Mismatch in tuple type length", vnode.loc);
                }
                let p = vnode.parameters[i];
                if (p.op == "optionalId") {
                    throw new errors_1.TypeError("Optional identifiers are not allowed in tuple context", vnode.loc);
                }
                else if (p.op == "ellipsisId") {
                    if (i + 1 != vnode.parameters.length) {
                        throw new errors_1.TypeError("Ellipsis identifier must be at last position in tuple", vnode.loc);
                    }
                    hasEllipsis = true;
                    let v = this.createVar(p, scope, false, isConst);
                    if (!v.type) {
                        v.type = new types_1.TupleType(rtypeStripped.types.slice(i));
                        if (isConst) {
                            v.type = new types_1.RestrictedType(v.type, { isConst: true });
                        }
                    }
                    else {
                        let lt = types_1.RestrictedType.strip(v.type);
                        if (!(lt instanceof types_1.TupleType)) {
                            throw new errors_1.TypeError("Ellipsis identifier in a tuple context must be of tuple type", vnode.loc);
                        }
                        if (lt.types.length != rtypeStripped.types.length - i) {
                            throw new errors_1.TypeError("Mismatch in tuple type length", vnode.loc);
                        }
                        if (rnode && rnode.op == "tuple") {
                            for (let j = i; j < rnode.parameters.length; j++) {
                                let r = rnode.parameters[j];
                                this.checkIsAssignableNode(lt.types[j - i], r, scope);
                            }
                        }
                        else {
                            let t = new types_1.TupleType(rtypeStripped.types.slice(i));
                            this.checkIsAssignableType(v.type, t, vnode.loc, "assign", true);
                        }
                    }
                    break;
                }
                else {
                    let r;
                    if (rnode && rnode.op == "tuple") {
                        r = rnode.parameters[i];
                    }
                    this.checkVarAssignment(isConst, scope, p, rtypeStripped.types[i], r);
                }
            }
            if (!hasEllipsis && rtypeStripped.types.length != vnode.parameters.length) {
                throw new errors_1.TypeError("Mismatch in tuple type length", vnode.loc);
            }
        }
        else if (vnode.op == "array") {
            let rtypeStripped = types_1.RestrictedType.strip(rtype);
            if (!(rtypeStripped instanceof types_1.ArrayLiteralType) && !(rtypeStripped instanceof types_1.ArrayType) && !(rtypeStripped instanceof types_1.SliceType)) {
                throw new errors_1.TypeError("Expected an expression of array type or slice type", vnode.loc);
            }
            let hasEllipsis = false;
            let hasOptional = false;
            for (let i = 0; i < vnode.parameters.length; i++) {
                if (rtypeStripped instanceof types_1.ArrayType && i >= rtypeStripped.size) {
                    throw new errors_1.TypeError("Mismatch in array type length", vnode.loc);
                }
                let p = vnode.parameters[i];
                if (p.op == "ellipsisId") {
                    if (i + 1 != vnode.parameters.length) {
                        throw new errors_1.TypeError("Ellipsis identifier must be at last position in array", vnode.loc);
                    }
                    hasEllipsis = true;
                    let v = this.createVar(p, scope, false, isConst);
                    if (!v.type) {
                        if (rtypeStripped instanceof types_1.ArrayLiteralType) {
                            for (let j = i; j < rnode.parameters.length; j++) {
                                // TODO: Check that all elements of the array have the same type
                                //                                this.checkIsAssignableNode(Static.t_json, rnode.parameters[j]);
                                //                                rtype.types[j] = rnode.parameters[j].type;
                            }
                            //                            v.type = new SliceType(Static.t_json);
                            throw new errors_1.TodoError();
                        }
                        else if (rtypeStripped instanceof types_1.ArrayType) {
                            v.type = new types_1.ArrayType(rtypeStripped.elementType, rtypeStripped.size - i);
                            // TODO: Check whether the array slice can be copied at all
                            // TODO: Clone the restrictions of the array
                            throw new errors_1.TodoError();
                        }
                        else if (rtypeStripped instanceof types_1.SliceType) {
                            // TODO: Clone the restrictions of the array
                            v.type = rtype;
                            throw new errors_1.TodoError();
                        }
                    }
                    else {
                        let lt = types_1.RestrictedType.strip(v.type);
                        if (rtypeStripped instanceof types_1.ArrayLiteralType) {
                            let rt;
                            if (lt instanceof types_1.ArrayType) {
                                if (rtypeStripped.types.length - i != lt.size) {
                                    throw new errors_1.TypeError("Mismatch in array type length", vnode.loc);
                                }
                                rt = lt.elementType;
                            }
                            else if (lt instanceof types_1.SliceType) {
                                rt = lt.getElementType();
                            }
                            else {
                                throw new errors_1.TypeError("Ellipsis identifier must be of array type, slice type, string or json", vnode.loc);
                            }
                            for (let j = i; j < rnode.parameters.length; j++) {
                                this.checkIsAssignableNode(rt, rnode.parameters[j], scope);
                            }
                        }
                        else if (rtypeStripped instanceof types_1.ArrayType) {
                            if (!(lt instanceof types_1.ArrayType)) {
                                throw new errors_1.TypeError("Ellipsis identifier must be of array type", vnode.loc);
                            }
                            if (lt.size != rtypeStripped.size - i) {
                                throw new errors_1.TypeError("Mismatch in array size", vnode.loc);
                            }
                            this.checkIsAssignableType(lt.elementType, rtypeStripped.elementType, vnode.loc, "assign", true);
                        }
                        else if (rtypeStripped instanceof types_1.SliceType) {
                            if (!(lt instanceof types_1.SliceType)) {
                                throw new errors_1.TypeError("Ellipsis identifier must be of slice type", vnode.loc);
                            }
                            this.checkIsAssignableType(lt.getElementType(), rtypeStripped.getElementType(), vnode.loc, "assign", true);
                        }
                    }
                    break;
                }
                else {
                    if (p.op == "optionalId") {
                        if (rtype instanceof types_1.ArrayType || rtype instanceof types_1.ArrayLiteralType) {
                            throw new errors_1.TypeError("Optional identifiers are not allowed in array context", vnode.loc);
                        }
                        hasOptional = true;
                    }
                    else {
                        if (hasOptional) {
                            throw new errors_1.TypeError("Non-optional identifier is not allowed after optional identifiers", vnode.loc);
                        }
                    }
                    let r;
                    let rt;
                    if (rtype instanceof types_1.ArrayLiteralType) {
                        rt = rtype.types[i];
                        r = rnode.parameters[i];
                    }
                    else if (rtype instanceof types_1.ArrayType || rtype instanceof types_1.SliceType) {
                        rt = rtype.getElementType();
                    }
                    this.checkVarAssignment(isConst, scope, p, rt, r);
                }
            }
            if (!hasEllipsis && rtype instanceof types_1.ArrayType && rtype.size != vnode.parameters.length) {
                throw new errors_1.TypeError("Mismatch in tuple type length", vnode.loc);
            }
        }
        else if (vnode.op == "object") {
            let rtypeStripped = types_1.RestrictedType.strip(rtype);
            if (!(rtypeStripped instanceof types_1.ObjectLiteralType) && (!helper.isMap(rtypeStripped) || helper.isString(this.mapKeyType(rtypeStripped)))) {
                throw new errors_1.TypeError("Expected an expression of type object literal or map[string]...", vnode.loc);
            }
            let hasEllipsis = false;
            for (let i = 0; i < vnode.parameters.length; i++) {
                let kv = vnode.parameters[i];
                if (kv.op == "ellipsisId") {
                    if (i + 1 != vnode.parameters.length) {
                        throw new errors_1.TypeError("Ellipsis identifier must be at last position in object", vnode.loc);
                    }
                    hasEllipsis = true;
                    let v = this.createVar(kv, scope, false, isConst);
                    if (!v.type) {
                        if (rtypeStripped instanceof types_1.ObjectLiteralType) {
                            let valueType = null; // TODO
                            for (let j = i; j < rnode.parameters.length; j++) {
                                //                                this.checkIsAssignableNode(Static.t_json, rnode.parameters[j].lhs);
                            }
                            v.type = new types_1.PointerType(new types_1.MapType(Static_1.Static.t_string, valueType), "strong");
                            throw new errors_1.TodoError();
                        }
                        else if (rtypeStripped instanceof types_1.TemplateStructType) {
                            v.type = rtype;
                        }
                    }
                    else {
                        let lt = types_1.RestrictedType.strip(v.type);
                        if (rtypeStripped instanceof types_1.ObjectLiteralType) {
                            let rt;
                            if (helper.isMap(lt) && helper.isString(this.mapKeyType(lt))) {
                                rt = this.mapValueType(lt);
                            }
                            else {
                                throw new errors_1.TypeError("Ellipsis identifier must be of map type", vnode.loc);
                            }
                            for (let j = i; j < rnode.parameters.length; j++) {
                                this.checkIsAssignableNode(rt, rnode.parameters[j].lhs, scope);
                            }
                        }
                        else if (helper.isMap(rtypeStripped)) {
                            if (!helper.isMap(lt) || !helper.isString(this.mapKeyType(lt))) {
                                throw new errors_1.TypeError("Ellipsis identifier must be of type map<string, ...>", vnode.loc);
                            }
                            this.checkIsAssignableType(this.mapKeyType(lt), this.mapValueType(rtypeStripped), vnode.loc, "assign", true);
                        }
                    }
                }
                else {
                    let p = kv.lhs;
                    let name = kv.name.value;
                    let optional = (kv.op == "optionalKeyValue");
                    let r;
                    let rt;
                    if (rtype instanceof types_1.ObjectLiteralType) {
                        if (!optional && !rtype.types.has(name)) {
                            throw new errors_1.TypeError("Object literal has no key '" + name + "'", p.loc);
                        }
                        rt = rtype.types.get(name);
                        r = rnode.parameters[i].lhs;
                        throw new errors_1.TodoError("Find matching node in literal");
                    }
                    else if (helper.isMap(rtype)) {
                        rt = this.mapValueType(rtype);
                    }
                    this.checkVarAssignment(isConst, scope, p, rt, r);
                }
            }
        }
    }
    checkAssignment(scope, vnode, rtype, rnode = null) {
        let rtypeStripped = types_1.RestrictedType.strip(rtype);
        if (vnode.op == "tuple") {
            if (!(rtypeStripped instanceof types_1.TupleType) && !(rtypeStripped instanceof types_1.TupleLiteralType)) {
                throw new errors_1.TypeError("Expected tuple expression on right hand side", vnode.loc);
            }
            let hasEllipsis = false;
            for (let i = 0; i < vnode.parameters.length; i++) {
                if (i >= rtypeStripped.types.length) {
                    throw new errors_1.TypeError("Mismatch in tuple type length", vnode.loc);
                }
                let p = vnode.parameters[i];
                if (p.op == "optionalAssign") {
                    throw new errors_1.TypeError("Optional identifiers are not allowed in tuple context", vnode.loc);
                }
                else if (p.op == "ellipsisAssign") {
                    if (i + 1 != vnode.parameters.length) {
                        throw new errors_1.TypeError("Ellipsis identifier must be at last position in tuple", vnode.loc);
                    }
                    this.checkExpression(p.lhs, scope);
                    this.checkIsAssignable(p.lhs, scope);
                    hasEllipsis = true;
                    if (!(p.lhs.type instanceof types_1.TupleType)) {
                        throw new errors_1.TypeError("Ellipsis identifier in a tuple context must be of tuple type", vnode.loc);
                    }
                    if (p.lhs.type.types.length != rtypeStripped.types.length - i) {
                        throw new errors_1.TypeError("Mismatch in tuple type length", vnode.loc);
                    }
                    if (rnode && rnode.op == "tuple") {
                        for (let j = i; j < rnode.parameters.length; j++) {
                            let r = rnode.parameters[j];
                            this.checkIsAssignableNode(p.lhs.type.types[j - i], r, scope);
                        }
                    }
                    else {
                        let t = new types_1.TupleType(rtypeStripped.types.slice(i));
                        this.checkIsAssignableType(p.lhs.type, t, vnode.loc, "assign", true);
                    }
                    break;
                }
                else {
                    let r;
                    if (rnode && rnode.op == "tuple") {
                        r = rnode.parameters[i];
                    }
                    this.checkAssignment(scope, p, rtypeStripped.types[i], r);
                }
            }
            if (!hasEllipsis && rtypeStripped.types.length != vnode.parameters.length) {
                throw new errors_1.TypeError("Mismatch in tuple type length", vnode.loc);
            }
            // The type of the right-hand side might have been inferred. In this case, compute the new type
            if (rnode && rnode.op == "tuple") {
                let types = [];
                for (let p of rnode.parameters) {
                    types.push(p.type);
                }
                rnode.type = new types_1.TupleType(types);
            }
        }
        else if (vnode.op == "array") {
            if (!(rtypeStripped instanceof types_1.ArrayLiteralType) && !(rtypeStripped instanceof types_1.ArrayType) && !(rtypeStripped instanceof types_1.SliceType)) {
                throw new errors_1.TypeError("Expected an expression of array type or slice type", vnode.loc);
            }
            let hasEllipsis = false;
            let hasOptional = false;
            for (let i = 0; i < vnode.parameters.length; i++) {
                if (rtypeStripped instanceof types_1.ArrayType && i >= rtypeStripped.size) {
                    throw new errors_1.TypeError("Mismatch in array type length", vnode.loc);
                }
                let p = vnode.parameters[i];
                if (p.op == "ellipsisAssign") {
                    if (i + 1 != vnode.parameters.length) {
                        throw new errors_1.TypeError("Ellipsis identifier must be at last position in array", vnode.loc);
                    }
                    hasEllipsis = true;
                    this.checkExpression(p.lhs, scope);
                    this.checkIsAssignable(p.lhs, scope);
                    if (rtypeStripped instanceof types_1.ArrayLiteralType) {
                        let rt;
                        if (p.lhs.type instanceof types_1.ArrayType) {
                            if (rtypeStripped.types.length - i != p.lhs.type.size) {
                                throw new errors_1.TypeError("Mismatch in array type length", vnode.loc);
                            }
                            rt = p.lhs.type.elementType;
                        }
                        else if (p.lhs.type instanceof types_1.SliceType) {
                            rt = p.lhs.type.getElementType();
                        }
                        else {
                            throw new errors_1.TypeError("Ellipsis identifier must be of array type or slice type", vnode.loc);
                        }
                        for (let j = i; j < rnode.parameters.length; j++) {
                            this.checkIsAssignableNode(rt, rnode.parameters[j], scope);
                        }
                    }
                    else if (rtypeStripped instanceof types_1.ArrayType) {
                        if (!(p.lhs.type instanceof types_1.ArrayType)) {
                            throw new errors_1.TypeError("Ellipsis identifier must be of array type", vnode.loc);
                        }
                        if (p.lhs.type.size != rtypeStripped.size - i) {
                            throw new errors_1.TypeError("Mismatch in array size", vnode.loc);
                        }
                        this.checkIsAssignableType(p.lhs.type.elementType, rtypeStripped.elementType, vnode.loc, "assign", true);
                    }
                    else if (rtype instanceof types_1.SliceType) {
                        if (!(p.lhs.type instanceof types_1.SliceType)) {
                            throw new errors_1.TypeError("Ellipsis identifier must be of slice type", vnode.loc);
                        }
                        this.checkIsAssignableType(p.lhs.type.getElementType(), rtypeStripped.getElementType(), vnode.loc, "assign", true);
                    }
                    break;
                }
                else {
                    if (p.op == "optionalAssign") {
                        if (rtypeStripped instanceof types_1.ArrayType || rtypeStripped instanceof types_1.ArrayLiteralType) {
                            throw new errors_1.TypeError("Optional identifiers are not allowed in array context", vnode.loc);
                        }
                        hasOptional = true;
                    }
                    else {
                        if (hasOptional) {
                            throw new errors_1.TypeError("Non-optional identifier is not allowed after optional identifiers", vnode.loc);
                        }
                    }
                    let r;
                    let rt;
                    if (rtype instanceof types_1.ArrayLiteralType) {
                        rt = rtype.types[i];
                        r = rnode.parameters[i];
                    }
                    else if (rtype instanceof types_1.ArrayType || rtype instanceof types_1.SliceType) {
                        rt = rtype.getElementType();
                    }
                    this.checkAssignment(scope, p, rt, r);
                }
            }
            if (!hasEllipsis && rtype instanceof types_1.ArrayType && rtype.size != vnode.parameters.length) {
                throw new errors_1.TypeError("Mismatch in tuple type length", vnode.loc);
            }
        }
        else if (vnode.op == "object") {
            if (!(rtypeStripped instanceof types_1.ObjectLiteralType) && (!helper.isMap(rtype) || !helper.isString(this.mapKeyType(rtype)))) {
                throw new errors_1.TypeError("Expected an expression of type object literal or map[string]...", vnode.loc);
            }
            for (let i = 0; i < vnode.parameters.length; i++) {
                let kv = vnode.parameters[i];
                if (kv.op == "ellipsisAssign") {
                    if (i + 1 != vnode.parameters.length) {
                        throw new errors_1.TypeError("Ellipsis identifier must be at last position in object", vnode.loc);
                    }
                    this.checkExpression(kv.lhs, scope);
                    this.checkIsAssignable(kv.lhs, scope);
                    if (rtype instanceof types_1.ObjectLiteralType) {
                        let rt;
                        if (helper.isMap(kv.lhs.type) && helper.isString(this.mapKeyType(kv.lhs.type))) {
                            rt = this.mapValueType(kv.lhs.type);
                        }
                        else {
                            throw new errors_1.TypeError("Ellipsis identifier must be of map type or json", vnode.loc);
                        }
                        for (let j = i; j < rnode.parameters.length; j++) {
                            this.checkIsAssignableNode(rt, rnode.parameters[j].lhs, scope);
                        }
                    }
                    else if (helper.isMap(rtypeStripped)) {
                        if (!helper.isMap(kv.lhs.type) || !helper.isString(this.mapKeyType(kv.lhs.type))) {
                            throw new errors_1.TypeError("Ellipsis identifier must be of type map[string]...", vnode.loc);
                        }
                        this.checkIsAssignableType(this.mapKeyType(kv.lhs.type), this.mapValueType(rtypeStripped), vnode.loc, "assign", true);
                    }
                }
                else {
                    let p = kv.lhs;
                    let name = kv.name.value;
                    let optional = (p.op == "optionalKeyValue");
                    let r;
                    let rt;
                    if (rtypeStripped instanceof types_1.ObjectLiteralType) {
                        if (!optional && !rtypeStripped.types.has(name)) {
                            throw new errors_1.TypeError("Object literal has no key '" + name + "'", p.loc);
                        }
                        rt = rtypeStripped.types.get(name);
                        r = rnode.parameters[i].lhs;
                        throw new errors_1.TodoError("Find matching node in literal");
                    }
                    else if (helper.isMap(rtypeStripped)) {
                        rt = this.mapValueType(rtypeStripped);
                    }
                    this.checkAssignment(scope, p, rt, r);
                }
            }
        }
        else {
            this.checkExpression(vnode, scope);
            this.checkIsAssignable(vnode, scope);
            if (rnode) {
                this.checkIsAssignableNode(vnode.type, rnode, scope);
            }
            else {
                this.checkIsAssignableType(vnode.type, rtype, vnode.loc, "assign", true);
            }
        }
    }
    checkStatements(statements, scope) {
        let scopeExit = new scope_1.ScopeExit();
        scopeExit.fallthrough = scope;
        for (let i = 0; i < statements.length; i++) {
            let st = statements[i];
            this.checkStatement(st, scope, scopeExit);
        }
        return scopeExit;
    }
    checkStatement(snode, scope, scopeExit) {
        if (snode.op == "comment") {
            return;
        }
        if (!scopeExit.fallthrough) {
            throw new errors_1.TypeError("Unreachable code", snode.loc);
        }
        switch (snode.op) {
            case "return":
                let f = scope.envelopingFunction();
                if (!f) {
                    throw new errors_1.TypeError("'return' outside of function body", snode.loc);
                }
                if (!snode.lhs) {
                    if (f.type.returnType != Static_1.Static.t_void && !f.hasNamedReturnVariables) {
                        throw new errors_1.TypeError("Mismatch in return type", snode.loc);
                    }
                }
                else {
                    this.checkExpression(snode.lhs, scope);
                    this.checkIsAssignableNode(f.type.returnType, snode.lhs, scope);
                }
                if (!scopeExit.returns)
                    scopeExit.returns = [];
                scopeExit.returns.push(scope);
                scopeExit.fallthrough = null;
                return;
            case "break":
                if (!scope.isInForLoop()) {
                    throw new errors_1.TypeError("'break' outside of loop", snode.loc);
                }
                if (!scopeExit.breaks)
                    scopeExit.breaks = [];
                scopeExit.breaks.push(scope);
                scopeExit.fallthrough = null;
                return;
            case "continue":
                if (!scope.isInForLoop()) {
                    throw new errors_1.TypeError("'continue' outside of loop", snode.loc);
                }
                if (!scopeExit.continues)
                    scopeExit.continues = [];
                scopeExit.continues.push(scope);
                scopeExit.fallthrough = null;
                return;
            case "if":
                {
                    let s;
                    // Assignment in the if-clause?
                    if (snode.lhs) {
                        snode.lhs.scope = new scope_1.Scope(scope);
                        let initScopeExit = new scope_1.ScopeExit();
                        initScopeExit.fallthrough = scope;
                        this.checkStatement(snode.lhs, snode.lhs.scope, initScopeExit);
                        if (initScopeExit.returns || initScopeExit.breaks || initScopeExit.continues) {
                            throw new errors_1.TypeError("break, return and continue are not allowed inside the initialization statement of an if clause.", snode.loc);
                        }
                        // initScopeExit can be ignored furtheron, since it does neither return nor break or continue.
                        s = new scope_1.Scope(snode.lhs.scope);
                    }
                    else {
                        s = new scope_1.Scope(scope);
                    }
                    snode.scope = s;
                    // Check the if-clause
                    this.checkExpression(snode.condition, s);
                    this.checkIsAssignableType(Static_1.Static.t_bool, snode.condition.type, snode.condition.loc, "assign", true);
                    if (snode.condition.op == "bool" && snode.condition.value == "false") {
                        // Do not type check the if-clause, because it does not execute
                        snode.statements = [];
                    }
                    snode.scopeExit = this.checkStatements(snode.statements, s);
                    scopeExit.merge(snode.scopeExit);
                    // Check the else clause
                    if (snode.elseBranch) {
                        this.checkStatement(snode.elseBranch, scope, scopeExit);
                        scopeExit.merge(snode.elseBranch.scopeExit);
                        if (!snode.scopeExit.fallthrough && !snode.elseBranch.scopeExit.fallthrough && snode.elseBranch.op == "else") {
                            scopeExit.fallthrough = null;
                        }
                    }
                    return;
                }
            case "else":
                {
                    let s = new scope_1.Scope(scope);
                    snode.scope = s;
                    snode.scopeExit = this.checkStatements(snode.statements, s);
                    scopeExit.merge(snode.scopeExit);
                    return;
                }
            case "for":
                {
                    let forScope;
                    if (snode.condition) {
                        snode.condition.scope = new scope_1.Scope(scope);
                        if (snode.condition.op == ";;") {
                            if (snode.condition.lhs) {
                                let initScopeExit = new scope_1.ScopeExit();
                                initScopeExit.fallthrough = scope;
                                this.checkStatement(snode.condition.lhs, snode.condition.scope, initScopeExit);
                                if (initScopeExit.returns || initScopeExit.breaks || initScopeExit.continues) {
                                    throw new errors_1.TypeError("break, return and continue are not allowed inside the initialization statement of a for loop.", snode.loc);
                                }
                            }
                            if (snode.condition.condition) {
                                this.checkExpression(snode.condition.condition, snode.condition.scope);
                                this.checkIsAssignableType(Static_1.Static.t_bool, snode.condition.condition.type, snode.condition.condition.loc, "assign", true);
                            }
                            if (snode.condition.rhs) {
                                let loopScopeExit = new scope_1.ScopeExit();
                                loopScopeExit.fallthrough = scope;
                                this.checkStatement(snode.condition.rhs, snode.condition.scope, loopScopeExit);
                                if (loopScopeExit.returns || loopScopeExit.breaks || loopScopeExit.continues) {
                                    throw new errors_1.TypeError("break, return and continue are not allowed inside the loop statement of a for loop.", snode.loc);
                                }
                            }
                        }
                        else {
                            this.checkStatement(snode.condition, snode.condition.scope, scopeExit);
                        }
                        forScope = new scope_1.Scope(snode.condition.scope);
                    }
                    else {
                        forScope = new scope_1.Scope(scope);
                    }
                    forScope.forLoop = true;
                    snode.scope = forScope;
                    snode.scopeExit = this.checkStatements(snode.statements, forScope);
                    scopeExit.merge(snode.scopeExit, true);
                    return;
                }
            case "var":
            case "let":
                if (!snode.rhs) {
                    if (snode.op == "let") {
                        throw new errors_1.ImplementationError("let without initialization");
                    }
                    if (snode.lhs.op == "id") {
                        let v = this.createVar(snode.lhs, scope, true);
                        snode.lhs.type = v.type;
                    }
                    else if (snode.lhs.op == "tuple") {
                        for (let p of snode.lhs.parameters) {
                            let v = this.createVar(p, scope, true);
                            p.type = v.type;
                        }
                    }
                    else {
                        throw new errors_1.TodoError("Implementation error");
                    }
                }
                else {
                    this.checkExpression(snode.rhs, scope);
                    this.checkVarAssignment(snode.op == "let", scope, snode.lhs, snode.rhs.type, snode.rhs);
                    if (TypeChecker.hasLocalReference(snode.rhs.type)) {
                        throw new errors_1.TypeError("Right hand side of assignment must not be a local reference", snode.rhs.loc);
                    }
                }
                break;
            case "=":
                this.checkExpression(snode.rhs, scope);
                this.checkAssignment(scope, snode.lhs, snode.rhs.type, snode.rhs);
                if (TypeChecker.hasLocalReference(snode.rhs.type)) {
                    throw new errors_1.TypeError("Right hand side of assignment must not be a local reference", snode.rhs.loc);
                }
                break;
            case "+=":
            case "*=":
            case "/=":
            case "-=":
                this.checkExpression(snode.lhs, scope);
                this.checkIsAssignable(snode.lhs, scope);
                this.checkExpression(snode.rhs, scope);
                if (snode.op == "+=" && helper.isString(snode.lhs.type)) {
                    this.checkIsString(snode.rhs);
                }
                else if (helper.isUnsafePointer(snode.lhs.type)) {
                    if (snode.op == "*=" || snode.op == "/=") {
                        throw new errors_1.TypeError("'" + snode.op + "' is an invalid operation on pointers", snode.loc);
                    }
                    this.checkIsIntNumber(snode.rhs);
                }
                else {
                    this.checkIsNumber(snode.lhs);
                    this.checkIsNumber(snode.rhs);
                    if (snode.rhs.op == "int" || snode.rhs.op == "float") {
                        this.unifyLiterals(snode.lhs.type, snode.rhs, scope, snode.loc);
                    }
                    else {
                        this.checkIsAssignableType(snode.lhs.type, snode.rhs.type, snode.loc, "assign", true);
                    }
                }
                break;
            case "<<=":
            case ">>=":
                {
                    this.checkExpression(snode.lhs, scope);
                    this.checkIsAssignable(snode.lhs, scope);
                    this.checkExpression(snode.rhs, scope);
                    if (helper.isUnsafePointer(snode.lhs.type)) {
                        if (snode.rhs.op == "int") {
                            this.unifyLiterals(Static_1.Static.t_uint, snode.rhs, scope, snode.loc);
                        }
                        else {
                            this.checkIsAssignableType(Static_1.Static.t_uint, snode.rhs.type, snode.loc, "assign", true);
                        }
                    }
                    else {
                        this.checkIsIntNumber(snode.lhs);
                        if (snode.rhs.op == "int") {
                            this.unifyLiterals(Static_1.Static.t_uint, snode.rhs, scope, snode.loc);
                        }
                        else {
                            this.checkIsAssignableType(Static_1.Static.t_uint, snode.rhs.type, snode.loc, "assign", true);
                        }
                    }
                    break;
                }
            case "%=":
            case "&=":
            case "&^=":
            case "|=":
            case "^=":
                this.checkExpression(snode.lhs, scope);
                this.checkIsAssignable(snode.lhs, scope);
                this.checkExpression(snode.rhs, scope);
                if (helper.isUnsafePointer(snode.lhs.type)) {
                    if (snode.op == "%=") {
                        throw new errors_1.TypeError("'%=' is an invalid operation on pointers", snode.loc);
                    }
                    if (snode.rhs.op == "int") {
                        this.unifyLiterals(snode.lhs.type, snode.rhs, scope, snode.loc);
                    }
                    else {
                        this.checkIsAssignableType(Static_1.Static.t_int, snode.rhs.type, snode.loc, "assign", true);
                    }
                }
                else {
                    this.checkIsIntNumber(snode.lhs);
                    this.checkIsIntNumber(snode.rhs);
                    if (snode.rhs.op == "int" || snode.rhs.op == "float") {
                        this.unifyLiterals(snode.lhs.type, snode.rhs, scope, snode.loc);
                    }
                    else {
                        this.checkIsAssignableType(snode.lhs.type, snode.rhs.type, snode.loc, "assign", true);
                    }
                }
                break;
            /*
        case "in":
            this.checkExpression(snode.rhs, scope);
            let [tindex1, tindex2] = this.checkIsEnumerable(snode.rhs);
            if (snode.lhs.op == "tuple") {
//                    if (snode.lhs.parameters[0].op != "id" || snode.lhs.parameters[0].value != "_") {
                if (snode.lhs.parameters[0].value != "_") {
                    this.checkExpression(snode.lhs.parameters[0], scope);
                    this.checkIsMutable(snode.lhs.parameters[0], scope);
                    this.checkIsAssignableType(snode.lhs.parameters[0].type, tindex1, snode.loc, "assign", true);
                }
                if (snode.lhs.parameters[1].value != "_") {
                    this.checkExpression(snode.lhs.parameters[1], scope);
                    this.checkIsMutable(snode.lhs.parameters[1], scope);
                    this.checkIsAssignableType(snode.lhs.parameters[1].type, tindex2, snode.loc, "assign", true);
                }
            } else {
                if (snode.lhs.value != "_") {
                    this.checkExpression(snode.lhs, scope);
                    this.checkIsMutable(snode.lhs, scope);
                    this.checkIsAssignableType(snode.lhs.type, tindex1, snode.loc, "assign", true);
                }
            }
            break; */
            case "let_in":
                {
                    this.checkExpression(snode.rhs, scope);
                    let [tindex1, tindex2] = this.checkIsEnumerable(snode.rhs);
                    if (snode.lhs.op == "tuple") {
                        if (snode.lhs.parameters[0].value != "_") {
                            let v1 = this.createVar(snode.lhs.parameters[0], scope, false, true);
                            if (v1.type) {
                                this.checkIsAssignableType(v1.type, tindex1, snode.loc, "assign", true);
                            }
                            else {
                                v1.type = tindex1;
                            }
                        }
                        if (snode.lhs.parameters[1].value != "_") {
                            let v2 = this.createVar(snode.lhs.parameters[1], scope, false, true);
                            if (v2.type) {
                                this.checkIsAssignableType(v2.type, tindex2, snode.loc, "assign", true);
                            }
                            else {
                                v2.type = tindex2;
                            }
                            //                        if (v2.type != Static.t_string) {
                            v2.isForLoopPointer = true;
                            //                        }
                        }
                    }
                    else {
                        let v = this.createVar(snode.lhs, scope, false, true);
                        if (v.type) {
                            this.checkIsAssignableType(v.type, tindex1, snode.loc, "assign", true);
                        }
                        else {
                            v.type = tindex1;
                        }
                        //                    if (v.type != Static.t_string) {
                        v.isForLoopPointer = true;
                        //                    }
                    }
                    break;
                }
            case "yield":
                break;
            case "yield_continue":
                break;
            case "spawn":
                {
                    this.checkExpression(snode.rhs, scope);
                    if (snode.rhs.op != "(") {
                        throw new errors_1.ImplementationError();
                    }
                    if (!(snode.rhs.lhs.type instanceof types_1.FunctionType)) {
                        throw new errors_1.ImplementationError();
                    }
                    if (snode.rhs.lhs.type.returnType != Static_1.Static.t_void) {
                        throw new errors_1.TypeError("Functions invoked via 'spawn' must return void", snode.loc);
                    }
                    // Functions invoked via `spawn` must not accept local references.
                    for (let p of snode.rhs.lhs.type.parameters) {
                        if (helper.isLocalReference(p.type)) {
                            throw new errors_1.TypeError("Functions invoked via 'spawn' must not use local references in parameter types", p.loc);
                        }
                    }
                    if (snode.rhs.lhs.op == "." && !helper.isSafePointer(snode.rhs.lhs.lhs.type)) {
                        throw new errors_1.TypeError("Function calls via 'spawn' must use a pointer to the object", snode.loc);
                    }
                    if (snode.rhs.lhs.op == "." && helper.isLocalReference(snode.rhs.lhs.lhs.type)) {
                        throw new errors_1.TypeError("Function calls via 'spawn' must use not use a local-pointer to the object ", snode.loc);
                    }
                    break;
                }
            case "copy":
            case "move":
                this.checkExpression(snode.lhs, scope);
                this.checkExpression(snode.rhs, scope);
                if (!helper.isSlice(snode.lhs.type) || !helper.isSlice(snode.rhs.type)) {
                    throw new errors_1.TypeError("'" + snode.op + "' is only allowed on slices", snode.loc);
                }
                let t = types_1.RestrictedType.strip(snode.lhs.type);
                let e = types_1.RestrictedType.strip(t.getElementType());
                if (helper.isConst(snode.lhs.type)) {
                    throw new errors_1.TypeError("'" + snode.op + "' requires a non-const slice as its first argument", snode.lhs.loc);
                }
                if (helper.isConst(t.arrayType)) {
                    throw new errors_1.TypeError("'" + snode.op + "' requires a non-const slice as its first argument", snode.lhs.loc);
                }
                if (snode.op == "move" && !(helper.isPureValue(e) || helper.isConst(e) || e == Static_1.Static.t_string) && helper.isConst(snode.rhs.type)) {
                    throw new errors_1.TypeError("'move' requires a non-const slice as its second argument when slice elements are neither const nor pure values", snode.lhs.loc);
                }
                let t2 = types_1.RestrictedType.strip(snode.rhs.type);
                let e2 = types_1.RestrictedType.strip(t2.getElementType());
                if (!this.checkTypeEquality(e, e2, snode.loc, false)) {
                    throw new errors_1.TypeError("'" + snode.op + "' requires two slices of the same type", snode.loc);
                }
                break;
            case "slice":
                this.checkExpression(snode.parameters[0], scope);
                this.checkExpression(snode.parameters[1], scope);
                this.checkExpression(snode.parameters[2], scope);
                if (!helper.isSlice(snode.parameters[0].type)) {
                    throw new errors_1.TypeError("'slice' is only allowed on slices", snode.loc);
                }
                if (helper.isLocalReference(snode.parameters[0].type)) {
                    throw new errors_1.TypeError("'slice' is not allowed on local references", snode.loc);
                }
                this.checkIsAssignable(snode.parameters[0], scope, true);
                this.checkIsPlatformIntNumber(snode.parameters[1]);
                this.checkIsPlatformIntNumber(snode.parameters[2]);
                break;
            case "println":
                for (let i = 0; i < snode.parameters.length; i++) {
                    this.checkExpression(snode.parameters[i], scope);
                }
                break;
            case "push": // Push is handled together with tryPush and append which are both expressions
            case "append":
            default:
                this.checkExpression(snode, scope);
                if (snode.type instanceof types_1.ArrayLiteralType || snode.type instanceof types_1.ObjectLiteralType || snode.type instanceof types_1.TupleLiteralType) {
                    throw new errors_1.TypeError("Cannot infer type", snode.loc);
                }
        }
        return;
    }
    checkExpression(enode, scope) {
        switch (enode.op) {
            case "null":
                enode.type = Static_1.Static.t_null;
                break;
            case "bool":
                enode.type = Static_1.Static.t_bool;
                break;
            case "str":
                enode.type = Static_1.Static.t_string;
                break;
            case "rune":
                enode.type = Static_1.Static.t_rune;
                break;
            case "int":
                // TODO: Check ranges and use t_uint if required
                enode.type = Static_1.Static.t_int;
                break;
            case "float":
                // TODO: Check ranges
                enode.type = Static_1.Static.t_double;
                break;
            case "id":
                // TODO: ellipsis, optional
                let element = scope.resolveElement(enode.value);
                if (!element) {
                    throw new errors_1.TypeError("Unknown identifier " + enode.value, enode.loc);
                }
                enode.type = element.type;
                break;
            case "++":
            case "--":
                this.checkExpression(enode.lhs, scope);
                if (helper.isUnsafePointer(enode.lhs.type)) {
                    // Do nothing by intention
                }
                else {
                    this.checkIsIntNumber(enode.lhs);
                }
                this.checkIsAssignable(enode.lhs, scope);
                enode.type = enode.lhs.type;
                break;
            case "unary-":
                this.checkExpression(enode.rhs, scope);
                this.checkIsSignedNumber(enode.rhs);
                if (enode.rhs.op == "int" || enode.rhs.op == "float") {
                    enode.op = enode.rhs.op;
                    enode.value = (-parseFloat(enode.rhs.value)).toString(); // TODO: BigNumber
                }
                enode.type = helper.stripType(enode.rhs.type);
                break;
            case "unary+":
                this.checkExpression(enode.rhs, scope);
                this.checkIsNumber(enode.rhs);
                if (enode.rhs.op == "int" || enode.rhs.op == "float") {
                    enode.op = enode.rhs.op;
                    enode.value = enode.rhs.value;
                }
                enode.type = helper.stripType(enode.rhs.type);
                break;
            case "unary^":
                this.checkExpression(enode.rhs, scope);
                this.checkIsIntNumber(enode.rhs);
                if (enode.rhs.op == "int") {
                    enode.op = enode.rhs.op;
                    enode.value = (~parseInt(enode.rhs.value)).toString();
                }
                enode.type = helper.stripType(enode.rhs.type);
                break;
            case "unary!":
                this.checkExpression(enode.rhs, scope);
                this.checkIsBool(enode.rhs);
                if (enode.rhs.op == "bool") {
                    enode.op = enode.rhs.op;
                    enode.value = enode.rhs.value == "true" ? "false" : "true";
                }
                enode.type = Static_1.Static.t_bool;
                break;
            case "unary*":
                {
                    this.checkExpression(enode.rhs, scope);
                    this.checkIsPointer(enode.rhs);
                    let t = helper.stripType(enode.rhs.type);
                    enode.type = t.elementType;
                    if (helper.stripType(enode.type) instanceof types_1.InterfaceType) {
                        throw new errors_1.TypeError("Interfaces cannot be dereferenced", enode.loc);
                    }
                    break;
                }
            case "unary&":
                {
                    this.checkExpression(enode.rhs, scope);
                    if (enode.isUnifyableLiteral()) {
                        enode.type = this.defaultLiteralType(enode);
                    }
                    else if (enode.rhs.isLiteral()) {
                        enode.type = new types_1.PointerType(enode.rhs.type, "strong");
                    }
                    else {
                        this.checkIsAddressable(enode.rhs, scope, true, true);
                        enode.type = new types_1.PointerType(enode.rhs.type, "local_reference");
                        // A reference to a non-mutable variable (e.g. 'let') must not be dereferenced and assigned to -> const
                        if (!this.checkIsAssignable(enode.rhs, scope, false)) {
                            enode.type = helper.applyConst(enode.type, enode.loc);
                        }
                    }
                    break;
                }
            case "+":
            case "*":
            case "/":
            case "-":
            case ">":
            case "<":
            case "<=":
            case ">=":
                this.checkExpression(enode.lhs, scope);
                this.checkExpression(enode.rhs, scope);
                if ((enode.op == "+" || enode.op == ">" || enode.op == "<" || enode.op == ">=" || enode.op == "<=") && helper.isString(enode.lhs.type)) {
                    this.checkIsString(enode.rhs);
                    if (enode.lhs.op == "str" && enode.rhs.op == "str") {
                        enode.op = "str";
                        enode.value = enode.lhs.value + enode.rhs.value;
                    }
                    if (enode.op == "+" || enode.op == "str") {
                        enode.type = Static_1.Static.t_string;
                    }
                    else {
                        enode.type = Static_1.Static.t_bool;
                    }
                }
                else if (helper.isUnsafePointer(enode.lhs.type)) {
                    if (enode.op == "*" || enode.op == "/") {
                        throw new errors_1.TypeError("'" + enode.op + "' is an invalid operation on pointers", enode.loc);
                    }
                    if (enode.op == "+" || enode.op == "-") {
                        this.checkIsIntNumber(enode.rhs);
                    }
                    else {
                        this.checkIsAssignableType(enode.lhs.type, enode.rhs.type, enode.loc, "assign", true);
                    }
                    enode.type = helper.stripType(enode.lhs.type);
                }
                else if (helper.isUnsafePointer(enode.rhs.type)) {
                    if (enode.op == "*" || enode.op == "/") {
                        throw new errors_1.TypeError("'" + enode.op + "' is an invalid operation on pointers", enode.loc);
                    }
                    if (enode.op == "+" || enode.op == "-") {
                        this.checkIsIntNumber(enode.lhs);
                    }
                    else {
                        this.checkIsAssignableType(enode.lhs.type, enode.rhs.type, enode.loc, "assign", true);
                    }
                    enode.type = helper.stripType(enode.rhs.type);
                }
                else {
                    this.checkIsNumber(enode.lhs);
                    this.checkIsNumber(enode.rhs);
                    // If lhs and rhs are constants, compute at compile time
                    if ((enode.lhs.op == "int" || enode.lhs.op == "float") && (enode.rhs.op == "int" || enode.rhs.op == "float")) {
                        // TODO: parse in a BigNumber representation
                        let l = parseFloat(enode.lhs.value);
                        let r = parseFloat(enode.rhs.value);
                        switch (enode.op) {
                            case "+":
                                enode.value = (l + r).toString();
                                break;
                            case "*":
                                enode.value = (l * r).toString();
                                break;
                            case "/":
                                // TODO: integer division
                                enode.value = (l / r).toString();
                                break;
                            case "-":
                                enode.value = (l - r).toString();
                                break;
                            case ">":
                                enode.value = (l > r) ? "true" : "false";
                                break;
                            case "<":
                                enode.value = (l < r) ? "true" : "false";
                                break;
                            case "<=":
                                enode.value = (l <= r) ? "true" : "false";
                                break;
                            case ">=":
                                enode.value = (l >= r) ? "true" : "false";
                                break;
                        }
                        if (enode.lhs.op == "float" || enode.rhs.op == "float") {
                            enode.op = "float";
                            enode.lhs.type = Static_1.Static.t_double;
                            enode.rhs.type = Static_1.Static.t_double;
                        }
                        else {
                            enode.op = "int";
                        }
                    }
                    else if (enode.op == "<" && this.checkIsUnsignedNumber(enode.lhs, false) && enode.rhs.op == "int" && enode.rhs.value == "0") {
                        enode.op = "bool";
                        enode.value = "false";
                    }
                    else if (enode.op == ">" && this.checkIsUnsignedNumber(enode.rhs, false) && enode.lhs.op == "int" && enode.lhs.value == "0") {
                        enode.op = "bool";
                        enode.value = "false";
                    }
                    else if (enode.lhs.op == "int" || enode.lhs.op == "float" || enode.lhs.op == "rune") {
                        this.unifyLiterals(enode.rhs.type, enode.lhs, scope, enode.loc);
                    }
                    else if (enode.rhs.op == "int" || enode.rhs.op == "float" || enode.rhs.op == "rune") {
                        this.unifyLiterals(enode.lhs.type, enode.rhs, scope, enode.loc);
                    }
                    else {
                        this.checkIsAssignableType(enode.lhs.type, enode.rhs.type, enode.loc, "assign", true);
                    }
                    if (enode.op == "+" || enode.op == "-" || enode.op == "*" || enode.op == "/" || enode.op == "float" || enode.op == "int") {
                        enode.type = helper.stripType(enode.lhs.type);
                    }
                    else {
                        enode.type = Static_1.Static.t_bool;
                    }
                }
                break;
            case "||":
            case "&&":
                this.checkExpression(enode.lhs, scope);
                this.checkExpression(enode.rhs, scope);
                this.checkIsBool(enode.lhs);
                this.checkIsBool(enode.rhs);
                enode.type = Static_1.Static.t_bool;
                break;
            case "&":
            case "|":
            case "^":
            case "&^":
            case "%":
            case "<<":
            case ">>":
                this.checkExpression(enode.lhs, scope);
                this.checkExpression(enode.rhs, scope);
                this.checkIsIntNumberOrUnsafePointer(enode.lhs);
                this.checkIsIntNumber(enode.rhs);
                if (enode.lhs.op == "int" && enode.rhs.op == "int") {
                    // TODO: parse in a BigNumber representation
                    let l = parseFloat(enode.lhs.value);
                    let r = parseFloat(enode.rhs.value);
                    switch (enode.op) {
                        case "&":
                            enode.value = (l & r).toString();
                            break;
                        case "|":
                            enode.value = (l | r).toString();
                            break;
                        case "^":
                            enode.value = (l ^ r).toString();
                            break;
                        case "&^":
                            enode.value = (l & ~r).toString();
                            break;
                        case "%":
                            enode.value = (l % r).toString();
                            break;
                        case "<<":
                            enode.value = (l << r).toString();
                            break;
                        case ">>":
                            enode.value = (l >> r).toString();
                            break;
                    }
                    enode.op = "int";
                }
                else if (enode.lhs.op == "int") {
                    if (enode.op == "<<" || enode.op == ">>") {
                        this.unifyLiterals(Static_1.Static.t_uint, enode.lhs, scope, enode.loc);
                        this.checkIsUnsignedNumber(enode.rhs);
                    }
                    else {
                        this.unifyLiterals(enode.rhs.type, enode.lhs, scope, enode.loc);
                    }
                }
                else if (enode.rhs.op == "int") {
                    if (enode.op == "<<" || enode.op == ">>") {
                        this.unifyLiterals(Static_1.Static.t_uint, enode.rhs, scope, enode.loc);
                    }
                    else {
                        this.unifyLiterals(enode.lhs.type, enode.rhs, scope, enode.loc);
                    }
                }
                else {
                    if (enode.op == "<<" || enode.op == ">>") {
                        this.checkIsUnsignedNumber(enode.rhs);
                    }
                    else if (helper.isUnsafePointer(enode.lhs.type)) {
                        this.checkIsAssignableType(Static_1.Static.t_uint, enode.rhs.type, enode.rhs.loc, "assign", true);
                    }
                    else {
                        this.checkIsAssignableType(enode.lhs.type, enode.rhs.type, enode.loc, "assign", true);
                    }
                }
                enode.type = helper.stripType(enode.lhs.type);
                break;
            case "==":
            case "!=":
                {
                    this.checkExpression(enode.lhs, scope);
                    this.checkExpression(enode.rhs, scope);
                    let tl = helper.stripType(enode.lhs.type);
                    if (tl instanceof types_1.OrType && !tl.stringsOnly()) {
                        throw new errors_1.TypeError("Or'ed types cannot be compared", enode.lhs.loc);
                    }
                    let tr = helper.stripType(enode.rhs.type);
                    if (tr instanceof types_1.OrType && !tr.stringsOnly()) {
                        throw new errors_1.TypeError("Or'ed types cannot be compared", enode.rhs.loc);
                    }
                    if ((enode.lhs.op == "int" || enode.lhs.op == "float") && (enode.rhs.op == "int" || enode.rhs.op == "float")) {
                        // Compare two number literals
                        // TODO: parse in a BigNumber representation
                        let l = parseFloat(enode.lhs.value);
                        let r = parseFloat(enode.rhs.value);
                        if (enode.op == "==") {
                            enode.value = (l == r) ? "true" : "false";
                        }
                        else {
                            enode.value = (l != r) ? "true" : "false";
                        }
                        enode.op = "bool";
                    }
                    else if (enode.lhs.op == "str" && enode.rhs.op == "str") {
                        // Compare two string literals?
                        if (enode.op == "==") {
                            enode.value = (enode.lhs.value == enode.rhs.value) ? "true" : "false";
                        }
                        else {
                            enode.value = (enode.lhs.value != enode.rhs.value) ? "true" : "false";
                        }
                        enode.op = "bool";
                    }
                    else if (enode.lhs.op == "null" && enode.rhs.op == "null") {
                        // Compare two nulls?
                        enode.value = (enode.op == "==" ? "true" : "false");
                        enode.op = "bool";
                    }
                    else if (enode.lhs.isUnifyableLiteral()) {
                        if ((tr instanceof types_1.PointerType || tr instanceof types_1.SliceType) && enode.lhs.op != "null") {
                            throw new errors_1.TypeError("Pointers and literals cannot be compared", enode.loc);
                        }
                        this.unifyLiterals(enode.rhs.type, enode.lhs, scope, enode.loc);
                    }
                    else if (enode.rhs.isUnifyableLiteral()) {
                        if ((tl instanceof types_1.PointerType || tl instanceof types_1.SliceType) && enode.rhs.op != "null") {
                            throw new errors_1.TypeError("Pointers and literals cannot be compared", enode.loc);
                        }
                        this.unifyLiterals(enode.lhs.type, enode.rhs, scope, enode.loc);
                    }
                    else {
                        this.checkIsAssignableType(tl, tr, enode.loc, "compare", true);
                    }
                    enode.type = Static_1.Static.t_bool;
                    break;
                }
            case ".":
                {
                    this.checkExpression(enode.lhs, scope);
                    let type = helper.stripType(enode.lhs.type);
                    let name = enode.name.value;
                    if (type instanceof types_1.PackageType) {
                        if (!type.pkg.scope.elements.has(name)) {
                            throw new errors_1.TypeError("Unknown identifier " + name + " in package " + type.pkg.pkgPath, enode.name.loc);
                        }
                        enode.type = type.pkg.scope.elements.get(name).type;
                        break;
                    }
                    let objectType = type;
                    let isConst = helper.isConst(enode.lhs.type);
                    if (type instanceof types_1.PointerType || type instanceof types_1.UnsafePointerType) {
                        isConst = helper.isConst(type.elementType);
                        objectType = helper.stripType(type.elementType);
                    }
                    else if (type instanceof types_1.StructType) {
                        objectType = type;
                        type = new types_1.PointerType(type, "reference");
                    }
                    else {
                        throw new errors_1.TypeError("Unknown field or method " + name + " in " + type.toString(), enode.name.loc);
                    }
                    if (objectType instanceof types_1.StructType) {
                        let field = objectType.field(name);
                        if (field) {
                            enode.type = field.type;
                            if (isConst) {
                                enode.type = helper.applyConst(enode.type, enode.loc);
                            }
                        }
                        else {
                            let method = objectType.method(name);
                            if (!method) {
                                throw new errors_1.TypeError("Unknown field or method " + name + " in " + objectType.toString(), enode.name.loc);
                            }
                            this.checkIsAssignableType(method.objectType, type, enode.loc, "assign", true);
                            enode.type = method;
                        }
                    }
                    else if (objectType instanceof types_1.InterfaceType) {
                        let method = objectType.method(name);
                        if (!method) {
                            throw new errors_1.TypeError("Unknown method " + name + " in " + objectType.toString(), enode.name.loc);
                        }
                        this.checkIsAssignableType(method.objectType, type, enode.loc, "assign", true);
                        enode.type = method;
                    }
                    else {
                        throw new errors_1.TypeError("Unknown field or method " + name + " in " + objectType.toString(), enode.name.loc);
                    }
                    break;
                }
            case ":":
                {
                    this.checkExpression(enode.lhs, scope);
                    let index1 = 0;
                    let index2 = 0;
                    let indicesAreNumbers = 0;
                    if (enode.parameters[0]) {
                        this.checkExpression(enode.parameters[0], scope);
                        this.checkIsIntNumber(enode.parameters[0]);
                        if (enode.parameters[0].op == "int") {
                            index1 = parseInt(enode.parameters[0].value);
                            indicesAreNumbers++;
                        }
                    }
                    if (enode.parameters[1]) {
                        this.checkExpression(enode.parameters[1], scope);
                        this.checkIsIntNumber(enode.parameters[1]);
                        if (enode.parameters[1].op == "int") {
                            index2 = parseInt(enode.parameters[1].value);
                            indicesAreNumbers++;
                        }
                    }
                    if (indicesAreNumbers == 2 && index1 > index2) {
                        throw new errors_1.TypeError("Index out of range", enode.rhs.loc);
                    }
                    let isConst = helper.isConst(enode.lhs.type);
                    let t = helper.stripType(enode.lhs.type);
                    let elementType = this.checkIsIndexable(enode.lhs, index1);
                    if (t instanceof types_1.ArrayType) {
                        this.checkIsAddressable(enode.lhs, scope, false);
                        this.checkIsIndexable(enode.lhs, index2, true);
                        enode.type = new types_1.SliceType(t, "local_reference");
                    }
                    else if (t instanceof types_1.UnsafePointerType) {
                        enode.type = new types_1.SliceType(enode.lhs.type, "local_reference");
                        if (isConst) {
                            enode.type = helper.applyConst(enode.type, enode.loc);
                        }
                    }
                    else if (helper.isMap(t)) {
                        throw new errors_1.TypeError("Ranges are not supported on maps", enode.loc);
                    }
                    else if (t instanceof types_1.SliceType) {
                        let isTakeExpr = helper.isTakeExpression(enode.lhs);
                        if ((t.mode == "unique" || t.mode == "strong") && !isTakeExpr) {
                            enode.type = new types_1.SliceType(t.arrayType, "reference");
                        }
                        else {
                            // For slices the type remains the same
                            enode.type = enode.lhs.type;
                        }
                    }
                    else if (t == Static_1.Static.t_string) {
                        enode.type = Static_1.Static.t_string;
                    }
                    else {
                        throw new errors_1.ImplementationError();
                    }
                    break;
                }
            case "[":
                {
                    this.checkExpression(enode.lhs, scope);
                    this.checkExpression(enode.rhs, scope);
                    let isConst = helper.isConst(enode.lhs.type);
                    let t = helper.stripType(enode.lhs.type);
                    if (t instanceof types_1.TupleType) {
                        this.checkIsIntNumber(enode.rhs);
                        if (enode.rhs.op != "int") {
                            throw new errors_1.TypeError("Index inside a tuple must be a constant number", enode.lhs.loc);
                        }
                        let index = parseInt(enode.rhs.value);
                        enode.type = this.checkIsIndexable(enode.lhs, index);
                    }
                    else if (t instanceof types_1.ArrayType) {
                        this.checkIsPlatformIntNumber(enode.rhs);
                        let index = 0;
                        if (enode.rhs.op == "int") {
                            index = parseInt(enode.rhs.value);
                        }
                        enode.type = this.checkIsIndexable(enode.lhs, index);
                    }
                    else if (helper.isMap(t)) {
                        isConst = isConst || helper.isConst(t.elementType);
                        if (enode.rhs.isUnifyableLiteral()) {
                            this.unifyLiterals(this.mapKeyType(t), enode.rhs, scope, enode.rhs.loc);
                        }
                        else {
                            this.checkIsAssignableType(this.mapKeyType(t), enode.rhs.type, enode.rhs.loc, "assign", true);
                        }
                        enode.type = this.mapValueType(t);
                    }
                    else if (t instanceof types_1.SliceType) {
                        this.checkIsPlatformIntNumber(enode.rhs);
                        enode.type = t.getElementType();
                        isConst = isConst || helper.isConst(t.arrayType);
                    }
                    else if (t == Static_1.Static.t_string) {
                        this.checkIsPlatformIntNumber(enode.rhs);
                        enode.type = Static_1.Static.t_byte;
                    }
                    else if (t instanceof types_1.UnsafePointerType) {
                        this.checkIsPlatformIntNumber(enode.rhs);
                        enode.type = t.elementType;
                    }
                    else {
                        throw new errors_1.TypeError("[] operator is not allowed on " + enode.lhs.type.toString(), enode.loc);
                    }
                    if (isConst) {
                        enode.type = helper.applyConst(enode.type, enode.loc);
                    }
                    break;
                }
            case "(":
                {
                    this.checkExpression(enode.lhs, scope);
                    if (enode.parameters) {
                        for (let pnode of enode.parameters) {
                            if (pnode.op == "unary...") {
                                this.checkExpression(pnode.rhs, scope);
                            }
                            else {
                                this.checkExpression(pnode, scope);
                            }
                        }
                    }
                    let t = helper.stripType(enode.lhs.type);
                    if (t instanceof types_1.TemplateType) {
                        let result = this.checkTemplateFunctionArguments(t, enode.parameters, scope, enode.loc);
                        let types = [];
                        for (let i = 0; i < t.templateParameterNames.length; i++) {
                            let tt = result.get(t.templateParameterNames[i]);
                            types.push(tt);
                        }
                        let f = this.instantiateTemplateFunction(t, types, enode.loc);
                        //                    console.log("Instantiate template", f.name);
                        enode.type = f.type.returnType;
                        enode.lhs.type = f.type;
                        t = f.type;
                    }
                    else if (t instanceof types_1.FunctionType) {
                        this.checkFunctionArguments(t, enode.parameters, scope, enode.loc);
                        enode.type = t.returnType;
                    }
                    else {
                        throw new errors_1.TypeError("Expression is not a function", enode.loc);
                    }
                    if (!helper.isConst(t.objectType) && enode.lhs.op == "." && !helper.isSafePointer(enode.lhs.lhs.type) && !helper.isUnsafePointer(enode.lhs.lhs.type)) {
                        this.checkIsMutable(enode.lhs.lhs, scope, true);
                    }
                    break;
                }
            case "genericInstance":
                this.checkExpression(enode.lhs, scope);
                enode.type = this.instantiateTemplateFunctionFromNode(enode, scope).type;
                break;
            case "make":
                {
                    enode.lhs.type = this.createType(enode.lhs, scope);
                    if (enode.parameters.length > 0) {
                        if (enode.parameters.length > 2) {
                            throw new errors_1.TypeError("make accepts at most two parameters", enode.loc);
                        }
                        this.checkExpression(enode.parameters[0], scope);
                        this.checkIsPlatformIntNumber(enode.parameters[0], true);
                        if (enode.parameters.length == 2) {
                            this.checkExpression(enode.parameters[1], scope);
                            this.checkIsPlatformIntNumber(enode.parameters[1], true);
                        }
                        enode.type = new types_1.SliceType(new types_1.ArrayType(enode.lhs.type, -1), "strong");
                    }
                    else {
                        enode.type = new types_1.PointerType(enode.lhs.type, "strong");
                    }
                    break;
                }
            case "tuple":
                {
                    let types = [];
                    for (let p of enode.parameters) {
                        this.checkExpression(p, scope);
                        types.push(p.type);
                    }
                    let t = new types_1.TupleLiteralType(types);
                    enode.type = t;
                    if (enode.lhs) {
                        let ct = this.createType(enode.lhs, scope, "variable_toplevel");
                        this.unifyLiterals(ct, enode, scope, enode.loc);
                    }
                    break;
                }
            case "array":
                {
                    let types = [];
                    if (enode.parameters) {
                        for (var i = 0; i < enode.parameters.length; i++) {
                            let p = enode.parameters[i];
                            if (p.op == "...") {
                                // Do nothing by intention
                            }
                            else {
                                this.checkExpression(p, scope);
                                types.push(p.type);
                            }
                        }
                    }
                    let t = new types_1.ArrayLiteralType(types);
                    enode.type = t;
                    if (enode.lhs) {
                        let ct = this.createType(enode.lhs, scope, "variable_toplevel");
                        this.unifyLiterals(ct, enode, scope, enode.loc);
                    }
                    break;
                }
            case "object":
                {
                    let types = new Map();
                    if (enode.parameters) {
                        for (let p of enode.parameters) {
                            this.checkExpression(p.lhs, scope);
                            types.set(p.name.value, p.lhs.type);
                        }
                    }
                    let t = new types_1.ObjectLiteralType(types);
                    enode.type = t;
                    if (enode.lhs) {
                        let ct = this.createType(enode.lhs, scope, "variable_toplevel");
                        this.unifyLiterals(ct, enode, scope, enode.loc);
                    }
                    break;
                }
            case "=>":
                {
                    let f = new scope_1.Function();
                    f.loc = enode.loc;
                    f.node = enode;
                    f.scope = new scope_1.Scope(scope);
                    f.scope.func = f;
                    f.type = new types_1.FunctionType();
                    f.type.loc = enode.loc;
                    enode.scope = f.scope;
                    enode.type = f.type;
                    if (enode.parameters) {
                        for (let pnode of enode.parameters) {
                            let original_pnode = pnode;
                            var p = new scope_1.FunctionParameter();
                            if (pnode.op == "ellipsisParam") {
                                p.ellipsis = true;
                                pnode = pnode.lhs;
                            }
                            p.name = pnode.name.value;
                            for (let param of f.type.parameters) {
                                if (param.name == p.name) {
                                    throw new errors_1.TypeError("Duplicate parameter name " + p.name, pnode.loc);
                                }
                            }
                            p.type = this.createType(pnode, enode.scope, "parameter");
                            if (p.ellipsis && !(p.type instanceof types_1.SliceType)) {
                                throw new errors_1.TypeError("Ellipsis parameters must be of a slice type", pnode.loc);
                            }
                            p.loc = pnode.loc;
                            f.type.parameters.push(p);
                            f.scope.registerElement(p.name, p);
                        }
                    }
                    // Return type?
                    if (enode.lhs) {
                        f.type.returnType = this.createType(enode.lhs, f.scope, "parameter_toplevel");
                        if (enode.lhs.op == "tupleType") {
                            for (let i = 0; i < enode.lhs.parameters.length; i++) {
                                let pnode = enode.lhs.parameters[i];
                                if (i == 0) {
                                    f.hasNamedReturnVariables = !!pnode.name;
                                }
                                else if ((f.hasNamedReturnVariables && !pnode.name) || (!f.hasNamedReturnVariables && !!pnode.name)) {
                                    throw new errors_1.TypeError("Either all or no return variables are named", pnode.loc);
                                }
                            }
                        }
                        if (enode.lhs.op == "tupleType") {
                            for (let i = 0; i < enode.lhs.parameters.length; i++) {
                                let pnode = enode.lhs.parameters[i];
                                if (pnode.name) {
                                    let v = new scope_1.Variable();
                                    v.isResult = true;
                                    v.loc = pnode.loc;
                                    v.name = pnode.name.value;
                                    v.type = f.type.returnType.types[i];
                                    f.scope.registerElement(v.name, v);
                                    if (!f.namedReturnVariables) {
                                        f.namedReturnVariables = [];
                                    }
                                    f.namedReturnVariables = [v];
                                }
                            }
                        }
                        else {
                            let v = new scope_1.Variable();
                            v.isResult = true;
                            v.loc = enode.loc;
                            v.name = "return";
                            v.type = f.type.returnType;
                            f.unnamedReturnVariable = v;
                        }
                    }
                    // Return expression or return statements?
                    if (enode.rhs) {
                        this.checkExpression(enode.rhs, enode.scope);
                        if (!f.type.returnType) {
                            f.type.returnType = enode.rhs.type;
                            let v = new scope_1.Variable();
                            v.isResult = true;
                            v.loc = enode.loc;
                            v.name = "return";
                            v.type = f.type.returnType;
                            f.unnamedReturnVariable = v;
                        }
                        else {
                            this.checkIsAssignableNode(f.type.returnType, enode.rhs, scope);
                        }
                    }
                    else {
                        enode.scopeExit = this.checkStatements(enode.statements, enode.scope);
                    }
                    break;
                }
            case "is":
                {
                    this.checkExpression(enode.lhs, scope);
                    let t = this.createType(enode.rhs, scope, "variable_toplevel");
                    enode.rhs.type = t;
                    if (helper.isOrType(enode.lhs.type)) {
                        let ot = helper.stripType(enode.lhs.type);
                        let found = false;
                        for (var option of ot.types) {
                            if (this.checkTypeEquality(option, t, enode.loc, false)) {
                                found = true;
                                break;
                            }
                        }
                        if (!found) {
                            throw new errors_1.TypeError("Type " + t.toString() + " is not part of " + ot.toString(), enode.rhs.loc);
                        }
                    }
                    else {
                        this.checkIsInterface(enode.lhs);
                        if (!helper.isSafePointer(t) || !helper.isStruct(types_1.RestrictedType.strip(types_1.RestrictedType.strip(t).elementType))) {
                            throw new errors_1.TypeError("Interface can only contain pointers to structs, but not " + t.toString(), enode.loc);
                        }
                    }
                    enode.type = Static_1.Static.t_bool;
                    break;
                }
            case "typeCast":
                {
                    let t = this.createType(enode.lhs, scope, "parameter_toplevel");
                    this.checkExpression(enode.rhs, scope);
                    let right = types_1.RestrictedType.strip(enode.rhs.type);
                    let left = types_1.RestrictedType.strip(t);
                    if ((left == Static_1.Static.t_float || left == Static_1.Static.t_double) && helper.isIntNumber(right)) {
                        // Ints can be converted to floats
                        enode.type = t;
                    }
                    else if (helper.isIntNumber(left) && (right == Static_1.Static.t_float || right == Static_1.Static.t_double)) {
                        // Floats can be converted to ints
                        enode.type = t;
                    }
                    else if (left == Static_1.Static.t_float && right == Static_1.Static.t_double) {
                        // Doubles can be converted to floats
                        enode.type = t;
                    }
                    else if (left == Static_1.Static.t_double && right == Static_1.Static.t_float) {
                        // Floats can be converted to doubles
                        enode.type = t;
                    }
                    else if (left == Static_1.Static.t_rune && helper.isUInt32Number(right)) {
                        // Runes can be converted to uint32
                        enode.type = t;
                    }
                    else if (helper.isUInt32Number(left) && right == Static_1.Static.t_rune) {
                        // Uint32 can be converted to a rune
                        enode.type = t;
                        //                } else if (helper.isInt32Number(t) && right instanceof UnsafePointerType) {
                        //                    // Unsafe pointers can be converted to 32-bit integers
                        //                    enode.type = t;
                    }
                    else if (left instanceof types_1.UnsafePointerType && (right instanceof types_1.UnsafePointerType || right instanceof types_1.PointerType || helper.isString(right) || helper.isInt32Number(right))) {
                        // Unsafe pointers to anything, safe pointers to anything, strings and 32-bit integers can be converted to any unsafe pointer
                        enode.type = t;
                    }
                    else if ((left == Static_1.Static.t_bool || helper.isIntNumber(left)) && (right == Static_1.Static.t_bool || helper.isIntNumber(right)) && t != right) {
                        // bool and all integers can be converted into each other
                        enode.type = t;
                    }
                    else if (helper.isString(left) && right instanceof types_1.UnsafePointerType) {
                        // An unsafe pointer can be converted to a string by doing nothing. This is an unsafe cast.
                        enode.type = t;
                    }
                    else if (helper.isString(left) && right instanceof types_1.SliceType && !helper.isLocalReference(right) && !helper.isConst(enode.rhs.type) && (right.getElementType() == Static_1.Static.t_byte || right.getElementType() == Static_1.Static.t_char)) {
                        // A slice of bytes or chars can be converted to a string by taking ownership away from the slice
                        enode.type = t;
                    }
                    else if (left instanceof types_1.SliceType && (left.mode == "strong" || left.mode == "unique" || helper.isConst(left.arrayType)) && (left.getElementType() == Static_1.Static.t_byte || left.getElementType() == Static_1.Static.t_char) && helper.isString(right)) {
                        // A string can be casted into a unique or owning sequence of bytes or chars by copying it.
                        // A string can be casted to a const slice without copying it.
                        enode.type = t;
                    }
                    else if (helper.isComplexOrType(right)) {
                        let ok = false;
                        for (let ot of right.types) {
                            if (this.checkIsAssignableType(t, ot, enode.loc, "assign", false)) {
                                enode.type = t;
                                ok = true;
                                break;
                            }
                        }
                        if (!ok) {
                            throw new errors_1.TypeError("Conversion from " + right.toString() + " to " + t.toString() + " is not possible", enode.loc);
                        }
                    }
                    else if (this.checkIsAssignableType(t, enode.rhs.type, enode.loc, "assign", false)) {
                        /*
                        // null can be casted, especially when it is assigned to interface{}
                        if (right != Static.t_null) {
                            throw new TypeError("Conversion from " + right.toString() + " to " + t.toString() + " does not require a cast", enode.loc);
                        }
                        */
                        enode.type = t;
                    }
                    else {
                        throw new errors_1.TypeError("Conversion from " + enode.rhs.type.toString() + " to " + t.toString() + " is not possible", enode.loc);
                        //                    throw new TodoError("conversion not possible or not implemented")
                    }
                    break;
                }
            case "take":
                this.checkExpression(enode.lhs, scope);
                switch (enode.lhs.op) {
                    case "id":
                    case "[":
                    case ".":
                        enode.lhs.flags |= parser_1.AstFlags.ZeroAfterAssignment;
                        this.checkIsMutable(enode.lhs, scope);
                        break;
                    case "typeCast":
                        let n = enode.lhs;
                        while (n.op == "typeCast" && helper.isOrType(n.rhs.type)) {
                            n = n.rhs;
                        }
                        n.flags |= parser_1.AstFlags.ZeroAfterAssignment;
                        this.checkIsMutable(n, scope);
                        break;
                    default:
                        throw new errors_1.TypeError("take() can only be applied on variables, object fields or slice/array elements", enode.lhs.loc);
                }
                enode.type = enode.lhs.type;
                break;
            case "len":
                this.checkExpression(enode.lhs, scope);
                if (!helper.isString(enode.lhs.type) && !helper.isArray(enode.lhs.type) && !helper.isSlice(enode.lhs.type)) {
                    throw new errors_1.TypeError("'len' is only allowed on strings, arrays and slices", enode.loc);
                }
                enode.type = Static_1.Static.t_int;
                break;
            case "cap":
                this.checkExpression(enode.lhs, scope);
                if (!helper.isSlice(enode.lhs.type)) {
                    throw new errors_1.TypeError("'cap' is only allowed on slices", enode.loc);
                }
                enode.type = Static_1.Static.t_int;
                break;
            case "clone":
                {
                    this.checkExpression(enode.lhs, scope);
                    if (!helper.isSlice(enode.lhs.type)) {
                        throw new errors_1.TypeError("'clone' is only allowed on slices", enode.loc);
                    }
                    let t = types_1.RestrictedType.strip(enode.lhs.type);
                    if (!helper.isPureValue(t.getElementType())) {
                        throw new errors_1.TypeError("'clone' cannot work on slices which contain pointer-like types", enode.loc);
                    }
                    enode.type = new types_1.SliceType(t.arrayType, "unique");
                    break;
                }
            case "sizeof":
            case "aligned_sizeof":
                enode.lhs.type = this.createType(enode.lhs, scope, "default");
                enode.type = Static_1.Static.t_int;
                break;
            case "max":
            case "min":
                enode.lhs.type = this.createType(enode.lhs, scope, "default");
                this.checkIsNumber(enode.lhs, true);
                enode.type = enode.lhs.type;
                break;
            case "pop": {
                this.checkExpression(enode.lhs, scope);
                if (!helper.isSlice(enode.lhs.type)) {
                    throw new errors_1.TypeError("'pop' is only allowed on slices", enode.loc);
                }
                this.checkIsAssignable(enode.lhs, scope, true);
                let t = types_1.RestrictedType.strip(enode.lhs.type);
                enode.type = t.getElementType();
                break;
            }
            case "push":
            case "tryPush":
            case "append":
                {
                    if (enode.parameters.length < 2) {
                        throw new errors_1.TypeError("'" + enode.op + "' expects at least two arguments", enode.loc);
                    }
                    let e;
                    for (let i = 0; i < enode.parameters.length; i++) {
                        let p = enode.parameters[i];
                        let expand = false;
                        if (i > 0) {
                            if (p.op == "unary...") {
                                expand = true;
                                p = p.rhs;
                            }
                        }
                        this.checkExpression(p, scope);
                        if (i == 0) {
                            if (enode.op == "push" || enode.op == "tryPush" || enode.op == "append") {
                                this.checkIsAssignable(p, scope, true);
                            }
                            if (!helper.isSlice(p.type)) {
                                throw new errors_1.TypeError("First argument to '" + enode.op + "' must be a slice", p.loc);
                            }
                            let t = types_1.RestrictedType.strip(p.type);
                            if (helper.isConst(t.arrayType)) {
                                throw new errors_1.TypeError("First argument to '" + enode.op + "' must be a non-const slice", p.loc);
                            }
                            if (enode.op == "append") {
                                if (!helper.isUnique(t) && !helper.isStrong(t)) {
                                    throw new errors_1.TypeError("First argument to 'append' must be an owning pointer", p.loc);
                                }
                            }
                            e = t.getElementType();
                            if (helper.isPureValue(e)) {
                                // Remove constness in case of pure values.
                                e = types_1.RestrictedType.strip(e);
                            }
                        }
                        else {
                            if (expand) {
                                if (!helper.isSlice(p.type)) {
                                    throw new errors_1.TypeError("'...' must be followed by a slice", p.loc);
                                }
                                let e2 = types_1.RestrictedType.strip(p.type).getElementType();
                                if (helper.isPureValue(e)) {
                                    // Remove constness in case of pure values.
                                    e2 = types_1.RestrictedType.strip(e2);
                                }
                                this.checkTypeEquality(e, e2, p.loc, true);
                            }
                            else {
                                this.checkIsAssignableNode(e, p, scope, true);
                            }
                        }
                    }
                    if (enode.op == "append") {
                        enode.type = Static_1.Static.t_void;
                    }
                    else if (enode.op == "tryPush") {
                        enode.type = Static_1.Static.t_bool;
                    }
                    else if (enode.op == "push") {
                        enode.type = Static_1.Static.t_void;
                    }
                    else {
                        throw new errors_1.ImplementationError();
                    }
                    break;
                }
            case "coroutine":
                enode.type = Static_1.Static.t_coroutine;
                break;
            case "resume":
                enode.type = Static_1.Static.t_void;
                this.checkExpression(enode.lhs, scope);
                this.checkTypeEquality(Static_1.Static.t_coroutine, enode.lhs.type, enode.loc, true);
                break;
            case "ellipsisId":
            case "unary...":
                throw new errors_1.TypeError("'...' is not allowed in this context", enode.loc);
            case "optionalId":
                throw new errors_1.TypeError("'?' is not allowed in this context", enode.loc);
            default:
                throw new errors_1.ImplementationError(enode.op);
        }
    }
    checkGlobalVariable(v, scope) {
        if (v.node.rhs) {
            this.checkExpression(v.node.rhs, scope);
            if (!v.type) {
                if (v.node.rhs.isUnifyableLiteral()) {
                    v.type = this.defaultLiteralType(v.node.rhs);
                }
                else {
                    v.type = v.node.rhs.type;
                }
            }
            else {
                this.checkIsAssignableNode(v.type, v.node.rhs, scope);
            }
            /*
            if (helper.isSafePointer(v.node.rhs.type) || helper.isSlice(v.node.rhs.type)) {
                if (v.node.rhs.op != "id" && v.node.rhs.op != "take" && v.node.rhs.op != "array" && v.node.rhs.op != "object") {
                    throw new TypeError("Right hand side of assignment must be wrapped in take()", v.node.rhs.loc);
                }
            }
            */
            this.checkGroupsInSingleAssignment(v.type, scope.resolveGroup(v), null, v.node.rhs, false, scope, v.loc);
        }
        else {
            throw new errors_1.TypeError("Global variable " + v.name + " must be initialized", v.node.loc);
        }
    }
    defaultLiteralType(node) {
        if (node.op == "unary&") {
            this.defaultLiteralType(node.rhs);
            if (node.rhs.type instanceof types_1.ArrayType) {
                node.type = new types_1.SliceType(node.rhs.type, "strong");
            }
            else {
                node.type = new types_1.PointerType(node.rhs.type, "strong");
            }
        }
        else if (node.type instanceof types_1.ArrayLiteralType) {
            for (let pnode of node.parameters) {
                if (pnode.op == "...") {
                    throw new errors_1.TypeError("Cannot infer default type of array literal using the ... operator", pnode.loc);
                }
                this.defaultLiteralType(pnode);
            }
            if (node.parameters.length == 0) {
                throw new errors_1.TypeError("Cannot infer default type of []", node.loc);
            }
            else {
                let t = node.parameters[0].type;
                for (let i = 1; i < node.parameters.length; i++) {
                    if (!this.checkIsAssignableType(t, node.parameters[i].type, node.loc, "assign", false)) {
                        if (!(t instanceof types_1.OrType)) {
                            let o = new types_1.OrType();
                            o.types.push(t);
                            t = o;
                        }
                        t.types.push(node.parameters[i].type);
                    }
                }
                // TODO: Set the group of this new slice to unbound
                node.type = new types_1.SliceType(new types_1.ArrayType(t, -1), "strong");
            }
        }
        else if (node.type instanceof types_1.TupleLiteralType) {
            let types = [];
            for (let pnode of node.parameters) {
                this.defaultLiteralType(pnode);
                types.push(pnode.type);
            }
            node.type = new types_1.TupleType(types);
        }
        else if (node.type instanceof types_1.ObjectLiteralType) {
            throw new errors_1.TypeError("Cannot infer default type of object literal", node.loc);
            /*
            let s = new StructType();
            this.structs.push(s);
            node.type = s;
            for(let i = 0; i < node.parameters.length; i++) {
                let pnode = node.parameters[i];
                this.defaultLiteralType(pnode.lhs);
                let f = new StructField();
                f.type = pnode.lhs.type;
                f.name = pnode.name.value;
                (node.type as StructType).fields.push(f);
            }
            */
        }
        return node.type;
    }
    unifyLiterals(t, node, scope, loc, doThrow = true, templateParams = null, allowPointerIndirection = true) {
        if (templateParams && t instanceof types_1.GenericParameter && templateParams.has(t.name)) {
            t = templateParams.get(t.name);
        }
        if (helper.isOrType(t)) {
            let orType = helper.stripType(t);
            for (let o of orType.types) {
                if (this.unifyLiterals(o, node, scope, loc, false)) {
                    return true;
                }
            }
            if (doThrow) {
                throw new errors_1.TypeError("Literal of type " + node.type.toString() + " is not an option of " + t.toString(), node.loc);
            }
            return false;
        }
        if (helper.isAny(t)) {
            node.type = this.defaultLiteralType(node);
            return true;
        }
        if (allowPointerIndirection && helper.isSafePointer(t) && node.op == "object") {
            if (!this.unifyLiterals(this.pointerElementType(t), node, scope, loc, doThrow, templateParams, false)) {
                return false;
            }
            node.type = new types_1.PointerType(this.pointerElementType(t), "strong");
            return true;
        }
        if (allowPointerIndirection && helper.isSlice(t) && node.op == "array") {
            if (!this.unifyLiterals(this.sliceArrayType(t), node, scope, loc, doThrow, templateParams, false)) {
                return false;
            }
            node.type = new types_1.SliceType(this.sliceArrayType(t), "strong");
            return true;
        }
        if (templateParams && t instanceof types_1.GenericParameter) {
            node.type = this.defaultLiteralType(node);
            templateParams.set(t.name, node.type);
            return true;
        }
        switch (node.op) {
            case "int":
                node.type = t;
                t = helper.stripType(t);
                // TODO: Check range
                if (helper.isNumber(t)) {
                    return true;
                }
                if (t instanceof types_1.UnsafePointerType) {
                    // TODO: Check range
                    return true;
                }
                if (!doThrow) {
                    return false;
                }
                throw new errors_1.TypeError("Type mismatch between integer number and " + t.toString(), loc);
            case "float":
                node.type = t;
                t = helper.stripType(t);
                // TODO: Check range
                if (t == Static_1.Static.t_float || t == Static_1.Static.t_double) {
                    return true;
                }
                if (!doThrow) {
                    return false;
                }
                throw new errors_1.TypeError("Type mismatch between floating point number and " + t.toString(), loc);
            case "str":
                node.type = t;
                if (helper.isString(t)) {
                    return true;
                }
                else if (helper.isStringLiteralType(t)) {
                    if (t.name == node.value) {
                        return true;
                    }
                }
                if (!doThrow) {
                    return false;
                }
                throw new errors_1.TypeError("Type mismatch between string and " + t.toString(), loc);
            case "rune":
                node.type = t;
                t = helper.stripType(t);
                if (t == Static_1.Static.t_rune) {
                    return true;
                }
                else if ((t == Static_1.Static.t_char || t == Static_1.Static.t_int8) && node.numValue <= 127) {
                    node.value = node.numValue.toString();
                    return true;
                }
                else if ((t == Static_1.Static.t_byte || t == Static_1.Static.t_uint8) && node.numValue <= 255) {
                    node.value = node.numValue.toString();
                    return true;
                }
                else if (t == Static_1.Static.t_uint16 && node.numValue <= 65535) {
                    node.value = node.numValue.toString();
                    return true;
                }
                else if (t == Static_1.Static.t_int16 && node.numValue <= 32768) {
                    node.value = node.numValue.toString();
                    return true;
                }
                else if (t == Static_1.Static.t_int32 && node.numValue <= 2147483647) {
                    node.value = node.numValue.toString();
                    return true;
                }
                else if (t == Static_1.Static.t_uint32 || Static_1.Static.t_uint64 || Static_1.Static.t_int64) {
                    node.value = node.numValue.toString();
                    return true;
                }
                if (!doThrow) {
                    return false;
                }
                throw new errors_1.TypeError("Type mismatch between rune and " + t.toString(), loc);
            case "array":
                {
                    if (helper.isArray(t)) {
                        let arrayType = helper.stripType(t);
                        // An array?
                        if (arrayType.size != -1) {
                            // Count array elements
                            let count = 0;
                            let lastParameterIsEllipsis = false;
                            if (node.parameters) {
                                count = node.parameters.length;
                                for (var i = 0; i < node.parameters.length; i++) {
                                    let pnode = node.parameters[i];
                                    if (pnode.op == "...") {
                                        lastParameterIsEllipsis = true;
                                    }
                                }
                            }
                            if (count != arrayType.size) {
                                if (!lastParameterIsEllipsis || count > arrayType.size) {
                                    throw new errors_1.TypeError("Mismatch in array size", node.loc);
                                }
                                else {
                                    // Note that this literal is incomplete
                                    node.flags |= parser_1.AstFlags.FillArray;
                                }
                            }
                        }
                        if (node.parameters) {
                            let elementType = this.arrayElementType(t);
                            for (var i = 0; i < node.parameters.length; i++) {
                                let pnode = node.parameters[i];
                                node.flags |= (pnode.flags & parser_1.AstFlags.FillArray);
                                if (pnode.op == "...") {
                                    if (arrayType.size == -1) {
                                        throw new errors_1.TypeError("The ... operator is not allowed in slice literals", pnode.loc);
                                    }
                                    continue;
                                }
                                if (!this.checkIsAssignableNode(elementType, pnode, scope, doThrow)) {
                                    return false;
                                }
                            }
                        }
                        node.type = t;
                        return true;
                    }
                    if (!doThrow) {
                        return false;
                    }
                    throw new errors_1.TypeError("Type mismatch between array literal and " + t.toString(), loc);
                }
            case "tuple":
                if (helper.isTupleType(t)) {
                    let tupleType = helper.stripType(t);
                    if (node.parameters.length != tupleType.types.length) {
                        throw new errors_1.TypeError("Mismatch in tuple length", node.loc);
                    }
                    for (let i = 0; i < node.parameters.length; i++) {
                        let pnode = node.parameters[i];
                        if (!this.checkIsAssignableNode(tupleType.types[i], pnode, scope, doThrow)) {
                            return false;
                        }
                        node.flags |= (pnode.flags & parser_1.AstFlags.FillArray);
                    }
                    node.type = t;
                    return true;
                }
                if (!doThrow) {
                    return false;
                }
                throw new errors_1.TypeError("Type mismatch between tuple literal and " + t.toString(), loc);
            case "object":
                if (helper.isMap(t)) {
                    let valueType = this.mapValueType(t);
                    let keyType = this.mapKeyType(t);
                    if (!node.parameters || node.parameters.length == 0) {
                        // Empty map
                        node.type = t;
                    }
                    else if (helper.isString(keyType)) {
                        // A map, e.g. "{foo: 42}"
                        for (let pnode of node.parameters) {
                            if (!this.checkIsAssignableNode(valueType, pnode.lhs, scope, doThrow)) {
                                return false;
                            }
                            node.flags |= (pnode.flags & parser_1.AstFlags.FillArray);
                        }
                    }
                    node.type = t;
                    return true;
                }
                else if (helper.isStruct(t)) {
                    let structType = helper.stripType(t);
                    // A struct initialization
                    if (node.parameters) {
                        for (let pnode of node.parameters) {
                            node.flags |= (pnode.flags & parser_1.AstFlags.FillArray);
                            let field = structType.field(pnode.name.value);
                            if (!field) {
                                throw new errors_1.TypeError("Unknown field " + pnode.name.value + " in " + t.toString(), pnode.name.loc);
                            }
                            if (!this.checkIsAssignableNode(field.type, pnode.lhs, scope, doThrow)) {
                                return false;
                            }
                        }
                    }
                    node.type = t;
                    return true;
                }
                if (!doThrow) {
                    return false;
                }
                throw new errors_1.TypeError("Type mismatch between object literal and " + t.toString(), loc);
            case "unary&":
                if (t instanceof types_1.PointerType) {
                    //                    if (node.rhs.op == "id") {}
                    let r = this.unifyLiterals(t, node.rhs, scope, loc, doThrow, templateParams);
                    node.type = node.rhs.type;
                    return r;
                }
                if (!doThrow) {
                    return false;
                }
                throw new errors_1.TypeError("Type mismatch between object literal and " + t.toString(), loc);
            case "null":
                if (helper.isSafePointer(t) || helper.isUnsafePointer(t) || helper.isSlice(t)) {
                    node.type = t;
                    return true;
                }
                if (!doThrow) {
                    return false;
                }
                throw new errors_1.TypeError("Type mismatch between null and " + t.toString(), loc);
            default:
                throw new errors_1.ImplementationError();
        }
    }
    /**
     * Checks whether the type of 'from' can be assigned to the type 'to'.
     */
    checkIsAssignableNode(to, from, scope, doThrow = true, templateParams = null) {
        if (from.isUnifyableLiteral()) {
            return this.unifyLiterals(to, from, scope, from.loc, doThrow, templateParams);
        }
        // "&local" can be a reference or even strong pointer
        if (from.op == "unary&" && from.rhs.op == "id" && (helper.isStrong(to) || helper.isReference(to))) {
            let element = scope.resolveElement(from.rhs.value);
            if (!element) {
                throw new errors_1.ImplementationError();
            }
            if (element instanceof scope_1.Variable) {
                let ptr = new types_1.PointerType(element.type, helper.isStrong(to) ? "strong" : "reference");
                if (!this.checkIsAssignableType(to, ptr, from.loc, "assign", doThrow, null, null, templateParams)) {
                    return false;
                }
                from.type = ptr;
                element.isReferenced = true;
                element.isReferencedWithRefcounting = true;
                return true;
            }
            throw new errors_1.TypeError("The & operator can produce a strong pointer or reference on local and global variables only", from.loc);
        }
        // For a stack-based array "local", the expression "local[:]" can be a reference or even strong slice pointer
        if (from.op == ":" && from.lhs.op == "id" && (helper.isStrong(to) || helper.isReference(to))) {
            let element = scope.resolveElement(from.lhs.value);
            if (!element) {
                throw new errors_1.ImplementationError();
            }
            if (element instanceof scope_1.Variable && helper.isArray(element.type)) {
                let ptr = new types_1.SliceType(element.type, helper.isStrong(to) ? "strong" : "reference");
                if (!this.checkIsAssignableType(to, ptr, from.loc, "assign", doThrow, null, null, templateParams)) {
                    return false;
                }
                from.type = ptr;
                element.isReferenced = true;
                element.isReferencedWithRefcounting = true;
                return true;
            }
        }
        return this.checkIsAssignableType(to, from.type, from.loc, "assign", doThrow, null, null, templateParams);
    }
    /**
     * Checks whether the type 'from' can be assigned to the type 'to'.
     * TODO: Remove unbox
     */
    checkIsAssignableType(to, from, loc, mode, doThrow = true, toRestrictions = null, fromRestrictions = null, templateParams = null) {
        if (toRestrictions == null) {
            toRestrictions = { isConst: false };
        }
        if (fromRestrictions == null) {
            fromRestrictions = { isConst: false };
        }
        // Determine const
        if (to instanceof types_1.RestrictedType) {
            toRestrictions = group_1.combineRestrictions(toRestrictions, to);
            to = to.elementType;
        }
        if (from instanceof types_1.RestrictedType) {
            fromRestrictions = group_1.combineRestrictions(fromRestrictions, from);
            from = from.elementType;
        }
        // A const-mismatch can be tolerated if the value is a pure value and if it is being copied.
        if (!toRestrictions.isConst && !!fromRestrictions.isConst && (mode == "pointer" || mode == "equal" || !helper.isPureValue(to))) {
            if (doThrow) {
                throw new errors_1.TypeError("Mismatch of const restriction on variables", loc);
            }
            return false;
        }
        if (templateParams && to instanceof types_1.GenericParameter) {
            if (templateParams.has(to.name)) {
                to = templateParams.get(to.name);
            }
            else {
                templateParams.set(to.name, from);
                return true;
            }
        }
        if (mode == "pointer" || mode == "compare") {
            if (from instanceof types_1.StructType && to != from && from.doesExtend(to)) {
                return true;
            }
        }
        if (to == from) {
            return true;
        }
        else if (to instanceof types_1.TupleType && from instanceof types_1.TupleType) {
            let ok = (to.types.length == from.types.length);
            for (let i = 0; i < to.types.length; i++) {
                if (!this.checkIsAssignableType(to.types[i], from.types[i], loc, "equal", false, toRestrictions, fromRestrictions, templateParams)) {
                    ok = false;
                    break;
                }
            }
            if (ok) {
                return true;
            }
        }
        else if (to instanceof types_1.OrType) {
            if (from instanceof types_1.OrType) {
                if ((mode == "assign" && from.types.length <= to.types.length) || from.types.length == to.types.length) {
                    let ok = true;
                    for (let f of from.types) {
                        let ok2 = false;
                        for (let t of to.types) {
                            if (this.checkTypeEquality(t, f, loc, false)) {
                                ok2 = true;
                                break;
                            }
                        }
                        if (!ok2) {
                            ok = false;
                            break;
                        }
                    }
                    if (ok) {
                        return true;
                    }
                }
            }
            else if (mode == "assign") {
                for (let o of to.types) {
                    if (this.checkIsAssignableType(o, from, loc, mode, false, toRestrictions, fromRestrictions, templateParams)) {
                        return true;
                    }
                }
            }
            // TODO: Else check for exact type equality
            /*        } else if (to instanceof PointerType && from == Static.t_null) {
                        // null can be assigned to any pointer type
                        if (mode == "assign" || mode == "compare") {
                            return true;
                        }
                    } else if (to == Static.t_null && from instanceof PointerType) {
                        if (mode == "compare") {
                            return true;
                        } */
        }
        else if (to instanceof types_1.PointerType && from instanceof types_1.UnsafePointerType) {
            if ((mode == "assign" || mode == "compare") && this.checkIsAssignableType(to.elementType, from.elementType, loc, "pointer", false, toRestrictions, fromRestrictions, templateParams)) {
                return true;
            }
        }
        else if (to instanceof types_1.PointerType && from instanceof types_1.PointerType) {
            if (mode == "compare" || to.mode == from.mode || (mode == "assign" &&
                (to.mode == "local_reference" ||
                    (to.mode == "reference" && (from.mode == "strong" || from.mode == "unique")) ||
                    (to.mode == "strong" && from.mode == "unique") ||
                    (to.mode == "unique" && from.mode == "strong")))) {
                if (this.checkIsAssignableType(to.elementType, from.elementType, loc, mode == "assign" ? "pointer" : "equal", false, toRestrictions, fromRestrictions, templateParams)) {
                    return true;
                }
            }
        }
        else if (to instanceof types_1.UnsafePointerType && (from == Static_1.Static.t_int || from == Static_1.Static.t_uint || from == Static_1.Static.t_null)) {
            // integers and null can be assigned to an usafe pointer type
            if (mode == "assign" || mode == "compare") {
                return true;
            }
        }
        else if ((to == Static_1.Static.t_int || to == Static_1.Static.t_uint || to == Static_1.Static.t_null) && from instanceof types_1.UnsafePointerType) {
            // integers and null can be assigned to an usafe pointer type
            if (mode == "compare") {
                return true;
            }
        }
        else if (to instanceof types_1.UnsafePointerType && (from instanceof types_1.UnsafePointerType || from instanceof types_1.PointerType)) {
            if (to.elementType == Static_1.Static.t_void) {
                // Safe and unsafe pointers to anything can be assigned to #void
                if (mode == "assign" || mode == "compare") {
                    return true;
                }
            }
            if (from.elementType == Static_1.Static.t_void) {
                // #void can be assigned to any unsafe pointer
                if (mode == "assign" || mode == "compare") {
                    return true;
                }
            }
            if (this.checkIsAssignableType(to.elementType, from.elementType, loc, (mode == "assign" || mode == "compare") ? "pointer" : "equal", false, toRestrictions, fromRestrictions, templateParams)) {
                return true;
            }
        }
        else if (to instanceof types_1.PointerType && from instanceof types_1.UnsafePointerType) {
            if (mode == "compare" && this.checkIsAssignableType(to.elementType, from.elementType, loc, "pointer", false, toRestrictions, fromRestrictions, templateParams)) {
                return true;
            }
        }
        else if (to instanceof types_1.ArrayType && from instanceof types_1.ArrayType) {
            if ((to.size == from.size || to.size == -1 || from.size == -1) && this.checkIsAssignableType(to.elementType, from.elementType, loc, "equal", false, toRestrictions, fromRestrictions, templateParams)) {
                return true;
            }
        }
        else if (to instanceof types_1.SliceType && from instanceof types_1.SliceType) {
            if (mode == "compare" || to.mode == from.mode || (mode == "assign" &&
                (to.mode == "local_reference" ||
                    (to.mode == "reference" && (from.mode == "strong" || from.mode == "unique")) ||
                    (to.mode == "strong" && from.mode == "unique") ||
                    (to.mode == "unique" && from.mode == "strong")))) {
                if (this.checkIsAssignableType(to.arrayType, from.arrayType, loc, "equal", false, toRestrictions, fromRestrictions, templateParams)) {
                    return true;
                }
            }
            /*        } else if (to instanceof SliceType && from == Static.t_null) {
                        // null can be assigned to any pointer type
                        if (mode == "assign" || mode == "compare") {
                            return true;
                        }
                    } else if (to == Static.t_null && from instanceof SliceType) {
                        // null can be assigned to any pointer type
                        if (mode == "compare") {
                            return true;
                        } */
        }
        else if (to instanceof types_1.MapType && from instanceof types_1.MapType) {
            if (this.checkIsAssignableType(to.keyType, from.keyType, loc, "equal", false, toRestrictions, fromRestrictions, templateParams) &&
                this.checkIsAssignableType(to.valueType, from.valueType, loc, "equal", false, toRestrictions, fromRestrictions, templateParams)) {
                return true;
            }
        }
        else if (to == Static_1.Static.t_any) {
            // Everything can be asssigned to the empty interface
            if (mode == "assign" || mode == "compare") {
                return true;
            }
        }
        else if (from == Static_1.Static.t_any) {
            // Everything can be asssigned to the empty interface
            if (mode == "compare") {
                return true;
            }
        }
        else if (to instanceof types_1.InterfaceType && mode == "pointer") {
            if (from instanceof types_1.InterfaceType) {
                // Check two interfaces
                if (from == to || from.hasBaseType(to)) {
                    return true;
                }
            }
            else if (from instanceof types_1.StructType) {
                let toMethods = to.getAllMethods();
                let fromMethods = from.getAllMethodsAndFields();
                let ok = true;
                for (let entry of toMethods.entries()) {
                    if (fromMethods.has(entry[0])) {
                        let fieldOrMethod = fromMethods.get(entry[0]);
                        if (!(fieldOrMethod instanceof types_1.FunctionType) || !this.checkFunctionEquality(entry[1], fieldOrMethod, loc, true, false)) {
                            ok = false;
                            if (doThrow) {
                                throw new errors_1.TypeError("Incompatible method signature for " + entry[0] + " in types " + from.toString() + " and " + to.toString(), loc);
                            }
                            break;
                        }
                    }
                    else {
                        ok = false;
                        if (doThrow) {
                            throw new errors_1.TypeError("Type " + from.toString() + " is missing method " + entry[0] + " as required by " + to.toString(), loc);
                        }
                        break;
                    }
                }
                if (ok) {
                    return true;
                }
            }
        }
        if (!doThrow) {
            return false;
        }
        throw new errors_1.TypeError("Type " + from.toString() + " cannot be " + (mode == "compare" ? "compared to type " : "assigned to type ") + to.toString(), loc);
    }
    /**
     * Type checks all parameters.
     */
    checkFunctionArguments(ft, args, scope, loc, doThrow = true) {
        if (args) {
            if (ft.parameters.length != args.length) {
                if (ft.requiredParameterCount() > args.length || (args.length > ft.parameters.length && !ft.hasEllipsis())) {
                    if (doThrow) {
                        throw new errors_1.TypeError("Supplied parameter count does not match function signature " + ft.toString(), loc);
                    }
                    return false;
                }
            }
            for (let i = 0; i < args.length; i++) {
                let pnode = args[i];
                if (pnode.op == "unary...") {
                    if (!ft.hasEllipsis()) {
                        if (doThrow) {
                            throw new errors_1.TypeError("Ellipsis not allowed here. Function is not variadic", pnode.loc);
                        }
                        return false;
                    }
                    if (i != ft.parameters.length - 1 || i != args.length - 1) {
                        if (doThrow) {
                            throw new errors_1.TypeError("Ellipsis must only appear with the last parameter", pnode.loc);
                        }
                        return false;
                    }
                    if (!this.checkIsAssignableNode(ft.lastParameter().type, pnode.rhs, scope, doThrow)) {
                        return false;
                    }
                }
                else {
                    if (ft.hasEllipsis() && i >= ft.parameters.length - 1) {
                        if (!this.checkIsAssignableNode(ft.lastParameter().type.getElementType(), pnode, scope, doThrow)) {
                            return false;
                        }
                    }
                    else {
                        if (!this.checkIsAssignableNode(ft.parameters[i].type, pnode, scope, doThrow)) {
                            return false;
                        }
                    }
                }
            }
        }
        else if (ft.parameters.length != 0 && (!ft.hasEllipsis || ft.parameters.length > 1)) {
            if (doThrow) {
                throw new errors_1.TypeError("Supplied parameters do not match function signature " + ft.toString(), loc);
            }
            return false;
        }
        return true;
    }
    checkTemplateFunctionArguments(t, args, scope, loc) {
        if (t.node.parameters.length == 0) {
            throw new errors_1.ImplementationError();
        }
        let s = new scope_1.Scope(scope);
        let result = new Map();
        for (let n of t.templateParameterNames) {
            let g = new types_1.GenericParameter();
            g.name = n;
            s.registerType(n, g);
        }
        let requiredParameterCount = t.node.parameters.length;
        let ellipsis = false; // TODO
        if (ellipsis) {
            requiredParameterCount--;
        }
        let lastParameter = t.node.parameters[t.node.parameters.length - 1];
        if (t.node.parameters.length != args.length) {
            if (requiredParameterCount > args.length || (args.length > t.node.parameters.length && !ellipsis)) {
                throw new errors_1.TypeError("Supplied parameter count does not match function signature of " + t.toString(), loc);
            }
        }
        for (let i = 0; i < args.length; i++) {
            let pnode = args[i];
            if (pnode.op == "unary...") {
                if (!ellipsis) {
                    throw new errors_1.TypeError("Ellipsis not allowed here. Function is not variadic", pnode.loc);
                }
                if (i != t.node.parameters.length - 1 || i != args.length - 1) {
                    throw new errors_1.TypeError("Ellipsis must only appear with the last parameter", pnode.loc);
                }
                this.checkIsAssignableNode(this.createType(lastParameter, s, "parameter"), pnode.rhs, scope, true, result);
            }
            else {
                if (ellipsis && i >= t.node.parameters.length - 1) {
                    this.checkIsAssignableNode(this.createType(lastParameter, s, "parameter").getElementType(), pnode, scope, true, result);
                }
                else {
                    this.checkIsAssignableNode(this.createType(t.node.parameters[i], s, "parameter"), pnode, scope, true, result);
                }
            }
        }
        for (let n of t.templateParameterNames) {
            if (!result.has(n)) {
                result.set(n, Static_1.Static.t_void);
            }
        }
        return result;
    }
    checkIsEnumerable(node) {
        let t = helper.stripType(node.type);
        if (helper.isMap(t)) {
            return [this.mapKeyType(t), this.mapValueType(t)];
        }
        else if (t instanceof types_1.ArrayType) {
            return [Static_1.Static.t_int, t.elementType];
        }
        else if (t instanceof types_1.SliceType) {
            return [Static_1.Static.t_int, t.getElementType()];
        }
        else if (t == Static_1.Static.t_string) {
            return [Static_1.Static.t_int, Static_1.Static.t_byte];
        }
        throw new errors_1.TypeError("The type " + t.toString() + " is not enumerable", node.loc);
    }
    checkIsIndexable(node, index, indexCanBeLength = false) {
        let t = helper.stripType(node.type);
        if (t instanceof types_1.ArrayType) {
            if (index < 0 || (!indexCanBeLength && index >= t.size) || (indexCanBeLength && index > t.size)) {
                throw new errors_1.TypeError("Index out of range", node.loc);
            }
            return t.elementType;
        }
        else if (t instanceof types_1.SliceType) {
            if (index < 0) {
                throw new errors_1.TypeError("Index out of range", node.loc);
            }
            return t.getElementType();
        }
        else if (t instanceof types_1.TupleType) {
            if (index < 0 || index >= t.types.length) {
                throw new errors_1.TypeError("The index " + index + " does not exist in the tuple " + t.name, node.loc);
            }
            return t.types[index];
        }
        else if (t instanceof types_1.UnsafePointerType || t instanceof types_1.PointerType) {
            return t.elementType;
        }
        else if (t == Static_1.Static.t_string) {
            return Static_1.Static.t_byte;
        }
        throw new errors_1.TypeError("The type " + t.toString() + " is not indexable", node.loc);
    }
    checkIsAddressable(node, scope, withAmpersand, doThrow = true) {
        switch (node.op) {
            case "id":
                let element = scope.resolveElement(node.value);
                if (element instanceof scope_1.Variable) {
                    element.isReferenced = true;
                    return true;
                }
                else if (element instanceof scope_1.FunctionParameter) {
                    element.isReferenced = true;
                    return true;
                }
                break;
            case ".":
                {
                    let t = types_1.RestrictedType.strip(node.lhs.type);
                    if (t instanceof types_1.PointerType || t instanceof types_1.UnsafePointerType) {
                        return true;
                    }
                    return this.checkIsAddressable(node.lhs, scope, false, doThrow);
                }
            case "[":
                let t = types_1.RestrictedType.strip(node.lhs.type);
                if (t instanceof types_1.SliceType || t instanceof types_1.UnsafePointerType) {
                    return true;
                }
                if (t instanceof types_1.ArrayType || t instanceof types_1.TupleType) {
                    return this.checkIsAddressable(node.lhs, scope, false, doThrow);
                }
                break;
            case "bool":
            case "int":
            case "float":
            case "str":
            case "array":
            case "tuple":
            case "object":
                // "&{x:1, y:2}" is allowed whereas "&({x:1, y:2}.x)"" is not allowed.
                if (withAmpersand) {
                    return true;
                }
        }
        if (doThrow) {
            throw new errors_1.TypeError("Cannot take address of intermediate value", node.loc);
        }
        return false;
    }
    checkIsPointer(node, doThrow = true) {
        let t = helper.stripType(node.type);
        if (t instanceof types_1.PointerType || t instanceof types_1.UnsafePointerType) {
            return true;
        }
        if (doThrow) {
            throw new errors_1.TypeError("Expected a pointer, but got " + node.type.toString(), node.loc);
        }
        return false;
    }
    checkIsString(node, doThrow = true) {
        if (helper.isString(node.type)) {
            return true;
        }
        if (doThrow) {
            throw new errors_1.TypeError("Expected a string, but got " + node.type.toString(), node.loc);
        }
        return false;
    }
    checkIsSignedNumber(node, doThrow = true) {
        let t = helper.stripType(node.type);
        if (t == Static_1.Static.t_float || t == Static_1.Static.t_double || t == Static_1.Static.t_int || t == Static_1.Static.t_char || t == Static_1.Static.t_int8 || t == Static_1.Static.t_int16 || t == Static_1.Static.t_int32 || t == Static_1.Static.t_int64) {
            return true;
        }
        if (doThrow) {
            throw new errors_1.TypeError("Expected a signed numeric type, but got " + node.type.toString(), node.loc);
        }
        return false;
    }
    checkIsUnsignedNumber(node, doThrow = true) {
        let t = helper.stripType(node.type);
        if (t == Static_1.Static.t_uint || t == Static_1.Static.t_byte || t == Static_1.Static.t_uint8 || t == Static_1.Static.t_uint16 || t == Static_1.Static.t_uint32 || t == Static_1.Static.t_uint64) {
            return true;
        }
        if (doThrow) {
            throw new errors_1.TypeError("Expected an unsigned numeric type, but got " + node.type.toString(), node.loc);
        }
        return false;
    }
    checkIsBool(node, doThrow = true) {
        let t = helper.stripType(node.type);
        if (t == Static_1.Static.t_bool) {
            return true;
        }
        if (doThrow) {
            throw new errors_1.TypeError("Expected a boolean type, but got " + node.type.toString(), node.loc);
        }
        return false;
    }
    checkIsNumber(node, doThrow = true) {
        if (helper.isNumber(node.type)) {
            return true;
        }
        if (doThrow) {
            throw new errors_1.TypeError("Expected a numeric type, but got " + node.type.toString(), node.loc);
        }
        return false;
    }
    checkIsIntNumber(node, doThrow = true) {
        if (helper.isIntNumber(node.type)) {
            return true;
        }
        if (doThrow) {
            throw new errors_1.TypeError("Expected an integer type, but got " + node.type.toString(), node.loc);
        }
        return false;
    }
    checkIsPlatformIntNumber(node, doThrow = true) {
        if (helper.isPlatformIntNumber(node.type)) {
            return true;
        }
        if (doThrow) {
            throw new errors_1.TypeError("Expected type int, but got " + node.type.toString(), node.loc);
        }
        return false;
    }
    checkIsIntNumberOrUnsafePointer(node, doThrow = true) {
        let t = helper.stripType(node.type);
        if (helper.isIntNumber(t)) {
            return true;
        }
        if (t instanceof types_1.UnsafePointerType) {
            return true;
        }
        if (doThrow) {
            throw new errors_1.TypeError("Expected a numeric or pointer type, but got " + node.type.toString(), node.loc);
        }
        return false;
    }
    checkIsInterface(node, doThrow = true) {
        if (helper.isInterface(node.type)) {
            return true;
        }
        if (!doThrow) {
            return false;
        }
        throw new errors_1.TypeError("Expected an interface type", node.loc);
    }
    checkFunctionEquality(a, b, loc, allowMoreRestrictions, doThrow = true) {
        if (a.parameters.length == b.parameters.length) {
            // Check the return type. And both functions are either both member functions or both non-member functions.
            if (this.checkTypeEquality(a.returnType, b.returnType, loc, false) && !!a.objectType == !!b.objectType) {
                let ok = true;
                for (let i = 0; i < a.parameters.length; i++) {
                    if (!this.checkTypeEquality(a.parameters[i].type, b.parameters[i].type, loc, false)) {
                        ok = false;
                        break;
                    }
                    if (a.parameters[i].ellipsis != b.parameters[i].ellipsis) {
                        ok = false;
                        break;
                    }
                }
                if (a.objectType) {
                    if (allowMoreRestrictions && (helper.isConst(a.objectType) && !helper.isConst(b.objectType))) {
                        ok = false;
                    }
                    else if (!allowMoreRestrictions && helper.isConst(a.objectType) != helper.isConst(b.objectType)) {
                        ok = false;
                    }
                }
                if (a.callingConvention != b.callingConvention) {
                    ok = false;
                }
                if (ok) {
                    return true;
                }
            }
        }
        if (doThrow) {
            throw new errors_1.TypeError("Type mismatch between " + a.toString() + " and " + b.toString(), loc);
        }
        return false;
    }
    checkTypeEquality(a, b, loc, doThrow = true) {
        if (a == b) {
            return true;
        }
        if (a instanceof types_1.RestrictedType && b instanceof types_1.RestrictedType) {
            if (a.isConst == b.isConst) {
                if (this.checkTypeEquality(a.elementType, b.elementType, loc, false)) {
                    return true;
                }
            }
        }
        else if (a instanceof types_1.PointerType && b instanceof types_1.PointerType) {
            if (this.checkTypeEquality(a.elementType, b.elementType, loc, false) && a.mode == b.mode) {
                return true;
            }
        }
        else if (a instanceof types_1.UnsafePointerType && b instanceof types_1.UnsafePointerType) {
            if (this.checkTypeEquality(a.elementType, b.elementType, loc, false)) {
                return true;
            }
        }
        else if (a instanceof types_1.SliceType && b instanceof types_1.SliceType && a.mode == b.mode) {
            if (this.checkTypeEquality(a.getElementType(), b.getElementType(), loc, false)) {
                return true;
            }
        }
        else if (a instanceof types_1.ArrayType && b instanceof types_1.ArrayType) {
            if (a.size == b.size && this.checkTypeEquality(a.elementType, b.elementType, loc, false)) {
                return true;
            }
        }
        else if (a instanceof types_1.TupleType && b instanceof types_1.TupleType) {
            if (a.types.length == b.types.length) {
                let ok = true;
                for (let i = 0; ok && i < a.types.length; i++) {
                    ok = ok && this.checkTypeEquality(a.types[i], b.types[i], loc, false);
                }
                if (ok) {
                    return true;
                }
            }
        }
        else if (a instanceof types_1.TemplateStructType && b instanceof types_1.TemplateStructType) {
            if (a.base == b.base) {
                let ok = true;
                for (let i = 0; ok && i < a.templateParameterTypes.length; i++) {
                    ok = ok && this.checkTypeEquality(a.templateParameterTypes[i], b.templateParameterTypes[i], loc, false);
                }
                if (ok) {
                    return true;
                }
            }
        }
        else if (a instanceof types_1.OrType && b instanceof types_1.OrType) {
            if (a.types.length == b.types.length) {
                let ok = true;
                for (let t of a.types) {
                    let ok2 = false;
                    for (let t2 of b.types) {
                        let eq = this.checkTypeEquality(t, t2, loc, false);
                        if (eq) {
                            ok2 = true;
                            break;
                        }
                    }
                    if (!ok2) {
                        ok = false;
                        break;
                    }
                }
                if (ok) {
                    return true;
                }
            }
        }
        else if (a instanceof types_1.FunctionType && b instanceof types_1.FunctionType) {
            if (this.checkFunctionEquality(a, b, loc, false, doThrow)) {
                return true;
            }
        }
        else if (a instanceof types_1.InterfaceType && b instanceof types_1.InterfaceType) {
            let m1 = a.getAllMethods();
            let m2 = b.getAllMethods();
            if (m1.size == m2.size) {
                let ok = true;
                for (let entry of m1.entries()) {
                    if (!m2.has(entry[0]) || !this.checkTypeEquality(m2.get(entry[0]), entry[1], loc, false)) {
                        ok = false;
                        break;
                    }
                }
                if (ok) {
                    return true;
                }
            }
        }
        if (doThrow) {
            throw new errors_1.TypeError("Type mismatch between " + a.toString() + " and " + b.toString(), loc);
        }
        return false;
    }
    // TODO: Move to helper.ts
    /**
     * Returns true if the expression given by `node` can be the target of an assignment.
     */
    checkIsAssignable(node, scope, doThrow = true) {
        if (!helper.isAssignable(node, scope)) {
            if (!doThrow) {
                return false;
            }
            throw new errors_1.TypeError("The expression is not assignable, because it is an intermediate value, the underlying variable is not mutable, or the type is const", node.loc);
        }
        return true;
    }
    checkIsMutable(node, scope, doThrow = true) {
        if (!helper.isMutable(node, scope)) {
            if (!doThrow) {
                return false;
            }
            throw new errors_1.TypeError("The expression is not mutable, because the variable is not mutable or the type is const", node.loc);
        }
        return true;
    }
    checkVariableType(t, loc) {
        if (types_1.RestrictedType.strip(t) instanceof types_1.InterfaceType) {
            throw new errors_1.TypeError("Interface types must be used together with a pointer", loc);
        }
        if (types_1.RestrictedType.strip(t) instanceof types_1.MapType) {
            throw new errors_1.TypeError("Map types must be used together with a pointer", loc);
        }
    }
    // TODO: Move to helper.ts
    mapKeyType(t) {
        t = helper.stripType(t);
        if (!(t instanceof types_1.PointerType)) {
            throw new errors_1.ImplementationError("Internal error");
        }
        t = helper.stripType(t.elementType);
        if (!(t instanceof types_1.MapType)) {
            throw new errors_1.ImplementationError("Internal error");
        }
        return t.keyType;
    }
    // TODO: Move to helper.ts
    mapValueType(t) {
        t = helper.stripType(t);
        if (!(t instanceof types_1.PointerType)) {
            throw new errors_1.ImplementationError("Internal error");
        }
        t = helper.stripType(t.elementType);
        if (!(t instanceof types_1.MapType)) {
            throw new errors_1.ImplementationError("Internal error");
        }
        return t.valueType;
    }
    orTypeIndex(o, t, exactMatch) {
        let index = 0;
        for (let ot of o.types) {
            if (this.checkIsAssignableType(ot, t, null, exactMatch ? "compare" : "assign", false)) {
                return index;
            }
            index++;
        }
        throw new Error("Internal error");
    }
    static hasStrongOrUniquePointers(t) {
        t = types_1.RestrictedType.strip(t);
        if ((t instanceof types_1.PointerType || t instanceof types_1.SliceType) && (t.mode == "strong" || t.mode == "unique")) {
            return true;
        }
        if (t instanceof types_1.TupleType) {
            for (let p of t.types) {
                if (this.hasStrongOrUniquePointers(p)) {
                    return true;
                }
            }
            return false;
        }
        else if (t instanceof types_1.ArrayType) {
            return this.hasStrongOrUniquePointers(t.elementType);
        }
        else if (t instanceof types_1.StructType) {
            for (let f of t.fields) {
                if (this.hasStrongOrUniquePointers(f.type)) {
                    return true;
                }
            }
            if (t.extends && this.hasStrongOrUniquePointers(t.extends)) {
                return true;
            }
            return false;
        }
        return false;
    }
    static hasReferenceOrStrongPointers(t) {
        t = types_1.RestrictedType.strip(t);
        if ((t instanceof types_1.PointerType || t instanceof types_1.SliceType) && (t.mode == "strong" || t.mode == "reference")) {
            return true;
        }
        if (t instanceof types_1.TupleType) {
            for (let p of t.types) {
                if (this.hasReferenceOrStrongPointers(p)) {
                    return true;
                }
            }
            return false;
        }
        else if (t instanceof types_1.ArrayType) {
            return this.hasReferenceOrStrongPointers(t.elementType);
        }
        else if (t instanceof types_1.StructType) {
            for (let f of t.fields) {
                if (this.hasStrongOrUniquePointers(f.type)) {
                    return true;
                }
            }
            if (t.extends && this.hasReferenceOrStrongPointers(t.extends)) {
                return true;
            }
            return false;
        }
        return false;
    }
    static hasLocalReference(t) {
        t = types_1.RestrictedType.strip(t);
        if ((t instanceof types_1.PointerType || t instanceof types_1.SliceType) && (t.mode == "local_reference")) {
            return true;
        }
        if (t instanceof types_1.TupleType) {
            for (let p of t.types) {
                if (this.hasLocalReference(p)) {
                    return true;
                }
            }
            return false;
        }
        else if (t instanceof types_1.ArrayType) {
            return this.hasLocalReference(t.elementType);
        }
        else if (t instanceof types_1.StructType) {
            for (let f of t.fields) {
                if (this.hasLocalReference(f.type)) {
                    return true;
                }
            }
            if (t.extends && this.hasLocalReference(t.extends)) {
                return true;
            }
            return false;
        }
        return false;
    }
    pointerElementType(t) {
        t = types_1.RestrictedType.strip(t);
        if (t instanceof types_1.PointerType || t instanceof types_1.UnsafePointerType) {
            return t.elementType;
        }
        return null;
    }
    sliceArrayType(t) {
        t = types_1.RestrictedType.strip(t);
        if (t instanceof types_1.SliceType) {
            return t.arrayType;
        }
        return null;
    }
    arrayElementType(t) {
        t = types_1.RestrictedType.strip(t);
        if (t instanceof types_1.ArrayType) {
            return t.elementType;
        }
        return null;
    }
    stringLiteralType(name) {
        if (this.stringLiteralTypes.has(name)) {
            return this.stringLiteralTypes.get(name);
        }
        let t = new types_1.StringLiteralType(name);
        this.stringLiteralTypes.set(name, t);
        return t;
    }
    checkGroupsInFunction(f) {
        if (!f.node.statements) {
            return;
        }
        let groups = f.type.createGroups();
        for (let pt of f.type.parameters) {
            let g = groups.get(pt.name);
            if (!g) {
                throw new errors_1.ImplementationError();
            }
            f.scope.setGroup(pt, g);
        }
        if (f.type.objectType) {
            let th = f.scope.resolveElement("this");
            if (!th) {
                throw new errors_1.ImplementationError();
            }
            let g = groups.get("this");
            if (!g) {
                throw new errors_1.ImplementationError();
            }
            f.scope.setGroup(th, g);
        }
        if (f.namedReturnVariables) {
            for (let i = 0; i < f.namedReturnVariables.length; i++) {
                let r = f.namedReturnVariables[i];
                let g = groups.get("return " + i.toString());
                if (!g) {
                    throw new errors_1.ImplementationError();
                }
                f.scope.setGroup(r, g);
            }
        }
        else if (f.unnamedReturnVariable) {
            let g = groups.get("return");
            if (!g) {
                throw new errors_1.ImplementationError();
            }
            f.scope.setGroup(f.unnamedReturnVariable, g);
        }
        for (let snode of f.node.statements) {
            this.checkGroupsInStatement(snode, f.scope);
        }
    }
    checkGroupsInStatement(snode, scope) {
        this.modifiedVariabes.clear();
        this.usedVariables.clear();
        switch (snode.op) {
            case "comment":
            case "yield":
                break;
            case "yield_continue":
                break;
            case "let_in": {
                let flags = TypeChecker.hasReferenceOrStrongPointers(snode.rhs.type) ? group_1.GroupCheckFlags.ForbidIsolates : group_1.GroupCheckFlags.AllowIsolates;
                let g = this.checkGroupsInExpression(snode.rhs, scope, flags);
                if (snode.lhs.op == "tuple") {
                    if (snode.lhs.parameters[0].value != "_") {
                        let v = scope.resolveElement(snode.lhs.parameters[0].value);
                        scope.setGroup(v, new group_1.Group(group_1.GroupKind.Free));
                    }
                    if (snode.lhs.parameters[1].value != "_") {
                        let v = scope.resolveElement(snode.lhs.parameters[1].value);
                        scope.setGroup(v, g ? g : new group_1.Group(group_1.GroupKind.Free));
                    }
                }
                else {
                    let v = scope.resolveElement(snode.lhs.value);
                    scope.setGroup(v, g ? g : new group_1.Group(group_1.GroupKind.Free));
                }
                break;
            }
            case "var":
            case "let":
                if (snode.rhs) {
                    this.checkGroupsInAssignment(snode, scope);
                }
                break;
            case "+=":
            case "*=":
            case "/=":
            case "-=":
            case "<<=":
            case ">>=":
            case "%=":
            case "&=":
            case "&^=":
            case "|=":
            case "^=":
            case "=":
                this.checkGroupsInAssignment(snode, scope);
                break;
            case "return": {
                let f = scope.envelopingFunction();
                if (!f) {
                    throw new errors_1.ImplementationError();
                }
                if (snode.lhs) {
                    // Returning a tuple?
                    if (f.namedReturnVariables) {
                        if (snode.lhs.op == "tuple") {
                            for (let i = 0; i < f.namedReturnVariables.length; i++) {
                                let group = scope.resolveGroup(f.namedReturnVariables[i]);
                                if (helper.isUnique(f.namedReturnVariables[i].type)) {
                                    group = null;
                                }
                                this.checkGroupsInSingleAssignment(f.namedReturnVariables[i].type, group, null, snode.lhs.parameters[i], false, scope, snode.loc);
                            }
                        }
                        else {
                            let flags = TypeChecker.hasReferenceOrStrongPointers(snode.lhs.type) ? group_1.GroupCheckFlags.ForbidIsolates : group_1.GroupCheckFlags.AllowIsolates;
                            let g = this.checkGroupsInExpression(snode.lhs, scope, flags);
                            for (let i = 0; i < f.namedReturnVariables.length; i++) {
                                let group = scope.resolveGroup(f.namedReturnVariables[i]);
                                if (helper.isUnique(f.namedReturnVariables[i].type)) {
                                    group = null;
                                }
                                this.checkGroupsInSingleAssignment(f.namedReturnVariables[i].type, group, g instanceof group_1.TupleGroup ? g.groups[i] : g, snode.lhs, i + 1 < f.namedReturnVariables.length, scope, snode.loc);
                            }
                        }
                    }
                    else {
                        if (!f.unnamedReturnVariable) {
                            throw new errors_1.ImplementationError();
                        }
                        let group = scope.resolveGroup(f.unnamedReturnVariable);
                        if (helper.isUnique(f.unnamedReturnVariable.type)) {
                            group = null;
                        }
                        this.checkGroupsInSingleAssignment(f.type.returnType, group, null, snode.lhs, false, scope, snode.loc);
                    }
                }
                break;
            }
            case "break":
                break;
            case "continue":
                break;
            case "else":
                snode.scope.resetGroups();
                for (let st of snode.statements) {
                    this.checkGroupsInStatement(st, snode.scope);
                }
                break;
            case "if":
                snode.scope.resetGroups();
                if (snode.lhs) {
                    this.checkGroupsInStatement(snode.lhs, snode.scope);
                    this.modifiedVariabes.clear();
                    this.usedVariables.clear();
                }
                this.checkExpression(snode.condition, snode.scope);
                for (let st of snode.statements) {
                    this.checkGroupsInStatement(st, snode.scope);
                }
                if (snode.elseBranch) {
                    this.checkGroupsInStatement(snode.elseBranch, scope);
                    //                    snode.elseBranch.scope.resetGroups();
                    //                    for(let st of snode.elseBranch.statements) {
                    //                        this.checkGroupsInStatement(st, snode.elseBranch.scope);
                    //                    }
                }
                /*
                if (snode.scopeExit.breaks) {
                    for (let c of snode.scopeExit.breaks) {
                        c.mergeScopes(scope, "reverted_subsequent");
                    }
                }
                if (snode.scopeExit.continues) {
                    for (let c of snode.scopeExit.continues) {
                        c.mergeScopes(scope, "reverted_subsequent");
                    }
                }*/
                if (snode.elseBranch) {
                    //                    snode.elseBranch.scope.resetGroups();
                    //                    for(let st of snode.elseBranch.statements) {
                    //                        this.checkGroupsInStatement(st, snode.elseBranch.scope);
                    //                    }
                    /*
                    if (snode.elseBranch.scopeExit.breaks) {
                        for (let c of snode.elseBranch.scopeExit.breaks) {
                            c.mergeScopes(scope, "reverted_subsequent");
                        }
                    }
                    if (snode.elseBranch.scopeExit.continues) {
                        for (let c of snode.elseBranch.scopeExit.continues) {
                            c.mergeScopes(scope, "reverted_subsequent");
                        }
                    }
                    */
                    if (snode.scopeExit.fallthrough && snode.elseBranch.scopeExit.fallthrough) {
                        snode.scope.mergeScopes(snode.elseBranch.scope, "conditional");
                        scope.mergeScopes(snode.scope, "subsequent");
                    }
                    else if (snode.scopeExit.fallthrough) {
                        scope.mergeScopes(snode.scope, "conditional");
                    }
                    else if (snode.elseBranch.scopeExit.fallthrough) {
                        scope.mergeScopes(snode.elseBranch.scope, "conditional");
                    }
                }
                else if (snode.scopeExit.fallthrough) {
                    scope.mergeScopes(snode.scope, "conditional");
                }
                break;
            case "for":
                {
                    // This line is required, because the check might run twice when inside a for loop.
                    snode.scope.resetGroups();
                    if (snode.condition) {
                        if (snode.condition.op == ";;") {
                            if (snode.condition.lhs) {
                                this.checkGroupsInStatement(snode.condition.lhs, snode.condition.scope);
                                this.modifiedVariabes.clear();
                                this.usedVariables.clear();
                            }
                            if (snode.condition.condition) {
                                this.checkGroupsInExpression(snode.condition.condition, snode.condition.scope, group_1.GroupCheckFlags.None);
                            }
                            if (snode.condition.rhs) {
                                this.checkGroupsInStatement(snode.condition.rhs, snode.condition.scope);
                            }
                        }
                        else {
                            this.checkGroupsInStatement(snode.condition, snode.condition.scope);
                        }
                    }
                    for (let st of snode.statements) {
                        this.checkGroupsInStatement(st, snode.scope);
                    }
                    let s = snode.scope;
                    // Join in all "continues"
                    if (snode.scopeExit.continues) {
                        for (let c of snode.scopeExit.continues) {
                            snode.scope.mergeScopes(c, "conditional");
                        }
                    }
                    // Check groups for a second run.
                    for (let st of snode.statements) {
                        this.checkGroupsInStatement(st, s);
                    }
                    // Join in all "breaks"
                    if (snode.scopeExit.breaks) {
                        for (let c of snode.scopeExit.breaks) {
                            scope.mergeScopes(c, "conditional");
                        }
                    }
                    // Join in the for scope
                    if (snode.scopeExit.fallthrough) {
                        scope.mergeScopes(snode.scope, "conditional");
                    }
                    break;
                }
            case "copy":
            case "move":
                this.checkGroupsInExpression(snode.lhs, scope, group_1.GroupCheckFlags.None);
                this.checkGroupsInExpression(snode.rhs, scope, group_1.GroupCheckFlags.None);
                break;
            case "slice":
                this.checkGroupsInExpression(snode.parameters[0], scope, group_1.GroupCheckFlags.None);
                this.checkGroupsInExpression(snode.parameters[1], scope, group_1.GroupCheckFlags.None);
                this.checkGroupsInExpression(snode.parameters[2], scope, group_1.GroupCheckFlags.None);
                break;
            /*
        case "spawn":
        {
            this.checkExpression(snode.rhs, scope);
            if (snode.rhs.op != "(") {
                throw new ImplementationError()
            }
            if (!(snode.rhs.lhs.type instanceof FunctionType)) {
                throw new ImplementationError()
            }
            if ((snode.rhs.lhs.type as FunctionType).returnType != Static.t_void) {
                throw new TypeError("Functions invoked via 'spawn' must return void", snode.loc);
            }
            break;
        }
        */
            case "println":
                for (let i = 0; i < snode.parameters.length; i++) {
                    this.checkGroupsInExpression(snode.parameters[i], scope, group_1.GroupCheckFlags.None);
                }
                break;
            case "push":
            case "append":
            default:
                this.checkGroupsInExpression(snode, scope, group_1.GroupCheckFlags.None);
        }
        //        return null;
    }
    checkGroupsInAssignment(snode, scope) {
        if (snode.lhs.op == "id") {
            this.checkGroupsInSingleAssignment(snode.lhs.type, snode.lhs, null, snode.rhs, false, scope, snode.loc);
        }
        else if (snode.lhs.op == "tuple") {
            let t = helper.stripType(snode.rhs.type);
            if (snode.rhs.op == "tuple") {
                for (let i = 0; i < snode.rhs.parameters.length; i++) {
                    this.checkGroupsInSingleAssignment(snode.lhs.parameters[i].type, snode.lhs.parameters[i], null, snode.rhs.parameters[i], false, scope, snode.loc);
                }
            }
            else {
                let flags = TypeChecker.hasReferenceOrStrongPointers(snode.rhs.type) ? group_1.GroupCheckFlags.ForbidIsolates : group_1.GroupCheckFlags.AllowIsolates;
                let g = this.checkGroupsInExpression(snode.rhs, scope, flags);
                for (let i = 0; i < snode.lhs.parameters.length; i++) {
                    this.checkGroupsInSingleAssignment(snode.lhs.parameters[i].type, snode.lhs.parameters[i], g instanceof group_1.TupleGroup ? g.groups[i] : g, snode.rhs, i + 1 < snode.lhs.parameters.length, scope, snode.loc);
                }
            }
        }
        else if (snode.lhs.op == ".") {
            if (helper.isUnsafePointer(snode.lhs.lhs.type)) {
                return;
            }
            if (helper.isSafePointer(snode.lhs.lhs.type) || helper.isStruct(snode.lhs.lhs.type)) {
                this.checkGroupsInSingleAssignment(snode.lhs.type, snode.lhs, null, snode.rhs, false, scope, snode.loc);
            }
            else {
                throw new errors_1.ImplementationError();
            }
        }
        else if (snode.lhs.op == "[") {
            this.checkGroupsInSingleAssignment(snode.lhs.type, snode.lhs, null, snode.rhs, false, scope, snode.loc);
        }
        else if (snode.lhs.op == "unary*") {
            this.checkGroupsInSingleAssignment(snode.lhs.type, snode.lhs, null, snode.rhs, false, scope, snode.loc);
        }
        else {
            throw new errors_1.ImplementationError();
        }
    }
    checkGroupsInSingleAssignment(ltype, lnode, rightGroup, rnode, rnodeReuse, scope, loc) {
        if (!rnode) {
            throw new errors_1.ImplementationError();
        }
        // Determine the groups of the LHS and the RHS
        if (!rightGroup) {
            // let isPointer = !helper.isPureValue(rnode.type);
            let flags = TypeChecker.hasReferenceOrStrongPointers(rnode.type) ? group_1.GroupCheckFlags.ForbidIsolates : group_1.GroupCheckFlags.AllowIsolates;
            rightGroup = this.checkGroupsInExpression(rnode, scope, flags);
        }
        let leftGroup = lnode instanceof parser_1.Node ? this.checkGroupsInExpression(lnode, scope, group_1.GroupCheckFlags.AllowIsolates | group_1.GroupCheckFlags.AllowUnavailableVariable) : lnode;
        // What are we assigning to?
        let lhsIsVariable = lnode instanceof parser_1.Node ? lnode.op == "id" : false;
        let lhsVariable = null;
        let lhsVariableScope = null;
        if (lhsIsVariable) {
            [lhsVariable, lhsVariableScope] = scope.resolveElementWithScope(lnode.value);
        }
        // Assigning to a value type? -> Nothing to do.
        // Assigning to a string? -> Nothing to do, because strings are not bound to any group.
        // Assigning to an unsafe pointer -> Nothing to do. Programmer hopefully knows what he is doing ...
        if (helper.isPureValue(ltype) || helper.isPureValue(rnode.type) || helper.isString(ltype) || helper.isUnsafePointer(ltype)) {
            // When assigning to a variable, set its group such that it becomes available
            if (lhsVariable) {
                if (lhsVariable instanceof scope_1.Variable && lhsVariable.isReferencedWithRefcounting) {
                    // The address of the variable is taken. Therefore, it is restricted to its scope.
                    // Use the bound-group of its scope.
                    scope.setGroup(lhsVariable, lhsVariableScope.group);
                }
                else {
                    // The variable is free.
                    scope.setGroup(lhsVariable, new group_1.Group(group_1.GroupKind.Free, lhsVariable.name));
                }
            }
            return;
        }
        // What is assigned?
        let rhsVariableName;
        let rhsIsVariable = false;
        if (rnode.op == "id" || (rnode.op == "take" && rnode.lhs.op == "id")) {
            rhsIsVariable = true;
            rhsVariableName = rnode.op == "id" ? rnode.value : rnode.lhs.value;
        }
        else if (rnode.op == ":" && (helper.isUnique(rnode.type) || helper.isStrong(rnode.type))) {
            let r = rnode;
            while (r.op == ":" || r.op == "take") {
                r = r.lhs;
            }
            if (r.op == "id") {
                rhsIsVariable = true;
                rhsVariableName = r.value;
            }
        }
        let rhsIsTakeExpr = helper.isTakeExpression(rnode);
        // The right hand side is an expression that evaluates to an isolate, and therefore the group is null
        if (!rightGroup) {
            // The isolate must be taken, even when assigned to a reference
            if (!rhsIsVariable && !rhsIsTakeExpr) {
                throw new errors_1.TypeError("Assignment of an expression that evaluates to an isolate is only allowed via a variable or take expression", loc);
            }
            if (TypeChecker.hasReferenceOrStrongPointers(rnode.type)) {
                // Accessing a strong pointer or reference inside an isolate is not allowed.
                // This should be guarded by the GroupCheckFlags.ForbidIsolates above. Just being paranoid here.
                throw new errors_1.ImplementationError();
            }
            rightGroup = new group_1.Group(group_1.GroupKind.Free);
        }
        let isArrayVariable = function (name) {
            let variable = scope.resolveElement(name);
            if (!(variable instanceof scope_1.Variable)) {
                return false;
            }
            return helper.isArray(variable.type);
        };
        // Assigning to a strong or unique pointer? Then the RHS must let go of its ownership.
        if (TypeChecker.hasStrongOrUniquePointers(ltype) || (!helper.isSafePointer(ltype) && !helper.isSlice(ltype) && !helper.isPureValue(ltype))) {
            if (rhsIsVariable) {
                // Make the RHS variable unavailable, since the LHS is now the owner and there can be one owner only
                if (!rnodeReuse) {
                    let rhsVariable = scope.resolveElement(rhsVariableName);
                    scope.setGroup(rhsVariable, null);
                }
                rnode.flags |= parser_1.AstFlags.ZeroAfterAssignment;
            }
            else if (rhsIsTakeExpr) {
                // Nothing special todo
            }
            else if (rnode.op == "unary&" && rnode.rhs.op == "id") {
                // Nothing special todo. We are referencing a stack variable and this variable remains accessible.
                // The variable may even have multiple owners. No owner can hold it longer than the stack frame exists
                // and the memory is free'd when the stack frame is removed.
            }
            else if (rnode.op == ":" && rnode.lhs.op == "id" && isArrayVariable(rnode.lhs.value)) {
                // Nothing special todo. We are slicing an array on the stack and this array remains accessible.
                // There may be multiple owning slices pointing to the array data. No owner can hold it longer than the stack frame exists
                // and the memory is free'd when the stack frame is removed.
            }
            else if (rnode.op == "typeCast" && rnode.rhs.op == "id" && helper.isOrType(rnode.rhs.type)) {
                // Taking ownership from an Or-Type? Then the underlying RHS variable is no longer accessible
                if (!rnodeReuse) {
                    let rhsVariable = scope.resolveElement(rhsVariableName);
                    scope.setGroup(rhsVariable, null);
                }
                rnode.rhs.flags |= parser_1.AstFlags.ZeroAfterAssignment;
            }
            else {
                throw new errors_1.TypeError("Assignment to an owning pointer (or data structure containing an owning pointer) is only allowed from a variable or take expression", loc);
            }
        }
        if (helper.isUnique(ltype)) {
            // Check that the RHS group is unbound, because the RHS is not neccessarily an isolate
            if (rightGroup.isBound(scope)) {
                throw new errors_1.TypeError("Assignment of a bound group to an isolate is not allowed", loc);
            }
        }
        // Assigning to a variable for the first time, and the unary& operator has been applied on it to create a reference or strong pointer to a stack variable?
        if (!leftGroup && lhsIsVariable && lhsVariable instanceof scope_1.Variable && lhsVariable.isReferencedWithRefcounting) {
            leftGroup = lhsVariableScope.group;
            scope.setGroup(lhsVariable, leftGroup);
        }
        // The if-clause is true when assigning to a variable that is not global.
        // The purpose of ignoring global is that setGroup should not be executed on a global variable.
        if (lhsIsVariable && (!(lhsVariable instanceof scope_1.Variable) || (!lhsVariable.isGlobal && !lhsVariable.isReferencedWithRefcounting))) {
            // Determine whether the variable being assigned to holds a value that needs destruction before assigning to it.
            if (scope.elementNeedsDestruction.get(lhsVariable)) {
                lnode.flags &= ~parser_1.AstFlags.EmptyOnAssignment;
            }
            else {
                // The variable being assigned to is currently not accessible.
                // Hence, it has no value that needs to be destructed before assignment.
                lnode.flags |= parser_1.AstFlags.EmptyOnAssignment;
            }
            // Set the group of the LHS variable to the RHS group
            scope.setGroup(lhsVariable, rightGroup);
        }
        else {
            // Assigning to an expression of type unique pointer?
            if (!leftGroup) {
                // Check that the RHS group is unbound
                if (rightGroup.isBound(scope)) {
                    throw new errors_1.TypeError("Assignment of a bound group to an isolate is not allowed", loc);
                }
                // Make the RHS group unavailable
                scope.makeGroupUnavailable(rightGroup);
            }
            else {
                // Join RHS and LHS
                scope.joinGroups(leftGroup, rightGroup, loc, true);
            }
        }
    }
    checkGroupsInExpression(enode, scope, flags) {
        let origFlags = flags;
        /*
        if ((flags & GroupCheckFlags.IsolatesMask) == 0) {
            if (this.checkIsPointer(enode, false)) {
                flags |= GroupCheckFlags.ForbidIsolates;
            } else {
                flags |= GroupCheckFlags.AllowIsolates;
            }
        }
        */
        flags = flags & group_1.GroupCheckFlags.AllowUnavailableVariableMask;
        switch (enode.op) {
            case "null":
                return new group_1.Group(group_1.GroupKind.Free);
            case "bool":
                break;
            case "str":
                break;
            case "rune":
                break;
            case "int":
                break;
            case "float":
                break;
            case "id":
                // TODO: ellipsis, optional
                let element = scope.resolveElement(enode.value);
                if (!element) {
                    throw new errors_1.ImplementationError();
                }
                if (element instanceof scope_1.TemplateFunction) {
                    return new group_1.Group(group_1.GroupKind.Free);
                }
                let g = scope.resolveGroup(element);
                if (!g || !scope.isGroupAvailable(g)) {
                    if ((origFlags & group_1.GroupCheckFlags.AllowUnavailableVariable) == 0) {
                        throw new errors_1.TypeError("Variable " + element.name + " is not available in this place", enode.loc);
                    }
                    //                    console.log(element.name + " is not available, but do not care");
                }
                if (element instanceof scope_1.Variable) {
                    if ((enode.flags & parser_1.AstFlags.ZeroAfterAssignment) == parser_1.AstFlags.ZeroAfterAssignment) {
                        if (this.modifiedVariabes.has(element)) {
                            throw new errors_1.TypeError("Variable " + element.name + " is taken more than once in the same statement", enode.loc);
                        }
                        if (this.usedVariables.has(element)) {
                            throw new errors_1.TypeError("Variable " + element.name + " is read and taken in the same statement", enode.loc);
                        }
                        this.modifiedVariabes.add(element);
                    }
                    else {
                        if (this.modifiedVariabes.has(element)) {
                            throw new errors_1.TypeError("Variable " + element.name + " is taken and read in the same statement", enode.loc);
                        }
                        this.usedVariables.add(element);
                    }
                }
                // Accessing a global isolate is like an expression that evaluates to an isolate. Therefore its Group is null
                if (element instanceof scope_1.Variable && element.isGlobal && helper.isUnique(element.type)) {
                    return null;
                }
                //                console.log("Group of " + element.name + " is " + (g ? g.name : null), scope.unavailableGroups);
                return g;
            case "++":
            case "--":
                return this.checkGroupsInExpression(enode.lhs, scope, flags);
            case "unary-":
            case "unary+":
            case "unary^":
            case "unary!":
                return this.checkGroupsInExpression(enode.rhs, scope, flags);
            case "unary&":
                {
                    if (helper.isUnique(enode.rhs.type) && (flags & group_1.GroupCheckFlags.ForbidIsolates) != 0) {
                        throw new errors_1.TypeError("Accessing a member in an isolate is not allowed", enode.loc);
                    }
                    return this.checkGroupsInExpression(enode.rhs, scope, flags | group_1.GroupCheckFlags.ForbidIsolates);
                }
            case "unary*":
                {
                    if (helper.isUnique(enode.rhs.type) && (flags & group_1.GroupCheckFlags.ForbidIsolates) != 0) {
                        throw new errors_1.TypeError("Accessing a member in an isolate is not allowed", enode.loc);
                    }
                    let g = this.checkGroupsInExpression(enode.rhs, scope, flags);
                    if (helper.isUnique(enode.type)) {
                        return null;
                    }
                    return g;
                }
            case "+":
            case "*":
            case "/":
            case "-":
            case ">":
            case "<":
            case "<=":
            case ">=":
            case "||":
            case "&&":
            case "&":
            case "|":
            case "^":
            case "&^":
            case "%":
            case "<<":
            case ">>":
            case "==":
            case "!=":
                this.checkGroupsInExpression(enode.lhs, scope, flags);
                this.checkGroupsInExpression(enode.rhs, scope, flags);
                break;
            case ".":
                {
                    let type = helper.stripType(enode.lhs.type);
                    if (type instanceof types_1.PackageType) {
                        break;
                    }
                    if (helper.isUnique(enode.lhs.type) && (flags & group_1.GroupCheckFlags.ForbidIsolates) != 0 && enode.lhs.op != "id" && enode.lhs.op != "(") {
                        throw new errors_1.TypeError("Accessing a member in an isolate is not allowed in this place", enode.loc);
                    }
                    let g = this.checkGroupsInExpression(enode.lhs, scope, flags);
                    if (helper.isUnique(enode.type)) {
                        return null;
                    }
                    return g;
                }
            case ":":
                {
                    let g = this.checkGroupsInExpression(enode.lhs, scope, flags);
                    if (enode.parameters[0]) {
                        this.checkGroupsInExpression(enode.parameters[0], scope, (flags | group_1.GroupCheckFlags.NoSideEffects) & group_1.GroupCheckFlags.NotIsolateMask);
                    }
                    if (enode.parameters[1]) {
                        this.checkGroupsInExpression(enode.parameters[1], scope, (flags | group_1.GroupCheckFlags.NoSideEffects) & group_1.GroupCheckFlags.NotIsolateMask);
                    }
                    if (helper.isUnique(enode.lhs.type) && (flags & group_1.GroupCheckFlags.ForbidIsolates) != 0 && enode.lhs.op != "id" && enode.lhs.op != "(") {
                        throw new errors_1.TypeError("Accessing a member in an isolate is not allowed", enode.loc);
                    }
                    return g;
                }
            case "[":
                {
                    let g = this.checkGroupsInExpression(enode.lhs, scope, flags);
                    this.checkGroupsInExpression(enode.rhs, scope, (flags | group_1.GroupCheckFlags.NoSideEffects) & group_1.GroupCheckFlags.NotIsolateMask);
                    if (helper.isUnique(enode.lhs.type) && (flags & group_1.GroupCheckFlags.ForbidIsolates) != 0 && enode.lhs.op != "id" && enode.lhs.op != "(") {
                        throw new errors_1.TypeError("Accessing a member in an isolate is not allowed", enode.loc);
                    }
                    if (helper.isUnique(enode.type)) {
                        return null;
                    }
                    return g;
                }
            case "spawn":
            case "(":
                {
                    if (enode.op == "spawn") {
                        enode = enode.rhs;
                    }
                    let g = this.checkGroupsInExpression(enode.lhs, scope, flags | group_1.GroupCheckFlags.ForbidIsolates);
                    //                if (!g) {
                    //                    throw new ImplementationError()
                    //                }
                    // When calling a non-member function, the default group is determined by the first parameter.
                    if (enode.lhs.op == "id") {
                        g = null;
                    }
                    let t = helper.stripType(enode.lhs.type);
                    if (!(t instanceof types_1.FunctionType)) {
                        throw new errors_1.ImplementationError();
                    }
                    return this.checkGroupsInFunctionArguments(t, g, enode.parameters, scope, enode.loc);
                }
            case "genericInstance":
                // return this.checkGroupsInExpression(enode.lhs, scope, flags);
                return new group_1.Group(group_1.GroupKind.Free);
            case "make":
                {
                    if (enode.parameters.length > 0) {
                        this.checkGroupsInExpression(enode.parameters[0], scope, flags);
                        if (enode.parameters.length == 2) {
                            this.checkGroupsInExpression(enode.parameters[1], scope, flags);
                        }
                    }
                    // The new array has been created on the heap.
                    // Therefore, it belongs to a new free group.
                    return new group_1.Group(group_1.GroupKind.Free);
                }
            case "tuple":
                {
                    let group = null;
                    for (let p of enode.parameters) {
                        let g = this.checkGroupsInExpression(p, scope, flags & group_1.GroupCheckFlags.NotIsolateMask);
                        if (!group) {
                            group = g;
                        }
                        else {
                            group = scope.joinGroups(group, g, enode.loc, true);
                        }
                    }
                    if (!group) {
                        return new group_1.Group(group_1.GroupKind.Free);
                    }
                    return group;
                }
            case "array":
                {
                    let t = enode.type;
                    if (helper.isSlice(enode.type)) {
                        t = this.sliceArrayType(t);
                    }
                    let group = null;
                    if (enode.parameters) {
                        for (var i = 0; i < enode.parameters.length; i++) {
                            let p = enode.parameters[i];
                            if (p.op == "...") {
                                continue;
                            }
                            let g = this.checkGroupsInExpression(p, scope, flags & group_1.GroupCheckFlags.NotIsolateMask);
                            if (!group) {
                                group = g;
                            }
                            else {
                                group = scope.joinGroups(group, g, enode.loc, true);
                            }
                        }
                    }
                    if (!group) {
                        return new group_1.Group(group_1.GroupKind.Free);
                    }
                    return group;
                }
            case "object":
                {
                    let t = enode.type;
                    if (helper.isSafePointer(enode.type)) {
                        t = this.pointerElementType(t);
                    }
                    t = helper.stripType(t);
                    let group = null;
                    if (t instanceof types_1.StructType) {
                        if (enode.parameters) {
                            for (let p of enode.parameters) {
                                let g = this.checkGroupsInExpression(p.lhs, scope, flags & group_1.GroupCheckFlags.NotIsolateMask);
                                if (!group) {
                                    group = g;
                                }
                                else {
                                    group = scope.joinGroups(group, g, enode.loc, true);
                                }
                            }
                        }
                    }
                    else if (t instanceof types_1.MapType) {
                        throw new errors_1.TodoError();
                    }
                    if (!group) {
                        return new group_1.Group(group_1.GroupKind.Free);
                    }
                    return group;
                }
            /*
            case "=>":
            {
                let f = new Function();
                f.loc = enode.loc;
                f.node = enode;
                f.scope = new Scope(scope);
                f.scope.func = f;
                f.type = new FunctionType();
                f.type.loc = enode.loc;
                enode.scope = f.scope;
                enode.type = f.type;
                if (enode.parameters) {
                    for(let pnode of enode.parameters) {
                        let original_pnode = pnode;
                        var p = new FunctionParameter();
                        if (pnode.op == "ellipsisParam") {
                            p.ellipsis = true;
                            pnode = pnode.lhs;
                        }
                        p.name = pnode.name.value;
                        for(let param of f.type.parameters) {
                            if (param.name == p.name) {
                                throw new TypeError("Duplicate parameter name " + p.name, pnode.loc);
                            }
                        }
                        p.type = this.injectVariableBoxes(this.createType(pnode, enode.scope, "parameter"), "parameter");
                        if (p.ellipsis && !(p.type instanceof SliceType)) {
                            throw new TypeError("Ellipsis parameters must be of a slice type", pnode.loc);
                        }
                        p.loc = pnode.loc;
                        f.type.parameters.push(p);
                        f.scope.registerElement(p.name, p);
                    }
                }
                if (enode.lhs) {
                    f.type.returnType = this.createType(enode.lhs, f.scope, "parameter");
                    if (enode.lhs.op == "tupleType") {
                        for(let i = 0; i < enode.lhs.parameters.length; i++) {
                            let pnode = enode.lhs.parameters[i];
                            if (pnode.name) {
                                let v = new Variable();
                                v.isResult = true;
                                v.loc = pnode.loc;
                                v.name = pnode.name.value;
                                v.type = this.injectVariableBoxes((f.type.returnType as TupleType).types[i], "variable");
                                f.scope.registerElement(v.name, v);
                                f.namedReturnTypes = true;
                            }
                        }
                    }
                }
                if (enode.rhs) {
                    this.checkExpression(enode.rhs, enode.scope);
                    if (!f.type.returnType) {
                        f.type.returnType = enode.rhs.type;
                    } else {
                        this.checkIsAssignableNode(f.type.returnType, enode.rhs);
                    }
                } else {
                    for(let s of enode.statements) {
                        this.checkStatement(s, enode.scope);
                    }
                }
                break;
            }
            */
            case "is":
                {
                    this.checkGroupsInExpression(enode.lhs, scope, flags);
                    break;
                }
            case "typeCast":
                {
                    // Converting a slice of bytes to a string, freezes the slice.
                    // This is similar to assigning the slice to a unique slice.
                    if (enode.type == Static_1.Static.t_string && helper.isSlice(enode.rhs.type)) {
                        let g = this.checkGroupsInExpression(enode.rhs, scope, flags);
                        this.checkGroupsInSingleAssignment(new types_1.SliceType(new types_1.ArrayType(Static_1.Static.t_byte, -1), "unique"), null, g, enode.rhs, false, scope, enode.loc);
                        return null;
                    }
                    if (helper.isSlice(enode.type) && enode.rhs.type == Static_1.Static.t_string && !helper.isConst(enode.type)) {
                        // Converting a string to a non-const slice implies copying it.
                        // The returned copy belongs to a free group.
                        return new group_1.Group(group_1.GroupKind.Free);
                    }
                    return this.checkGroupsInExpression(enode.rhs, scope, flags);
                }
            case "take":
                if ((flags & group_1.GroupCheckFlags.NoSideEffects) != 0) {
                    throw new errors_1.TypeError("Expression with side effects is not allowed in this place", enode.loc);
                }
                return this.checkGroupsInExpression(enode.lhs, scope, flags);
            case "cap":
            case "len":
                this.checkGroupsInExpression(enode.lhs, scope, flags);
                return null;
            case "sizeof":
            case "aligned_sizeof":
            case "max":
            case "min":
                return null;
            case "clone":
                this.checkGroupsInExpression(enode.lhs, scope, flags);
                return new group_1.Group(group_1.GroupKind.Free);
            case "pop":
                return this.checkGroupsInExpression(enode.lhs, scope, flags);
            case "push":
            case "tryPush":
            case "append":
                {
                    let group;
                    let ltype;
                    for (let i = 0; i < enode.parameters.length; i++) {
                        let p = enode.parameters[i];
                        let g = this.checkGroupsInExpression(p.op == "unary..." ? p.rhs : p, scope, group_1.GroupCheckFlags.None);
                        if (i == 0) {
                            group = g;
                            ltype = types_1.RestrictedType.strip(p.type).getElementType();
                        }
                        else {
                            if (p.op == "unary...") {
                                if (!helper.isPureValue(ltype)) {
                                    // TODO: Could be realized. Might need to zero the slice
                                    throw new errors_1.TypeError("Appending a slice of pointers is not supported.", p.loc);
                                }
                            }
                            else {
                                // Appending a pointer-like type? Then we must care about the groups involved
                                if (helper.isSafePointer(p.type) || helper.isSlice(p.type)) {
                                    this.checkGroupsInSingleAssignment(ltype, group, g, p, false, scope, p.loc);
                                }
                            }
                        }
                    }
                    if (enode.op == "append") {
                        return group;
                    }
                    return null;
                }
            case "resume":
                this.checkGroupsInExpression(enode.lhs, scope, flags);
                return null;
            case "coroutine":
                return this.globalGroup;
            default:
                throw new errors_1.ImplementationError(enode.op);
        }
        return null;
    }
    checkGroupsInFunctionArguments(ft, defaultGroup, args, scope, loc) {
        let groups = new Map();
        if (defaultGroup) {
            groups.set("default", defaultGroup);
        }
        if (args) {
            for (let i = 0; i < args.length; i++) {
                let rnode;
                let ltype;
                let param;
                if (args[i].op == "unary...") {
                    param = ft.lastParameter();
                    rnode = args[i].rhs;
                    ltype = param.type;
                }
                else {
                    if (ft.hasEllipsis() && i >= ft.parameters.length - 1) {
                        param = ft.lastParameter();
                        ltype = this.arrayElementType(this.sliceArrayType(param.type));
                    }
                    else {
                        param = ft.parameters[i];
                        ltype = param.type;
                    }
                    rnode = args[i];
                }
                let name = param.type.groupName;
                if (!name || name == "") {
                    name = "default";
                }
                let g;
                if (helper.isUnique(ltype)) {
                    g = null;
                }
                else if (groups.has(name)) {
                    g = groups.get(name);
                }
                else {
                    g = new group_1.Group(group_1.GroupKind.Free, param.name);
                    groups.set(name, g);
                }
                this.checkGroupsInSingleAssignment(ltype, g, null, rnode, false, scope, loc);
            }
        }
        // Void function?
        if (!ft.returnType || ft.returnType == Static_1.Static.t_void) {
            return null;
        }
        // The function returns a tuple type?
        if (ft.returnType instanceof types_1.TupleType) {
            let tupleg = new group_1.TupleGroup(group_1.GroupKind.Free);
            for (let i = 0; i < ft.returnType.types.length; i++) {
                let t = ft.returnType.types[i];
                let name = t.groupName;
                if (!name || name == "") {
                    name = "default";
                }
                let g;
                if (helper.isUnique(t)) {
                    g = new group_1.Group(group_1.GroupKind.Free);
                }
                else if (groups.has(name)) {
                    g = groups.get(name);
                }
                else {
                    let kind = group_1.GroupKind.Free;
                    if (TypeChecker.hasReferenceOrStrongPointers(t)) {
                        kind = group_1.GroupKind.Bound;
                    }
                    g = new group_1.Group(kind);
                    groups.set(name, g);
                }
                tupleg.groups.push(g);
            }
            return tupleg;
        }
        // Only one type is returned
        let name = ft.returnType.groupName;
        if (!name || name == "") {
            name = "default";
        }
        let g;
        if (groups.has(name) && !helper.isUnique(ft.returnType)) {
            return groups.get(name);
        }
        let kind = group_1.GroupKind.Free;
        if (TypeChecker.hasReferenceOrStrongPointers(ft.returnType)) {
            kind = group_1.GroupKind.Bound;
        }
        return new group_1.Group(kind, "return");
    }
    hasTemplateInstantiations() {
        return (this.templateFunctionInstantiations.size != 0 || this.templateTypeInstantiations.size != 0);
    }
}
exports.TypeChecker = TypeChecker;
//# sourceMappingURL=typecheck.js.map